#pragma once


//
// This class handles a track, ie. the list of keys which make an 'animation'
//
class CAnimTrack
{
public:

    // Constructors / destructors
    CAnimTrack(DWORD dwNumKeys, DWORD dwNumFrames, DWORD dwFrameRate, FLOAT fAnimFPS);
    virtual ~CAnimTrack()                               { m_ListKeys.clear(); }

    // Accessors
    DWORD           dwGetNumKeys()                      { return m_dwNumKeys; }
    DWORD           dwGetNumFrames()                    { return m_dwNumFrames; }
    DWORD           dwGetCurKey()                       { return m_dwCurKey; }
    FLOAT           fGetCurFrame()                      { return m_fCurFrame; }
    FLOAT           fGetKeyFromFrame(FLOAT fFrame)      { return fFrame / m_dwFrameRate; }
    FLOAT           fGetInterpolationFactor()           { return m_fInterpFactor; }
    FLOAT           fGetOverallSpeed()                  { return m_fOverallSpeed; }
    FLOAT           fGetAnimFPS()                       { return m_fAnimFPS; }
    DWORD           dwGetFrameRate()                    { return m_dwFrameRate; }
    CAnimKey*       lpGetKey(DWORD dwKeyIndex)          { ASSERT(dwKeyIndex >= 0  &&  dwKeyIndex < m_dwNumKeys); return &m_ListKeys[dwKeyIndex]; }

    void            vSetOverallSpeed(FLOAT fOverSpeed)  { ASSERT(fOverSpeed >= (FLOAT)0.0); m_fOverallSpeed = fOverSpeed; }
    void            vSetInterpolationFactor(FLOAT fFact){ m_fInterpFactor = fFact; }
    void            vActivateInterpolation(bool boActiv){ m_boActivateInterpolation = boActiv;  }
    void            vPushKey(CAnimKey& Key)             { Key.vPackData(); m_ListKeys.push_back(Key); }
    void            vPackKeys()                         { m_ListKeys.resize(m_ListKeys.size()); }

    // Processors
    virtual bool    boSetCurrentFrame(FLOAT fFrame);
    virtual bool    boRunForward(FLOAT fElapsedTime);
    void            vGetInterpolatedData(DWORD dwNumData,CAnimData* lpData,bool boDetectRecursInfinite = false);
    void            vGetInterpolatedBB(CBoundingBox* lpBB,bool boDetectRecursInfinite = false);

    void            vPushState()                        { m_dwCurKeySave = m_dwCurKey; m_fCurFrameSave = m_fCurFrame; m_fInterpFactorSave = m_fInterpFactor; m_boNextKeyIsInCurrentAnimSave = m_boNextKeyIsInCurrentAnim; }
    void            vPopState()                         { m_dwCurKey = m_dwCurKeySave; m_fCurFrame = m_fCurFrameSave; m_fInterpFactor = m_fInterpFactorSave; m_boNextKeyIsInCurrentAnim = m_boNextKeyIsInCurrentAnimSave; }

protected:

    typedef std::vector<CAnimKey> AnimListKeys;

    void            vSetNextAnim(CAnimTrack* lpAnim, DWORD dwFrame);

    // Variables provided by user
    DWORD           m_dwNumKeys;                        // num keys in the animation
    DWORD           m_dwNumFrames;                      // num frames in the animation
    FLOAT           m_fAnimFPS;                         // play speed of the animation, in frames per second
    FLOAT           m_fOverallSpeed;                    // allow changes in animation speed without modifying m_fAnimFPS. Real anim speed is m_fAnimFPS * m_fOverallSpeed
    CAnimTrack*     m_lpNextAnim;                       // next animation to chain to at the end of the animation
    DWORD           m_dwNextAnimFrame;                  // frame in the next animation to chain to
    bool            m_boActivateInterpolation;          // true if interpolation is on between keys, else false
    AnimListKeys    m_ListKeys;                         // list of keys of the animation

    // Variables set before animating
    DWORD           m_dwFrameRate;                      // num frames + 1 between two keys
    DWORD           m_dwRemainingFrames;                // num frames in the animation after the last key

    // Variables used during animation playing
    DWORD           m_dwCurKey;                         // current key of the animation
    FLOAT           m_fCurFrame;                        // current frame
    FLOAT           m_fInterpFactor;                    // Interpolation factor to use to blend current key with next key
    bool            m_boNextKeyIsInCurrentAnim;         // true if the next key is in the current animation, false if it's in m_lpNextAnim

    DWORD           m_dwCurKeySave;                     // used by push / pop state functions
    FLOAT           m_fCurFrameSave;                    // to save and retrieve the state
    FLOAT           m_fInterpFactorSave;                // of the track
    bool            m_boNextKeyIsInCurrentAnimSave;     //

    friend class CAnimMgr;
};
