#include "stdafx.h"
#include "TRViewer.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "AnimMgr.h"
#include "TRAnimTrack.h"


#define CRLF        "\15\12"

typedef std::map<DWORD,DWORD> MAPOBJID2NUMANIM;


void	vDumpVertex( TR2_VERTEX& v, std::string& str, bool boAddEOL )
{
	CString t;

	t.Format( "Coord: %ld, %ld, %ld", (long) v.sX, (long) v.sY, (long) v.sZ );

	str += t;

	if( boAddEOL ) str += CRLF;
}


void	vReadVertex( TR2_VERTEX& v, FILE* f, bool boAddEOL )
{
	long l1,l2,l3;

	fscanf( f, "Coord: %ld, %ld, %ld\n", &l1, &l2, &l3 );

	v.sX = l1;
	v.sY = l2;
	v.sZ = l3;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),f); }
}


void	vDumpFace3( TR2_FACE3& f, std::string& str, bool boDumpExtraIndex, bool boAddEOL )
{
	CString t;

	t.Format( "Indexes: %ld, %ld, %ld - TextureIndex: %ld", 
		(long) f.awVertices[0], (long) f.awVertices[1], (long) f.awVertices[2],
		f.wTextureIndex);

	str += t;

	if( boDumpExtraIndex )
		{
		t.Format( " - ExtraTextureIndex: %ld", f.wExtraTextureIndex );

		str += t;
		}

	if( boAddEOL ) str += CRLF;
}


void	vReadFace3( TR2_FACE3& f, FILE* ff, bool boDumpExtraIndex, bool boAddEOL )
{
	long l1, l2, l3, l4, l5;

	if( boDumpExtraIndex )
		{
		fscanf( ff, "Indexes: %ld, %ld, %ld - TextureIndex: %ld - ExtraTextureIndex: %ld", 
			&l1, &l2, &l3, &l4, &l5
			);

		f.wExtraTextureIndex = l5;
		}
	else
		{
		fscanf( ff, "Indexes: %ld, %ld, %ld - TextureIndex: %ld", 
			&l1, &l2, &l3, &l4
			);
		}

	f.awVertices[0] = l1;
	f.awVertices[1] = l2;
	f.awVertices[2] = l3;
	f.wTextureIndex = l4;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),ff); }
}


void	vDumpFace4( TR2_FACE4& f, std::string& str, bool boDumpExtraIndex, bool boAddEOL )
{
	CString t;

	t.Format( "Indexes: %ld, %ld, %ld, %ld - TextureIndex: %ld", 
		(long) f.awVertices[0], (long) f.awVertices[1], (long) f.awVertices[2], (long) f.awVertices[3],
		f.wTextureIndex);

	str += t;

	if( boDumpExtraIndex )
		{
		t.Format( " - ExtraTextureIndex: %ld", f.wExtraTextureIndex );

		str += t;
		}

	if( boAddEOL ) str += CRLF;
}


void	vReadFace4( TR2_FACE4& f, FILE* ff, bool boDumpExtraIndex, bool boAddEOL )
{
	long l1, l2, l3, l4, l5, l6;

	if( boDumpExtraIndex )
		{
		fscanf( ff, "Indexes: %ld, %ld, %ld, %ld - TextureIndex: %ld - ExtraTextureIndex: %ld", 
			&l1, &l2, &l3, &l6, &l4, &l5
			);

		f.wExtraTextureIndex = l5;
		}
	else
		{
		fscanf( ff, "Indexes: %ld, %ld, %ld, %ld - TextureIndex: %ld", 
			&l1, &l2, &l3, &l6, &l4
			);
		}

	f.awVertices[0] = l1;
	f.awVertices[1] = l2;
	f.awVertices[2] = l3;
	f.awVertices[3] = l6;
	f.wTextureIndex = l4;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),ff); }
}


void	vDumpWord( short s, std::string& str, bool boHexa, bool boAddEOL )
{
	CString t;

	if( boHexa )
		t.Format( "%.8x", s );
	else
		t.Format( "%ld", (long) s );

	str += t;

	if( boAddEOL ) str += CRLF;
}


void	vReadWord( short& s, FILE* ff, bool boHexa, bool boAddEOL )
{
	long l1;

	if( boHexa )
		{
		fscanf( ff, "%x", &l1 );
		}
	else
		{
		fscanf( ff, "%ld", &l1 );
		}

	s = l1;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),ff); }
}


void	vDumpLong( long l, std::string& str, bool boHexa, bool boAddEOL )
{
	CString t;

	if( boHexa )
		t.Format( "%.8x", l);
	else
		t.Format( "%ld", l);

	str += t;

	if( boAddEOL ) str += CRLF;
}


void	vReadLong( long& l, FILE* ff, bool boHexa, bool boAddEOL )
{
	long l1;

	if( boHexa )
		{
		fscanf( ff, "%x", &l1 );
		}
	else
		{
		fscanf( ff, "%ld", &l1 );
		}

	l = l1;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),ff); }
}



void	vDumpMesh( TR123_MESH* lpMesh, std::string& str )
{
	CString t;

	str += "Center: ";	vDumpVertex( lpMesh->Center, str, true );
	str += "lCollisionSize: "; vDumpLong( lpMesh->lCollisionSize, str, false, true );
	str += "sNumVertices: "; vDumpWord( lpMesh->sNumVertices, str, false, true );

	for( bit16 i = 0; i < lpMesh->sNumVertices; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		vDumpVertex( lpMesh->lpVertices[i], str, true );
		}

	str += "Lighting.LightingType: "; vDumpWord( (WORD) lpMesh->Lighting.LightingType, str, false, true );
	str += "Lighting.sNumData: "; vDumpWord( lpMesh->Lighting.sNumData, str, false, true );

	for( i = 0; i < lpMesh->Lighting.sNumData; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		if( lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL )
			{
			vDumpWord( lpMesh->Lighting.lpLights[i], str, false, true );
			}
		else
			{
			vDumpVertex( (TR2_VERTEX&)(lpMesh->Lighting.lpLights[i]), str, true );
			}
		}

	str += "sNumTexturedRectangles: "; vDumpWord( lpMesh->sNumTexturedRectangles, str, false, true );

	for( i = 0; i < lpMesh->sNumTexturedRectangles; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		vDumpFace4( lpMesh->lpTexturedRectangles[i], str, true, true );
		}

	str += "sNumTexturedTriangles: "; vDumpWord( lpMesh->sNumTexturedTriangles, str, false, true );

	for( i = 0; i < lpMesh->sNumTexturedTriangles; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		vDumpFace3( lpMesh->lpTexturedTriangles[i], str, true, true );
		}

	str += "sNumColoredRectangles: "; vDumpWord( lpMesh->sNumColoredRectangles, str, false, true );

	for( i = 0; i < lpMesh->sNumColoredRectangles; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		vDumpFace4( lpMesh->lpColoredRectangles[i], str, true, true );
		}

	str += "sNumColoredTriangles: "; vDumpWord( lpMesh->sNumColoredTriangles, str, false, true );

	for( i = 0; i < lpMesh->sNumColoredTriangles; ++i )
		{
		t.Format( "    #%.4d ", i );
		str += t;
		vDumpFace3( lpMesh->lpColoredTriangles[i], str, true, true );
		}

}


void	vReadMesh( TR123_MESH* lpMesh, FILE* ff )
{
	long l;

	fscanf( ff, "Center: " );	vReadVertex( lpMesh->Center, ff, false );
	fscanf( ff, "lCollisionSize: " ); vReadLong( lpMesh->lCollisionSize, ff, false, true );
	fscanf( ff, "sNumVertices: "); vReadWord( lpMesh->sNumVertices, ff, false, true );

	for( bit16 i = 0; i < lpMesh->sNumVertices; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		vReadVertex( lpMesh->lpVertices[i], ff, false );
		}

	fscanf( ff, "Lighting.LightingType: " ); vReadWord( *(short*)(&lpMesh->Lighting.LightingType), ff, false, true );
	fscanf( ff, "Lighting.sNumData: " ); vReadWord( lpMesh->Lighting.sNumData, ff, false, true );

	for( i = 0; i < lpMesh->Lighting.sNumData; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		if( lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL )
			{
			vReadWord( lpMesh->Lighting.lpLights[i], ff, false, false );
			}
		else
			{
			vReadVertex( (TR2_VERTEX&)(lpMesh->Lighting.lpLights[i]), ff, false );
			}
		}

	fscanf( ff, "sNumTexturedRectangles: "); vReadWord( lpMesh->sNumTexturedRectangles, ff, false, true );

	for( i = 0; i < lpMesh->sNumTexturedRectangles; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		vReadFace4( lpMesh->lpTexturedRectangles[i], ff, true, false );
		}

	fscanf( ff, CRLF );
	fscanf( ff, "sNumTexturedTriangles: "); vReadWord( lpMesh->sNumTexturedTriangles, ff, false, true );

	for( i = 0; i < lpMesh->sNumTexturedTriangles; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		vReadFace3( lpMesh->lpTexturedTriangles[i], ff, true, false );
		}

	fscanf( ff, "sNumColoredRectangles: "); vReadWord( lpMesh->sNumColoredRectangles, ff, false, true );

	for( i = 0; i < lpMesh->sNumColoredRectangles; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		vReadFace4( lpMesh->lpColoredRectangles[i], ff, true, false );
		}

	fscanf( ff, "sNumColoredTriangles: "); vReadWord( lpMesh->sNumColoredTriangles, ff, false, true );

	for( i = 0; i < lpMesh->sNumColoredTriangles; ++i )
		{
		fscanf( ff, "    #%ld ", &l );
		vReadFace3( lpMesh->lpColoredTriangles[i], ff, true, false );
		}
}


void	vDumpObjTextVert( TR2_OBJECT_TEXTURE_VERT& ot, std::string& str, bool boAddEOL )
{
	CString t;

	t.Format( "(%ld,%ld)", ot.bXPixel, ot.bYPixel );

	str += t;

	if( boAddEOL ) str += CRLF;
}


void	vReadObjTextVert( TR2_OBJECT_TEXTURE_VERT& ot, FILE* ff, bool boAddEOL )
{
	long l1, l2;

	fscanf( ff, "(%ld,%ld)", &l1, &l2 );

	ot.bXPixel = l1;
	ot.bYPixel = l2;

	if( boAddEOL ) { char sz[512]; fgets(sz,sizeof(sz),ff); }
}


void	vDumpObjText( TR2_OBJECT_TEXTURE* lpObjText, std::string& str, bool boAddEOL )
{
	str += "wAttribute: "; vDumpWord( lpObjText->wAttribute, str, true, false );
	str += ", wTile: "; vDumpWord( lpObjText->wTile, str, true, false );
	str += ", wUnknown: "; vDumpWord( lpObjText->Unknown, str, true, false );

	str += ", Vertices: ";

	vDumpObjTextVert( lpObjText->Vertices[0], str, false );
	vDumpObjTextVert( lpObjText->Vertices[1], str, false );
	vDumpObjTextVert( lpObjText->Vertices[2], str, false );
	vDumpObjTextVert( lpObjText->Vertices[3], str, false );

	str += ", XOfs/YOfs/XSize/YSize: ";	vDumpLong( lpObjText->dwXOffset, str, false, false );
	str += ","; vDumpLong( lpObjText->dwYOffset, str, false, false );
	str += ","; vDumpLong( lpObjText->dwXSize, str, false, false );
	str += ","; vDumpLong( lpObjText->dwYSize, str, false, true );
}


void	vReadObjText( TR2_OBJECT_TEXTURE* lpObjText, FILE* ff, bool boAddEOL )
{
	long l;
	short s;

	fscanf( ff, "wAttribute: "); vReadWord( s, ff, true, false ); lpObjText->wAttribute = s;
	fscanf( ff, ", wTile: "); vReadWord( s, ff, true, false ); lpObjText->wTile = s;
	fscanf( ff, ", wUnknown: "); vReadWord( s, ff, true, false ); lpObjText->Unknown = s;

	fscanf( ff, ", Vertices: ");

	vReadObjTextVert( lpObjText->Vertices[0], ff, false );
	vReadObjTextVert( lpObjText->Vertices[1], ff, false );
	vReadObjTextVert( lpObjText->Vertices[2], ff, false );
	vReadObjTextVert( lpObjText->Vertices[3], ff, false );

	fscanf( ff, ", XOfs/YOfs/XSize/YSize: ");	vReadLong( l, ff, false, false ); lpObjText->dwXOffset = l;
	fscanf( ff, ","); vReadLong( l, ff, false, false ); lpObjText->dwYOffset = l;
	fscanf( ff, ","); vReadLong( l, ff, false, false ); lpObjText->dwXSize = l;
	fscanf( ff, ","); vReadLong( l, ff, false, true ); lpObjText->dwYSize = l;
}


void vParseDirectory(MAPOBJID2NUMANIM& Map,TCHAR* lpszFilter)
{
    WIN32_FIND_DATA fd;

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    while(FindNextFile(h,&fd))
        {
        if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
            {
            CTRFile trFile;

            if(trFile.boLoad(fd.cFileName))
                {
                TR_FILE& m_trFile = trFile.GetTRFile();

                for(DWORD dwI = 0; dwI < m_trFile.tr123File.dwNumMoveables; dwI++)
                    {
                    TR2_MOVEABLE* lpMoveable = &m_trFile.tr123File.lpMoveables[dwI];

                    DWORD   dwObjId = lpMoveable->dwObjectId;
                    WORD    wStartAnim = lpMoveable->wAnimation;
                    WORD    wNextStartAnim;

                    if(dwI != m_trFile.tr123File.dwNumMoveables - 1)
                        {
                        wNextStartAnim = m_trFile.tr123File.lpMoveables[dwI+1].wAnimation;
                        if(wNextStartAnim == 0xFFFF)
                            {
                            if((dwI + 1) < m_trFile.tr123File.dwNumMoveables - 1)
                                {
                                wNextStartAnim = m_trFile.tr123File.lpMoveables[dwI+2].wAnimation;
                                }
                            else
                                {
                                wNextStartAnim = m_trFile.tr123File.dwNumAnimations;
                                }
                            }
                        }
                    else
                        wNextStartAnim = m_trFile.tr123File.dwNumAnimations;

                    WORD    wNumAnim = (wStartAnim != 0xFFFF) ? wNextStartAnim - wStartAnim : -1;
                    //char    szBuff[512];
                    //DWORD   dw;

                    MAPOBJID2NUMANIM::iterator  theIterator = Map.find(dwObjId);
                    if(theIterator == Map.end())
                        {
                        CString s;
                        s.Format(_T("Found a moveable (#%d) with an Id too large (#%d) !"),dwI,dwObjId);
                        AfxMessageBox(s);
                        }
                    else
                        {
                        DWORD   dwCurNum = theIterator->second;
                        if(dwCurNum == 0x8000)
                            {
                            }
                        else if(dwCurNum != wNumAnim)
                            {
                            CString s;
                            s.Format(_T("Found a moveable (#%d) with num anims different from a level to another (#%d in previous level / #%d in %s)"),dwObjId,dwCurNum,(DWORD)wNumAnim,fd.cFileName);
                            AfxMessageBox(s);
                            }
                        Map[dwObjId] = wNumAnim;
                        }
                    }
                }
            else
                {
                CString s;
                s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                AfxMessageBox(s);
                }
            }
        }

    FindClose(h);
}


bool    boChooseFolder(TCHAR* lpszBuf)
{
    BROWSEINFO   bi;
    ZeroMemory(&bi,sizeof(BROWSEINFO));

    TCHAR szBuf[MAX_PATH];
    lstrcpy(szBuf, __TEXT("Choose a folder to load the levels from:"));

    bi.hwndOwner  = NULL;
    bi.ulFlags    = 0;
    bi.lpszTitle  = static_cast<LPCSTR>(szBuf);
    bi.lpfn       = NULL;
    bi.lParam     = NULL;

    LPITEMIDLIST  pidl = SHBrowseForFolder(&bi);

    if(!pidl) return false;

    SHGetPathFromIDList(pidl,lpszBuf);

    return true;
}


void vOnGenerateNumAnims() 
{
    TCHAR szBuf[MAX_PATH];

    if(!boChooseFolder(szBuf)) return;

    CWaitCursor dummy;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(szBuf);

    MAPOBJID2NUMANIM Map;

    for(DWORD dwI = 0; dwI < 10000; dwI++)
        {
        Map[dwI] = 0x8000;
        }

    vParseDirectory(Map,_T("*.phd"));
    vParseDirectory(Map,_T("*.tr2"));

    SetCurrentDirectory(szCurDir);
}


void vParseDirectoryDebug(TCHAR* lpszFilter)
{
    WIN32_FIND_DATA fd;

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    while(FindNextFile(h,&fd))
        {
        if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
            {
            CTRFile trFile;

            if(trFile.boLoad(fd.cFileName))
                {
                TR_FILE& m_trFile = trFile.GetTRFile();

                for(DWORD dwI = 0; dwI < m_trFile.tr123File.dwNumStaticMeshes; dwI++)
                    {
                    TR2_STATICMESH* lpStaticMesh = &m_trFile.tr123File.lpStaticMeshes[dwI];
                    if(lpStaticMesh->wFlags < 2)
                        {
                        CString s;
                        s.Format(_T("Level %s: static mesh #%ld has flags %ld"),fd.cFileName,dwI,(long)lpStaticMesh->wFlags);
                        AfxMessageBox(s);
                        }
                    }
                }
            else
                {
                CString s;
                s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                AfxMessageBox(s);
                }
            }
        }

    FindClose(h);
}


void vCheckStaticMeshesFlags(TCHAR* lpszBuf)
{
    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    vParseDirectoryDebug(_T("*.phd"));
    vParseDirectoryDebug(_T("*.tr2"));

    SetCurrentDirectory(szCurDir);
}


void vCountSfx(TCHAR* lpszBuf)
{
    TCHAR szBuf[MAX_PATH];

    lstrcpy(szBuf,lpszBuf);
    lstrcat(szBuf,_T("\\main.sfx"));

    HANDLE hFile = CreateFile(szBuf,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_ARCHIVE,NULL);

    if(hFile != INVALID_HANDLE_VALUE)
        {
        DWORD   dwFileSize = GetFileSize(hFile,NULL);
        BYTE*   lpBuf = new BYTE [dwFileSize];
        BYTE*   lpBufSave = lpBuf;
        DWORD   dwNbRead;
        
        ReadFile(hFile,lpBuf,dwFileSize,&dwNbRead,NULL);

        DWORD   dwNumSfx = 0;

        while(dwFileSize)
            {
            lpBuf += 4; // skip RIFF
            DWORD dwSfxSize = *((DWORD*) lpBuf); lpBuf += 4 + dwSfxSize;
            dwFileSize -= 8 + dwSfxSize;
            dwNumSfx++;
            }

        delete [] lpBufSave;

        CString s;

        s.Format(_T("Num sfx:%ld"),dwNumSfx);
        AfxMessageBox(s);
        }

    CloseHandle(hFile);
}


void vExportAnmText(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr2");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

                    bitu16* lpsAnimatedTextures     = m_trFile.tr123File.lpAnimatedTextures;
                    WORD    wNumAnimatedTextures    = *lpsAnimatedTextures++;

                    DWORD dwNumText = 0;

                    while(wNumAnimatedTextures--)
                        {
                        bitu16   wNum = *lpsAnimatedTextures;  wNum++;

                        CTrObjGfx trAnmText(&trFile,lpsAnimatedTextures);

                        char sz[123];
                        sprintf(sz,"d:\\temp\\AnmText%.2d_%.2d.tga",dwNiv,dwNumText++);

                        trAnmText.vSaveTextureAsTGA(sz);

                        lpsAnimatedTextures += 1 + wNum;
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vExportSprSeq(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr2");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

                    for(DWORD dwI = 0; dwI < m_trFile.tr123File.dwNumSpriteSequences; dwI++)
                        {
                        TR2_SPRITE_SEQUENCE* lpSprSeq = &m_trFile.tr123File.lpSpriteSequences[dwI];

                        CTrLstSpr trSpr(lpSprSeq,&trFile);

                        char sz[123];
                        sprintf(sz,"d:\\temp\\Spr%.2d_%.2d.tga",dwNiv,dwI);

                        trSpr.vSaveTextureAsTGA(sz);
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vCountMeshes(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr2");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE&    m_trFile = trFile.GetTRFile();
                    enumTRType  trType = trFile.GetTRType();

                    dwNiv++;

                    DWORD dwNumMeshes = m_trFile.tr123File.dwNumMeshes;
                    DWORD dwNumMeshesCalc = 0;
                    DWORD dwNumMeshesCalc2 = 0;
                    DWORD dwNumMeshPointers = m_trFile.tr123File.dwNumMeshPointers;
                    DWORD dwNumMeshPointersCalc = 0;

                    DWORD dwNumZero = 0;
                    DWORD dwNumZero2 = 0;

                    DWORD* lpd = new DWORD [dwNumMeshPointers];

                    ZeroMemory(lpd,sizeof(DWORD) * dwNumMeshPointers);

                    DWORD dwI = 0;

                    TRListMoveable* lpList = m_trFile.tr123File.lpListMoveable;
                    for(TRListMoveable::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
                        {
                        CTrMvb* lpMvb = *theIterator;

                        dwNumMeshPointersCalc += lpMvb->dwGetNumMeshes();

                        DWORD   dwObjectId = lpMvb->dwGetObjectId();
                        bool    boIsLara =  ((trType == TR3_FILE) && dwObjectId == 315)  ||
                                            ((trType != TR3_FILE) && dwObjectId == 0);

                        TR2_MOVEABLE* lpMoveable = &m_trFile.tr123File.lpMoveables[dwI++];

                        for(DWORD dwJ = 0; dwJ < lpMvb->dwGetNumMeshes(); dwJ++)
                            {
                            lpd[lpMoveable->wStartingMesh + dwJ]++;

                            if(m_trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + dwJ] != 0) dwNumMeshesCalc++; else dwNumZero2++;
                            }
                        }

                    dwNumMeshesCalc++;  // take into account the real mesh #0

                    for(dwI = 0; dwI < m_trFile.tr123File.dwNumStaticMeshes; dwI++)
                        {
                        TR2_STATICMESH* lpStaticMesh = &m_trFile.tr123File.lpStaticMeshes[dwI];
                        if(m_trFile.tr123File.lpMeshPointers[lpStaticMesh->wMesh] == 0)
                            {
                            dwNumZero2++;
                            dwNumMeshesCalc--;
                            }
                        lpd[lpStaticMesh->wMesh]++;
                        }

                    CString smp("");
                    bool    boErr = false;

                    for(dwI = 0; dwI < dwNumMeshPointers; dwI++)
                        {
                        if(m_trFile.tr123File.lpMeshPointers[dwI] != 0) dwNumMeshesCalc2++; else dwNumZero++;
                        if(lpd[dwI] == 0)
                            {
                            smp.Format("%s\nMesh pointer #%ld unused",smp,dwI);
                            boErr = true;
                            }
                        if(lpd[dwI] > 1)
                            {
                            smp.Format("%s\nMesh pointer #%ld used %ld times",smp,lpd[dwI]);
                            boErr = true;
                            }
                        }

                    delete [] lpd;

                    dwNumMeshesCalc2++;

                    dwNumMeshPointersCalc += m_trFile.tr123File.dwNumStaticMeshes;
                    dwNumMeshesCalc += m_trFile.tr123File.dwNumStaticMeshes;

                    if(dwNumMeshesCalc != dwNumMeshes  ||  dwNumMeshPointersCalc != dwNumMeshPointers  ||  boErr)
                        {
                        CString s;

                        s.Format("Level: %s\n\nNum meshes: %ld - %ld - %ld\nNum zero mesh: %ld - %ld\nNum mesh pointers: %ld - %ld\n%s\n\nContinue ?",
                                fd.cFileName,dwNumMeshes,dwNumMeshesCalc,dwNumMeshesCalc2,
                                dwNumZero,dwNumZero2,
                                dwNumMeshPointers,dwNumMeshPointersCalc,smp);

                        if(AfxMessageBox(s,MB_YESNO) == IDNO) return;
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vCheckAnimCommands(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr2");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

                    // Only load the level: if there is an unknown command in a anim command list,
                    // the prog will abort (displaying first a msgbox with the unknown command) !
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


#define CMP_SIMPLE(lp1,lp2,prop,txt) \
    if(lp1->prop != lp2->prop) \
        { \
        t = _T(" - " txt " !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false; \
        } \



bool    boCompareQuad(TR2_FACE4* lpFace1,TR2_FACE4* lpFace2,bool boTextured,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    if(!boTextured  &&  (lpFace1->wTextureIndex != lpFace2->wTextureIndex))
        {
        t = _T(" - wTextureIndex on a quad face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }

    for(BYTE bJ = 0; bJ < 4; bJ++)
        {
        if(lpFace1->awVertices[bJ] != lpFace2->awVertices[bJ])
            {
            t = _T(" - vertex index on a quad face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        }

    return boOk;
}


bool    boCompareTri(TR2_FACE3* lpFace1,TR2_FACE3* lpFace2,bool boTextured,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    if(!boTextured  &&  (lpFace1->wTextureIndex != lpFace2->wTextureIndex))
        {
        t = _T(" - wTextureIndex on a tri face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }

    for(BYTE bJ = 0; bJ < 3; bJ++)
        {
        if(lpFace1->awVertices[bJ] != lpFace2->awVertices[bJ])
            {
            t = _T(" - vertex index on a tri face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        }

    return boOk;
}


bool    boCompareQuad(TR2_FACE4* lpFace1,TR2_FACE4* lpFace2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    if(lpFace1->wTextureIndex != lpFace2->wTextureIndex)
        {
        t = _T(" - wTextureIndex on a quad face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }

    for(BYTE bJ = 0; bJ < 4; bJ++)
        {
        if(lpFace1->awVertices[bJ] != lpFace2->awVertices[bJ])
            {
            t = _T(" - vertex index on a quad face !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        }

    return boOk;
}


bool    boCompareMesh(TR123_MESH* lpMesh1,TR123_MESH* lpMesh2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    if(lpMesh1->Center.sX != lpMesh2->Center.sX  ||  lpMesh1->Center.sY != lpMesh2->Center.sY  ||  lpMesh1->Center.sZ != lpMesh2->Center.sZ)
        {
        t = _T(" - collision sphere center !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,lCollisionSize,"collision sphere size")
    CMP_SIMPLE(lpMesh1,lpMesh2,Lighting.LightingType,"lighting type")
    CMP_SIMPLE(lpMesh1,lpMesh2,Lighting.sNumData,"lighting num data")
    else
        {
        long    lNumData = lpMesh1->Lighting.sNumData; if(lNumData < 0) lNumData = -lNumData;

        if(memcmp(lpMesh1->Lighting.lpLights,lpMesh2->Lighting.lpLights,sizeof(bit16) * lNumData) != 0)
            {
            t = _T(" - lighting values !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,sNumVertices,"num vertices")
    else
        {
        if(memcmp(lpMesh1->lpVertices,lpMesh2->lpVertices,lpMesh1->sNumVertices * sizeof(TR2_VERTEX)) != 0)
            {
            t = _T(" - vertices data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,sNumColoredRectangles,"num colored rectangles")
    else
        {
        for(DWORD dwFace = 0; dwFace < lpMesh1->sNumColoredRectangles; dwFace++)
            {
            TR2_FACE4* lpFace1 = &lpMesh1->lpColoredRectangles[dwFace];
            TR2_FACE4* lpFace2 = &lpMesh1->lpColoredRectangles[dwFace];

            boOk &= boCompareQuad(lpFace1,lpFace2,false,s,sh);
            }
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,sNumColoredTriangles,"num colored triangles")
    else
        {
        for(DWORD dwFace = 0; dwFace < lpMesh1->sNumColoredTriangles; dwFace++)
            {
            TR2_FACE3* lpFace1 = &lpMesh1->lpColoredTriangles[dwFace];
            TR2_FACE3* lpFace2 = &lpMesh1->lpColoredTriangles[dwFace];

            boOk &= boCompareTri(lpFace1,lpFace2,false,s,sh);
            }
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,sNumTexturedRectangles,"num textured rectangles")
    else
        {
        for(DWORD dwFace = 0; dwFace < lpMesh1->sNumTexturedRectangles; dwFace++)
            {
            TR2_FACE4* lpFace1 = &lpMesh1->lpTexturedRectangles[dwFace];
            TR2_FACE4* lpFace2 = &lpMesh1->lpTexturedRectangles[dwFace];

            boOk &= boCompareQuad(lpFace1,lpFace2,true,s,sh);
            }
        }

    CMP_SIMPLE(lpMesh1,lpMesh2,sNumTexturedTriangles,"num textured triangles")
    else
        {
        for(DWORD dwFace = 0; dwFace < lpMesh1->sNumTexturedTriangles; dwFace++)
            {
            TR2_FACE3* lpFace1 = &lpMesh1->lpTexturedTriangles[dwFace];
            TR2_FACE3* lpFace2 = &lpMesh1->lpTexturedTriangles[dwFace];

            boOk &= boCompareTri(lpFace1,lpFace2,true,s,sh);
            }
        }

    return boOk;
}


bool    boCompareStaticMesh(TR123_FILE_STRUCT& trFile1,TR2_STATICMESH* lpStatMesh1,TR123_FILE_STRUCT& trFile2,TR2_STATICMESH* lpStatMesh2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    CMP_SIMPLE(lpStatMesh1,lpStatMesh2,wFlags,"flags")

    if( lpStatMesh1->BoundingBox->sMinX != lpStatMesh2->BoundingBox->sMinX  ||
        lpStatMesh1->BoundingBox->sMinY != lpStatMesh2->BoundingBox->sMinY  ||
        lpStatMesh1->BoundingBox->sMinZ != lpStatMesh2->BoundingBox->sMinZ  ||
        lpStatMesh1->BoundingBox->sMaxX != lpStatMesh2->BoundingBox->sMaxX  ||
        lpStatMesh1->BoundingBox->sMaxY != lpStatMesh2->BoundingBox->sMaxY  ||
        lpStatMesh1->BoundingBox->sMaxZ != lpStatMesh2->BoundingBox->sMaxZ
      )
        {
        t = _T(" - bounding box !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF);
        s += sh + t;
        boOk = false;
        }

    boOk &= boCompareMesh(trFile1.lplpMeshes[trFile1.lpMeshPointers[lpStatMesh1->wMesh]],trFile2.lplpMeshes[trFile2.lpMeshPointers[lpStatMesh2->wMesh]],s,sh);

    return boOk;
}


bool    boCompareSpriteTexture(TR2_SPRITE_TEXTURE* lpSprText1,TR2_SPRITE_TEXTURE* lpSprText2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    if(memcmp(&lpSprText1->wWidth,&lpSprText2->wWidth,sizeof(TR2_SPRITE_TEXTURE) - sizeof(bitu16) - 2 * sizeof(bitu8)) != 0)
        {
        t = _T(" - sprite texture !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }

    return boOk;
}


bool    boCompareSprite(TR123_FILE_STRUCT& trFile1,TR2_SPRITE_SEQUENCE* lpSpr1,TR123_FILE_STRUCT& trFile2,TR2_SPRITE_SEQUENCE* lpSpr2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    CMP_SIMPLE(lpSpr1,lpSpr2,sNegativeLength,"num sprites")
    else
        {
        for(DWORD dwI = 0; dwI < -lpSpr1->sNegativeLength; dwI++)
            {
            t.Format(_T(" Sprite texture #%.4d"),dwI);
            boOk &= boCompareSpriteTexture(&trFile1.lpSpriteTextures[lpSpr1->sOffset + dwI],&trFile2.lpSpriteTextures[lpSpr2->sOffset + dwI],s,sh+t);
            }
        }

    CMP_SIMPLE(lpSpr1,lpSpr2,sOffset,"offset")

    return boOk;
}


DWORD dwGetRealFrameSize(enumTRType trType,bitu16* lpFrame,DWORD dwNumMeshes)
{
    bitu16* lpsFrame = lpFrame;

    while(dwNumMeshes--)
        {
        bitu16  wFrameData = *lpFrame; lpFrame++;

        if((wFrameData & 0xC000)  &&  (trType != TR1_FILE)) // single axis of rotation
            {
            }
        else    // 3 axis of rotation
            {
            WORD    wFrameData2 = *lpFrame; lpFrame++;
            }
        }

    return (lpFrame - lpsFrame) * 2;
}


bool    boCompareAnimation(DWORD dwNumMeshes,TR123_FILE_STRUCT& trFile1,TR2_ANIMATION* lpAnim1,TR123_FILE_STRUCT& trFile2,TR2_ANIMATION* lpAnim2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    CMP_SIMPLE(lpAnim1,lpAnim2,bFrameRate,"frame rate")
        CMP_SIMPLE(lpAnim1,lpAnim2,bFrameSize,"frame size")
        CMP_SIMPLE(lpAnim1,lpAnim2,dwFrameOffset,"frame offset")
    CMP_SIMPLE(lpAnim1,lpAnim2,wStateId,"state id")
    CMP_SIMPLE(lpAnim1,lpAnim2,Unknown,"Unknown")
    CMP_SIMPLE(lpAnim1,lpAnim2,sSpeed,"speed")
    CMP_SIMPLE(lpAnim1,lpAnim2,wAccelLo,"acceleration lo")
    CMP_SIMPLE(lpAnim1,lpAnim2,sAccelHi,"acceleration hi")
    CMP_SIMPLE(lpAnim1,lpAnim2,wFrameStart,"frame start")
    CMP_SIMPLE(lpAnim1,lpAnim2,wFrameEnd,"frame end")
    CMP_SIMPLE(lpAnim1,lpAnim2,wNextAnimation,"next animation")
    CMP_SIMPLE(lpAnim1,lpAnim2,wNextFrame,"next frame")
    CMP_SIMPLE(lpAnim1,lpAnim2,wNumStateChanges,"num state changes")
    CMP_SIMPLE(lpAnim1,lpAnim2,wStateChangeOffset,"state change offset")
    else
        {
        if(memcmp(&trFile1.lpStateChanges[lpAnim1->wStateChangeOffset],&trFile2.lpStateChanges[lpAnim2->wStateChangeOffset],sizeof(TR2_STATE_CHANGE) * lpAnim1->wNumStateChanges) != 0)
            {
            t = _T(" - state changes data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        else
            {
            for(DWORD dwStateChange1 = 0; dwStateChange1 < lpAnim1->wNumStateChanges; dwStateChange1++)
                {
                TR2_STATE_CHANGE* lpStateChange1 = &trFile1.lpStateChanges[lpAnim1->wStateChangeOffset + dwStateChange1];
                TR2_STATE_CHANGE* lpStateChange2 = &trFile2.lpStateChanges[lpAnim2->wStateChangeOffset + dwStateChange1];

                if(memcmp(&trFile1.lpAnimDispatches[lpStateChange1->wAnimDispatch],&trFile2.lpAnimDispatches[lpStateChange2->wAnimDispatch],lpStateChange1->wNumAnimDispatches * sizeof(TR2_ANIM_DISPATCH)) != 0)
                    {
                    t = _T(" - anim dispatch data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
                    }
                }
            }
        }

    CMP_SIMPLE(lpAnim1,lpAnim2,wNumAnimCommands,"num anim commands")
    else
        {
        TR2_ANIM_COMMAND*   lpAnimCmd1 = &trFile1.lpAnimCommands[lpAnim1->wAnimCommand];
        TR2_ANIM_COMMAND*   lpAnimCmd2 = &trFile2.lpAnimCommands[lpAnim2->wAnimCommand];

        for(DWORD dwAnimCmd1 = 0; dwAnimCmd1 < lpAnim1->wNumAnimCommands; dwAnimCmd1++)
            {
            CTRAnimCommand  AnimCommand1(lpAnimCmd1, lpAnim1->wFrameStart);
            WORD    wNumParams1 = AnimCommand1.wGetNumParams();

            CTRAnimCommand  AnimCommand2(lpAnimCmd2, lpAnim2->wFrameStart);
            WORD    wNumParams2 = AnimCommand2.wGetNumParams();

            if(wNumParams1 != -1  &&  wNumParams2 != -1)
                {
                if(wNumParams1 != wNumParams2)
                    {
                    t = _T(" - anim commands data (num params different) !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
                    }
                else if(memcmp(lpAnimCmd1,lpAnimCmd2,sizeof(TR2_ANIM_COMMAND) * (wNumParams1 + 1)) != 0)
                    {
                    t = _T(" - anim commands data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
                    }
                }
            else
                {
                t = _T(" - num param = -1 in a anim command data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
                }

            lpAnimCmd1 += wNumParams1 + 1;
            lpAnimCmd2 += wNumParams2 + 1;
            }

        }

    CMP_SIMPLE(lpAnim1,lpAnim2,wAnimCommand,"anim command offset")

    DWORD   dwNumFrames1    = lpAnim1->wFrameEnd - lpAnim1->wFrameStart + 1;
    DWORD   dwAnimNumKeys1  = (dwNumFrames1 - 1) / lpAnim1->bFrameRate + 1;
    
    DWORD   dwNumFrames2    = lpAnim2->wFrameEnd - lpAnim2->wFrameStart + 1;
    DWORD   dwAnimNumKeys2  = (dwNumFrames2 - 1) / lpAnim2->bFrameRate + 1;

    if(dwNumFrames1 != dwNumFrames2  ||  dwAnimNumKeys1 != dwAnimNumKeys2)
        {
        t = _T(" - anim num frames / keys !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
        }
    else
        {
        DWORD   dwFrameStep1    = lpAnim1->bFrameSize;
        DWORD   dwFrameStep2    = lpAnim2->bFrameSize;

        if(dwFrameStep1 != dwFrameStep2)
            {
            t.Format(_T("  - different framestep (%ld, %ld)" CRLF),dwFrameStep1,dwFrameStep2); s += sh + t;
            }

        for(DWORD dwKey1 = 0; dwKey1 < dwAnimNumKeys1; dwKey1++)
            {
            bitu16* lpFrame1        = &trFile1.lpFrames[lpAnim1->dwFrameOffset / 2];
            bitu16* lpFrame2        = &trFile2.lpFrames[lpAnim2->dwFrameOffset / 2];

            CString tt; tt.Format(_T("    Key #%.4d"),dwKey1);

            lpFrame1 += dwKey1 * dwFrameStep1;
            lpFrame2 += dwKey1 * dwFrameStep2;

            bit16   sBBLoX1  = (bit16) *lpFrame1; lpFrame1++;
            bit16   sBBHiX1  = (bit16) *lpFrame1; lpFrame1++;
            bit16   sBBLoY1  = (bit16) *lpFrame1; lpFrame1++;
            bit16   sBBHiY1  = (bit16) *lpFrame1; lpFrame1++;
            bit16   sBBLoZ1  = (bit16) *lpFrame1; lpFrame1++;
            bit16   sBBHiZ1  = (bit16) *lpFrame1; lpFrame1++;

            bit16   sTransX1 = (bit16) *lpFrame1; lpFrame1++;
            bit16   sTransY1 = (bit16) *lpFrame1; lpFrame1++;
            bit16   sTransZ1 = (bit16) *lpFrame1; lpFrame1++;

            bit16   sBBLoX2  = (bit16) *lpFrame2; lpFrame2++;
            bit16   sBBHiX2  = (bit16) *lpFrame2; lpFrame2++;
            bit16   sBBLoY2  = (bit16) *lpFrame2; lpFrame2++;
            bit16   sBBHiY2  = (bit16) *lpFrame2; lpFrame2++;
            bit16   sBBLoZ2  = (bit16) *lpFrame2; lpFrame2++;
            bit16   sBBHiZ2  = (bit16) *lpFrame2; lpFrame2++;

            bit16   sTransX2 = (bit16) *lpFrame2; lpFrame2++;
            bit16   sTransY2 = (bit16) *lpFrame2; lpFrame2++;
            bit16   sTransZ2 = (bit16) *lpFrame2; lpFrame2++;

            if( sBBLoX1 != sBBLoX2  ||  sBBLoY1 != sBBLoY2  ||  sBBLoZ1 != sBBLoZ2  ||
                sBBHiX1 != sBBHiX2  ||  sBBHiY1 != sBBHiY2  ||  sBBHiZ1 != sBBHiZ2  ||
                sTransX1 != sTransX2  ||  sTransY1 != sTransY2  ||  sTransZ1 != sTransZ2)
                {
                t = _T(" - bouding box / root mesh translation !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + tt + t; boOk = false;
                }

            if(dwFrameStep1 == dwFrameStep2)
                {
                DWORD dwSize = dwGetRealFrameSize(TR2_FILE,lpFrame1,dwNumMeshes);
                if(memcmp(lpFrame1, lpFrame2, dwSize) != 0)
                    {
                    t = _T(" - key data !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + tt + t; boOk = false;
                    }
                }
            }
        }

    return boOk;
}


bool    boCompareMoveable(CTRFile* lpTRFile1,TR123_FILE_STRUCT& trFile1,TR2_MOVEABLE* lpMvb1,CTRFile* lpTRFile2,TR123_FILE_STRUCT& trFile2,TR2_MOVEABLE* lpMvb2,CString& s,CString& sh)
{
    bool boOk = true;
    CString t;

    CMP_SIMPLE(lpMvb1,lpMvb2,dwFrameOffset,"Frame offset")
    CMP_SIMPLE(lpMvb1,lpMvb2,dwMeshTree,"MeshTree offset")
    CMP_SIMPLE(lpMvb1,lpMvb2,wAnimation,"Animation offset")
    CMP_SIMPLE(lpMvb1,lpMvb2,wNumMeshes,"num meshes")
    else
        {
        for(DWORD dwI = 0; dwI < lpMvb1->wNumMeshes; dwI++)
            {
            TR123_MESH* lpMesh1 = trFile1.lplpMeshes[trFile1.lpMeshPointers[lpMvb1->wStartingMesh + dwI]];
            TR123_MESH* lpMesh2 = trFile2.lplpMeshes[trFile2.lpMeshPointers[lpMvb2->wStartingMesh + dwI]];

            t.Format(_T("  Mesh #%.4d" CRLF),dwI);

            boOk &= boCompareMesh(lpMesh1,lpMesh2,s,sh+t);
            }

        if(memcmp(&trFile1.lpMeshTrees[lpMvb1->dwMeshTree],&trFile2.lpMeshTrees[lpMvb2->dwMeshTree],(lpMvb1->wNumMeshes - 1) * sizeof(TR2_MESHTREE) * 4) != 0)
            {
            t = _T(" - mesh tree !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }

        DWORD dwNumAnim1 = lpTRFile1->dwGetNumAnimsForMoveable(lpMvb1);
        DWORD dwNumAnim2 = lpTRFile2->dwGetNumAnimsForMoveable(lpMvb2);

        if(dwNumAnim1 != dwNumAnim2)
            {
            t = _T(" - num animations !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += sh + t; boOk = false;
            }
        else
            {
            for(DWORD dwAnim1 = 0; dwAnim1 < dwNumAnim1; dwAnim1++)
                {
                TR2_ANIMATION* lpAnim1 = &trFile1.lpAnimations[lpMvb1->wAnimation + dwAnim1];
                TR2_ANIMATION* lpAnim2 = &trFile2.lpAnimations[lpMvb2->wAnimation + dwAnim1];

                t.Format(_T("  Animation #%.4d"),dwAnim1);

                boOk &= boCompareAnimation(lpMvb1->wNumMeshes,trFile1,lpAnim1,trFile2,lpAnim2,s,sh+t);
                }
            }
        }

    CMP_SIMPLE(lpMvb1,lpMvb2,wStartingMesh,"starting mesh offset")

    return boOk;
}


void    vCompareLevels(TCHAR* lpszLevel1, TCHAR* lpszLevel2)
{
    static CString s;
    CString t;

    CTRFile File1;
    CTRFile File2;

    if(!File1.boLoad(lpszLevel1))
        {
        s.Format(_T("Can't load level %s"),lpszLevel1);
        AfxMessageBox(s);
        return;
        }

    TR123_FILE_STRUCT& trFile1 = File1.GetTRFile().tr123File;

    if(!File2.boLoad(lpszLevel2))
        {
        s.Format(_T("Can't load level %s"),lpszLevel2);
        AfxMessageBox(s);
        return;
        }

    TR123_FILE_STRUCT& trFile2 = File2.GetTRFile().tr123File;

    bool boOk = true;

    AfxMessageBox("Starting comparison...");

    // Compare number of objects
    DWORD   dwNumMeshes1 = trFile1.dwNumMeshes, dwNumMeshes2 = trFile2.dwNumMeshes;
    DWORD   dwNumStatMesh1 = trFile1.dwNumStaticMeshes, dwNumStatMesh2 = trFile2.dwNumStaticMeshes;
    DWORD   dwNumSprite1 = trFile1.dwNumSpriteSequences, dwNumSprite2 = trFile2.dwNumSpriteSequences;
    DWORD   dwNumMvb1 = trFile1.dwNumMoveables, dwNumMvb2 = trFile2.dwNumMoveables;

    if(dwNumMeshes1 != dwNumMeshes2)
        {
        t.Format(_T("Num meshes (%ld in first, %ld in second)" CRLF),dwNumMeshes1,dwNumMeshes2);
        s += t;
        boOk = false;
        }

    if(dwNumStatMesh1 != dwNumStatMesh2)
        {
        t.Format(_T("Num static meshes (%ld in first, %ld in second)" CRLF),dwNumStatMesh1,dwNumStatMesh2);
        s += t;
        boOk = false;
        }

    if(dwNumSprite1 != dwNumSprite2)
        {
        t.Format(_T("Num sprite sequences (%ld in first, %ld in second)" CRLF),dwNumSprite1,dwNumSprite2);
        s += t;
        boOk = false;
        }

    if(dwNumMvb1 != dwNumMvb2)
        {
        t.Format(_T("Num moveables (%ld in first, %ld in second)" CRLF),dwNumMvb1,dwNumMvb2);
        s += t;
        boOk = false;
        }

    // Compare static meshes
    if(dwNumStatMesh1 == dwNumStatMesh2)
        {
        for(DWORD dwStatMesh1 = 0; dwStatMesh1 < dwNumStatMesh1; dwStatMesh1++)
            {
            TR2_STATICMESH* lpStatMesh1 = &trFile1.lpStaticMeshes[dwStatMesh1];
            TR2_STATICMESH* lpStatMesh2 = NULL;
            DWORD           dwObjId = lpStatMesh1->dwObjectId;

            t.Format(_T("Static mesh id #%.4d"),dwObjId);

            for(DWORD dwStatMesh2 = 0; dwStatMesh2 < dwNumStatMesh2; dwStatMesh2++)
                {
                TR2_STATICMESH* lp = &trFile2.lpStaticMeshes[dwStatMesh2];
                if(lp->dwObjectId == dwObjId)
                    {
                    lpStatMesh2 = lp;
                    break;
                    }
                }

            if(!lpStatMesh2)
                {
                t.Format(_T("Can't find static mesh id #%ld in second level !" CRLF),dwObjId); s+= t;
                boOk = false;
                }
            else
                {
                boOk &= boCompareStaticMesh(trFile1,lpStatMesh1,trFile2,lpStatMesh2,s,t);
                }
            }
        }

    // Compare sprites
    if(dwNumSprite1 == dwNumSprite2)
        {
        for(DWORD dwSprite1 = 0; dwSprite1 < dwNumSprite1; dwSprite1++)
            {
            TR2_SPRITE_SEQUENCE*    lpSpr1 = &trFile1.lpSpriteSequences[dwSprite1];
            TR2_SPRITE_SEQUENCE*    lpSpr2 = NULL;
            DWORD                   dwObjId = lpSpr1->lObjectId;

            t.Format(_T("Sprite sequence id #%.4d"),dwObjId);

            for(DWORD dwSprite2 = 0; dwSprite2 < dwNumSprite2; dwSprite2++)
                {
                TR2_SPRITE_SEQUENCE* lp = &trFile2.lpSpriteSequences[dwSprite2];
                if(lp->lObjectId == dwObjId)
                    {
                    lpSpr2 = lp;
                    break;
                    }
                }

            if(!lpSpr2)
                {
                t.Format(_T("Can't find sprite sequence id #%ld in second level !" CRLF),dwObjId); s+= t;
                boOk = false;
                }
            else
                {
                boOk &= boCompareSprite(trFile1,lpSpr1,trFile2,lpSpr2,s,t);
                }
            }
        }

    // Compare moveables
    if(dwNumMvb1 == dwNumMvb2)
        {
        for(DWORD dwMvb1 = 0; dwMvb1 < dwNumMvb1; dwMvb1++)
            {
            TR2_MOVEABLE*    lpMvb1 = &trFile1.lpMoveables[dwMvb1];
            TR2_MOVEABLE*    lpMvb2 = NULL;
            DWORD            dwObjId = lpMvb1->dwObjectId;

            t.Format(_T("Moveable id #%.4d"),dwObjId);

            for(DWORD dwMvb2 = 0; dwMvb2 < dwNumMvb2; dwMvb2++)
                {
                TR2_MOVEABLE* lp = &trFile2.lpMoveables[dwMvb2];
                if(lp->dwObjectId == dwObjId)
                    {
                    lpMvb2 = lp;
                    break;
                    }
                }

            if(!lpMvb2)
                {
                t.Format(_T("Can't find moveable id #%ld in second level !" CRLF),dwObjId); s+= t;
                boOk = false;
                }
            else
                {
                boOk &= boCompareMoveable(&File1,trFile1,lpMvb1,&File2,trFile2,lpMvb2,s,t);
                }
            }
        }

    // Exact comparisons
    CString sh;

    if(trFile1.dwNumFrames != trFile2.dwNumFrames)
        {
        t.Format(_T("Num frames (%ld in first, %ld in second)" CRLF),trFile1.dwNumFrames,trFile2.dwNumFrames);
        s += t;
        boOk = false;
        }

    if(memcmp(trFile1.lpAnimations,trFile2.lpAnimations,trFile1.dwNumAnimations*sizeof(TR2_ANIMATION)) != 0)
        {
        t = _T("  - ANIMATIONS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpStateChanges,trFile2.lpStateChanges,trFile1.dwNumStateChanges*sizeof(TR2_STATE_CHANGE)) != 0)
        {
        t = _T("  - STATE CHANGE !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpAnimDispatches,trFile2.lpAnimDispatches,trFile1.dwNumAnimDispatches*sizeof(TR2_ANIM_DISPATCH)) != 0)
        {
        t = _T("  - ANIM DISPATCHES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpAnimCommands,trFile2.lpAnimCommands,trFile1.dwNumAnimCommands*sizeof(TR2_ANIM_COMMAND)) != 0)
        {
        t = _T("  - ANIM COMMANDS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpMeshTrees,trFile2.lpMeshTrees,trFile1.dwNumMeshTrees*sizeof(TR2_MESHTREE)) != 0)
        {
        t = _T("  - MESH TREES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpMoveables,trFile2.lpMoveables,trFile1.dwNumMoveables*sizeof(TR2_MOVEABLE)) != 0)
        {
        t = _T("  - MOVEABLES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpStaticMeshes,trFile2.lpStaticMeshes,trFile1.dwNumStaticMeshes*sizeof(TR2_STATICMESH)) != 0)
        {
        t = _T("  - STATIC MESHES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpSpriteSequences,trFile2.lpSpriteSequences,trFile1.dwNumSpriteSequences*sizeof(TR2_SPRITE_SEQUENCE)) != 0)
        {
        t = _T("  - SPRITE SEQUENCES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.lpItems,trFile2.lpItems,trFile1.dwNumItems*sizeof(TR123_ITEM)) != 0)
        {
        t = _T("  - ITEMS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    if(memcmp(trFile1.SoundMap.tr2.aData,trFile2.SoundMap.tr2.aData,sizeof(TR2_SOUNDMAP)) != 0)
        {
        t = _T("  - SOUND MAP !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    CMP_SIMPLE((&trFile1),(&trFile2),dwNumSoundDetails,"num sound details")
    else if(memcmp(trFile1.lpSoundDetails,trFile2.lpSoundDetails,trFile1.dwNumSoundDetails * sizeof(TR2_SOUND_DETAILS)) != 0)
        {
        t = _T("  - SOUND DETAILS !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    CMP_SIMPLE((&trFile1),(&trFile2),dwNumSampleIndices,"num sample indices")
    else if(memcmp(trFile1.lpSampleIndices,trFile2.lpSampleIndices,trFile1.dwNumSampleIndices * sizeof(bitu32)) != 0)
        {
        t = _T("  - SAMPLE INDICES !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!" CRLF); s += t; boOk = false;
        }

    FILE* f = fopen("c:\\trcmp.txt","wb");
    
    fwrite((LPCTSTR)s,1,s.GetLength(),f);
    fclose(f);

    if(boOk) AfxMessageBox("No difference found"); else AfxMessageBox("Differences found !");
}


void vCheckSounds(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr2");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

                    long   dwExpectedSampleIndiceIndex = 0;

                    for(DWORD dwSoundDetail = 0; dwSoundDetail < m_trFile.tr123File.dwNumSoundDetails; dwSoundDetail++)
                        {
                        TR2_SOUND_DETAILS* lpSoundDetail = &m_trFile.tr123File.lpSoundDetails[dwSoundDetail];

                        if(lpSoundDetail->sSample != dwExpectedSampleIndiceIndex)
                            {
                            CString s;
                            s.Format("Sample indice offset for sound detail #%ld in level %s is %ld but should be %ld",dwSoundDetail,fd.cFileName,lpSoundDetail->sSample,dwExpectedSampleIndiceIndex);
                            AfxMessageBox(s,MB_ICONEXCLAMATION);
                            }

                        DWORD dwNumSamples = (lpSoundDetail->sUnknown2 & 0x00FC) >> 2;

                        dwExpectedSampleIndiceIndex = lpSoundDetail->sSample + dwNumSamples;

                        if(trFile.GetTRType() != TR1_FILE)
                            for(DWORD dwSample = 0; dwSample < dwNumSamples - 1; dwSample++)
                                {
                                if( m_trFile.tr123File.lpSampleIndices[lpSoundDetail->sSample + dwSample] !=
                                    m_trFile.tr123File.lpSampleIndices[lpSoundDetail->sSample + dwSample + 1] - 1)
                                    {
                                    CString s;
                                    s.Format("Sample indice sequence is not consecutive. Sound detail #%ld in level %s", dwSoundDetail,fd.cFileName);
                                    AfxMessageBox(s,MB_ICONEXCLAMATION);
                                    }
                                }
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vGetNumSamplesPerSound(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.phd");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    std::vector<DWORD> ListSnd;

    ListSnd.resize(1000);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

					for( DWORD dwSoundMap = 0; dwSoundMap < 256; dwSoundMap++ )
						{
						DWORD	dwSoundDetail = m_trFile.tr123File.SoundMap.tr2.aData[ dwSoundMap ];

						if( dwSoundDetail != 0xFFFFFFFF )
							{
							TR2_SOUND_DETAILS* lpSoundDetail = &m_trFile.tr123File.lpSoundDetails[dwSoundDetail];

							DWORD dwNumSamples = (lpSoundDetail->sUnknown2 & 0x00FC) >> 2;

							DWORD   dwPrevNum = ListSnd[dwSoundMap];

							if(dwPrevNum != 0  &&  dwPrevNum != dwNumSamples)
								{
								//CString s;
								//s.Format("Same sounds with different num samples ! Sample indice #%.3d, num %ld & %ld",lpSoundDetail->sSample,dwPrevNum,dwNumSamples);
								//AfxMessageBox(s,MB_ICONEXCLAMATION);
								}

							ListSnd[dwSoundMap] = dwNumSamples > dwPrevNum ? dwNumSamples : dwPrevNum;
							}
						}
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    CString s;

    DWORD dw = 0;
	DWORD dwTot = 0;
	DWORD dwSnd = 0;
    CString t;

    for(DWORD dwI = 0; dwI < 370; dwI++)
        {
        DWORD dwNum = ListSnd[dwI];

		if( dwNum != 0 )
			{
			for( DWORD dwJ = 0; dwJ < dwNum; ++dwJ )
				{
				t.Format("%.4d -- Sound #%.3d" CRLF,dwI, dwSnd++);
				s += t;
				}
			dwTot += dwNum;
			}
        }

	t.Format("Num snd=%ld" CRLF,dwSnd);

    s += t;

	t.Format("Num tot=%ld" CRLF,dwTot);

    s += t;

/*    for(DWORD dwI = 0; dwI < 1000; dwI++)
        {
        DWORD dwNum = ListSnd[dwI];

        CString t;

        if(dwI != dw)
            t.Format("%.3d : %.3d !!!!!!!!!!!!!!!" CRLF,dwI, dwNum);
        else
            t.Format("%.3d : %.3d" CRLF,dwI, dwNum);

        s += t;
        if(dwI + dwNum > dw) dw = dwI + dwNum;
        }
*/
    FILE* f = fopen("c:\\trcmp.txt","wb");
    
    fwrite((LPCTSTR)s,1,s.GetLength(),f);
    fclose(f);

    AfxMessageBox("The end");
}


void vCompareNumTiles(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.phd");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

	CString s;

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

					DWORD dwNumTiles = m_trFile.tr123File.dwNumTextiles - 1;

					trFile.boSave("c:\\temp\\a.tr2");

					DWORD	dwNewNumTiles = m_trFile.tr123File.dwNumTextiles;

					CString t;

					t.Format("%s: %ld %ld" CRLF,fd.cFileName,dwNumTiles,dwNewNumTiles);

					s += t;

					FILE* f = fopen("c:\\trcmp.txt","wb");
    
					fwrite((LPCTSTR)s,1,s.GetLength(),f);
					fclose(f);
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void    vQuat2Euler_( CQuaternion& Quat, FLOAT* lpfX, FLOAT* lpfY, FLOAT* lpfZ )
{
    CVect3D&    vec3 = Quat.v3GetVector();
    FLOAT       fX = vec3[0], fY = vec3[1], fZ = vec3[2];
    FLOAT       fW = Quat.fGetScalar();

    FLOAT       fR00 = 1.0 - 2.0 * fY * fY - 2.0 * fZ * fZ;
    FLOAT       fR01 = 2.0 * fX * fY + 2.0 * fW * fZ;
    FLOAT       fR02 = 2.0 * fX * fZ - 2.0 * fW * fY;
    FLOAT       fR10 = 2.0 * fX * fY - 2.0 * fW * fZ;
    FLOAT       fR11 = 1.0 - 2.0 * fX * fX - 2.0 * fZ * fZ;
    FLOAT       fR12 = 2.0 * fY * fZ + 2.0 * fW * fX;
    FLOAT       fR20 = 2.0 * fX * fZ + 2.0 * fW * fY;
    FLOAT       fR21 = 2.0 * fY * fZ - 2.0 * fW * fX;
    FLOAT       fR22 = 1.0 - 2.0 * fX * fX - 2.0 * fY * fY;

#define ZXY

#ifdef YXZ
    *lpfX = asin( -fR12 );

    if( *lpfX < _PI_ / 2.0 )
        {
        if( *lpfX > -_PI_ / 2.0)
            {
            *lpfY = atan2( fR02, fR22 ) + _PI_;
            *lpfZ = atan2( fR10, fR11 ) + _PI_;
            }
        else
            {
            *lpfY = -atan2( -fR01, fR00 ) + _PI_;
            *lpfZ = 0.0;
            }
        }
    else
        {
        *lpfY = atan2( -fR01, fR00 );
        *lpfZ = 0.0;
        }

    *lpfX = (*lpfX + _PI_ / 2.0) * 2.0;
#endif
#ifdef ZXY
    *lpfX = -asin( fR21 );

    if( *lpfX < _PI_ / 2.0 )
        {
        if( *lpfX > -_PI_ / 2.0)
            {
            *lpfZ = atan2( fR01, fR11 );// + _PI_;
            *lpfY = atan2( fR20, fR22 );// + _PI_;
            }
        else
            {
            *lpfZ = -atan2( -fR02, fR00 );// + _PI_;
            *lpfY = 0.0;
            }
        }
    else
        {
        *lpfZ = atan2( fR02, fR00 );
        *lpfY = 0.0;
        }

    //*lpfX = (*lpfX + _PI_ / 2.0) * 2.0;
#endif
}


void    vTestEuler2Quat()
{
    FLOAT   fRapp = (FLOAT) (_PI_ / 180.0);
    FLOAT   fX = 130.0 * fRapp;
    FLOAT   fY = 30.0 * fRapp;
    FLOAT   fZ = 60.0 * fRapp;
    FLOAT   fX2 = 1000000.0, fY2 = 1000000.0, fZ2 = 1000000.0;

    CAnimData   Data( fX, fY, fZ, CVect3D(0,0,0), 0, 0 );

    CQuaternion& quat = const_cast<CQuaternion&>( Data.quatGetQuaternion() );

    vQuat2Euler_( quat, &fX2, &fY2, &fZ2 );

    CAnimData   Data2( fX2, fY2, fZ2, CVect3D(0,0,0), 0, 0 );

    CQuaternion& quat2 = const_cast<CQuaternion&>( Data2.quatGetQuaternion() );

    CVect3D&    vec  = quat.v3GetVector();
    CVect3D&    vec2 = quat2.v3GetVector();

    CString s;

    s.Format( "length=%f\n%f:%f\n%f:%f\n%f:%f\n\n%f,%f,%f,%f\n%f,%f,%f,%f", 
        quat.fLength2(), fX, fX2, fY, fY2, fZ, fZ2,
        vec[0],vec[1],vec[2],quat.fGetScalar(),
        vec2[0],vec2[1],vec2[2],quat2.fGetScalar()
         );

    AfxMessageBox( s );
}


void vCheckObjAttrFlag(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

	DWORD*	lp1 = new DWORD [ 65536 ];
	DWORD*	lp2 = new DWORD [ 65536 ];
	DWORD*	lp3 = new DWORD [ 65536 ];
	DWORD*	lp4 = new DWORD [ 65536 ];
	DWORD*	lp5 = new DWORD [ 65536 ];
	DWORD*	lp6 = new DWORD [ 65536 ];

	ZeroMemory( lp1, 65536 * 4 );
	ZeroMemory( lp2, 65536 * 4 );
	ZeroMemory( lp3, 65536 * 4 );
	ZeroMemory( lp4, 65536 * 4 );
	ZeroMemory( lp5, 65536 * 4 );
	ZeroMemory( lp6, 65536 * 4 );

	std::string strDump;

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					CString t;

					t.Format( "#######################################################" CRLF "Level %s" CRLF, trFile.lpszGetLevelName() );

					strDump = t;

					strDump.reserve( 5000000 );

					strDump += "Object textures:" CRLF CRLF;

					for(DWORD dwObjText = 0; dwObjText < m_trFile.tr123File.dwNumObjectTextures; dwObjText++)
						{
						TR2_OBJECT_TEXTURE*	lpText = &m_trFile.tr123File.lpObjectTextures[ dwObjText ];

						lp1[ lpText->wTile ]++;
						lp2[ lpText->Unknown ]++;
						lp4[ lpText->wAttribute ]++;

						t.Format( "Obj text #%.4d: ", dwObjText );

						strDump += t;

						vDumpObjText( lpText, strDump, true );
						}

 					for(DWORD dwStatMesh = 0; dwStatMesh < m_trFile.tr123File.dwNumStaticMeshes; dwStatMesh++)
						{
						TR2_STATICMESH*	lpStatMsh = &m_trFile.tr123File.lpStaticMeshes[ dwStatMesh ];

						lp6[ lpStatMsh->wFlags ]++;

						t.Format( CRLF "Static Mesh id #%.4d = mesh #%.4d", lpStatMsh->dwObjectId, lpStatMsh->wMesh );

						strDump += t;
						}

					for(DWORD dwMesh = 0; dwMesh < lp->dwNumMeshPointers; dwMesh++)
						{
						TR123_MESH* lpMesh = lp->lplpMeshes[ lp->lpMeshPointers[ dwMesh ] ];

						for(DWORD dwPoly = 0; dwPoly < lpMesh->sNumTexturedRectangles; dwPoly++)
							{
							lp3[ lpMesh->lpTexturedRectangles[ dwPoly ].wTextureIndex ]++;
							lp5[ lpMesh->lpTexturedRectangles[ dwPoly ].wExtraTextureIndex ]++;
							}

						for(dwPoly = 0; dwPoly < lpMesh->sNumTexturedTriangles; dwPoly++)
							{
							lp3[ lpMesh->lpTexturedTriangles[ dwPoly ].wTextureIndex ]++;
							lp5[ lpMesh->lpTexturedTriangles[ dwPoly ].wExtraTextureIndex ]++;
							}

						t.Format( CRLF "Mesh #%.4d" CRLF CRLF, dwMesh );

						strDump += t;

						vDumpMesh( lpMesh, strDump );
						}

					for(DWORD dwNumRoom = 0; dwNumRoom < lp->wNumRooms; dwNumRoom++)
						{
						TR123_ROOM*     lpRoom = &lp->lpRooms[dwNumRoom];
						TR2_ROOM_DATA*  lpRoomData = &lpRoom->RoomData;

						for(DWORD dwPoly = 0; dwPoly < lpRoomData->wNumRectangles; dwPoly++)
							{
							lp3[ lpRoomData->lpRectangles[ dwPoly ].wTextureIndex ]++;
							}

						for(dwPoly = 0; dwPoly < lpRoomData->wNumTriangles; dwPoly++)
							{
							lp3[ lpRoomData->lpTriangles[ dwPoly ].wTextureIndex ]++;
							}
						}

					bitu16* lpsAnimatedTextures     = lp->lpAnimatedTextures;
					WORD    wNumAnimatedTextures    = *lpsAnimatedTextures++;

					strDump += CRLF "Animated textures:" CRLF CRLF;

					while(wNumAnimatedTextures--)
						{
						bitu16   wNum = *lpsAnimatedTextures++;  wNum++;
						while(wNum--)
							{
							bitu16  wNumObjText = *lpsAnimatedTextures;

							lp3[ wNumObjText ]++;

							t.Format( "%.4d ", (long) wNumObjText );

							strDump += t;

							lpsAnimatedTextures++;
							}
						strDump += CRLF;
						}

					char szz[123];
					sprintf(szz,"c:\\tr%s.txt",trFile.lpszGetShortLevelName());

					FILE* f6 = fopen(szz,"wb");
    				fwrite((LPCTSTR)strDump.c_str(),1,strDump.length(),f6);
					fclose(f6);
                   }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    CString s1;
    CString s2;
    CString s3;
    CString s4;
    CString s5;

    DWORD dw = 0;

    for(DWORD dwI = 0; dwI < 65536; dwI++)
        {
        CString t1;
        CString t2;
        CString t3;
        CString t4;
        CString t5;

        if(lp1[dwI])
            t1.Format("%.4X : %ld" CRLF,dwI,lp1[dwI]);

        if(lp2[dwI])
            t2.Format("%.4X : %ld" CRLF,dwI,lp2[dwI]);

        if(lp3[dwI])
            t3.Format("%.4X : %ld" CRLF,dwI,lp3[dwI]);

        if(lp4[dwI])
            t4.Format("%.4X : %ld" CRLF,dwI,lp4[dwI]);

        if(lp5[dwI])
            t5.Format("%.4X : %ld" CRLF,dwI,lp5[dwI]);

        s1 += t1;
        s2 += t2;
        s3 += t3;
        s4 += t4;
        s5 += t5;
        }

    FILE* f1 = fopen("c:\\tr1.txt","wb");
    
    fwrite((LPCTSTR)s1,1,s1.GetLength(),f1);
    fclose(f1);

    FILE* f2 = fopen("c:\\tr2.txt","wb");
    
    fwrite((LPCTSTR)s2,1,s2.GetLength(),f2);
    fclose(f2);

    FILE* f3 = fopen("c:\\tr3.txt","wb");
    
    fwrite((LPCTSTR)s3,1,s3.GetLength(),f3);
    fclose(f3);

    FILE* f4 = fopen("c:\\tr4.txt","wb");
    
    fwrite((LPCTSTR)s4,1,s4.GetLength(),f4);
    fclose(f4);

    FILE* f5 = fopen("c:\\tr5.txt","wb");
    
    fwrite((LPCTSTR)s5,1,s5.GetLength(),f5);
    fclose(f5);

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vCheckAnimUnknown(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					BYTE	zeros[8] = {0,0,0,0,0,0,0,0};

					for( DWORD dwI = 0; dwI < lp->dwNumAnimations; ++dwI )
						{
						TR2_ANIMATION*	lpAnim = &lp->lpAnimations[ dwI ];

						if( memcmp( lpAnim->Unknown2, zeros, 8 ) != 0  &&  memcmp(&lpAnim->Unknown, zeros, 8) != 0 )
							{
							dwI=dwI;
							}
						}
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vGetStateId(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.phd");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    std::vector<DWORD> ListStateId;

    ListStateId.resize(1000);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					TR2_MOVEABLE*	lpMvb = &lp->lpMoveables[ 0 ];
					DWORD			dwNumAnim = trFile.dwGetNumAnimsForMoveable( lpMvb );

					for( DWORD dwI = 0; dwI < dwNumAnim; ++dwI )
						{
						TR2_ANIMATION*	lpAnim = &lp->lpAnimations[ lpMvb->wAnimation + dwI ];

						ListStateId[lpAnim->wStateId]++;
						}

					/*for( DWORD dwI = 0; dwI < lp->dwNumAnimations; ++dwI )
						{
						TR2_ANIMATION*	lpAnim = &lp->lpAnimations[ dwI ];

						ListStateId[lpAnim->wStateId]++;
						}*/
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    CString s;
    CString t;

    for(int i=0; i<1000; i++)
        {
		if( ListStateId[i] != 0 )
			{
			t.Format("<state id=\"%.3d\" name=\"\"/>" CRLF,i);
			s += t;
			}
        }

    FILE* f = fopen("c:\\trstate.txt","wb");
    
    fwrite((LPCTSTR)s,1,s.GetLength(),f);
    fclose(f);

    AfxMessageBox("The end");
}


void vGenerateXML(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

	std::string strLevels;

	strLevels.reserve( 100000 );

	strLevels += "\t<levels>" CRLF;

	std::string strMvb;

	strMvb.reserve( 100000 );

	strMvb += "\t<moveables name=\"Moveables\" icon=\"Moveables\">"  CRLF;

	std::string strSpr;

	strSpr.reserve( 100000 );

	strSpr += "\t<sprites name=\"Sprites\" icon=\"Sprites\">"  CRLF;

	std::string strSnd;

	strSnd.reserve( 100000 );

	strSnd += "\t<sounds name=\"Sounds\" icon=\"Sounds\">"  CRLF;

	CString t;

	DWORD*	lpMvbId = new DWORD [ 65536 ];

	ZeroMemory( lpMvbId, 65536 * 4 );

	DWORD*	lpSprId = new DWORD [ 65536 ];

	ZeroMemory( lpSprId, 65536 * 4 );

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					strLevels += "\t\t<level id=\"";
					strLevels += trFile.lpszGetShortLevelName();
					strLevels += "\" name=\"\" type=\"normal\" icon=\"Level\">" CRLF;

					strLevels += "\t\t\t<static_meshes name=\"Static meshes\" icon=\"StaticMeshes\">" CRLF;

 					for(DWORD dwStatMesh = 0; dwStatMesh < m_trFile.tr123File.dwNumStaticMeshes; dwStatMesh++)
						{
						TR2_STATICMESH*	lpStatMsh = &m_trFile.tr123File.lpStaticMeshes[ dwStatMesh ];

						t.Format( "%ld", lpStatMsh->dwObjectId );

						strLevels += "\t\t\t\t<static_mesh id=\"";
						strLevels += t;
						strLevels += "\" name=\"\" icon=\"Shape\"/>" CRLF;
						}

					for(DWORD dwMvb = 0; dwMvb < m_trFile.tr123File.dwNumMoveables; ++dwMvb)
						{
						TR2_MOVEABLE* lpMvb = &m_trFile.tr123File.lpMoveables[dwMvb];

						lpMvbId[lpMvb->dwObjectId]++;
						}

					for(DWORD dwSpr = 0; dwSpr < m_trFile.tr123File.dwNumSpriteSequences; ++dwSpr)
						{
						TR2_SPRITE_SEQUENCE* lpSpr = &m_trFile.tr123File.lpSpriteSequences[dwSpr];

						lpSprId[lpSpr->lObjectId]++;
						}

					strLevels += "\t\t\t</static_meshes>" CRLF;
					strLevels += "\t\t</level>" CRLF CRLF;
                   }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

	for( DWORD dwMvb = 0; dwMvb < 65536; ++dwMvb )
		{
		if( lpMvbId[dwMvb] != 0 )
			{
			t.Format( "%ld", dwMvb );

			strMvb += "\t\t<moveable id=\"";
			strMvb += t;
			strMvb += "\" name=\"\" icon=\"Tool\"/>" CRLF;
			}
		}

	for( DWORD dwSpr = 0; dwSpr < 65536; ++dwSpr )
		{
		if( lpSprId[dwSpr] != 0 )
			{
			t.Format( "%ld", dwSpr );

			strSpr += "\t\t<sprite id=\"";
			strSpr += t;
			strSpr += "\" name=\"\" icon=\"Tool\" type=\"list\"/>" CRLF;
			}
		}

	strLevels += "\t</levels>" CRLF CRLF;
	strMvb += "\t</moveables>" CRLF CRLF;
	strSpr += "\t</sprites>" CRLF CRLF;
	strSnd += "\t</sounds>" CRLF;

    FILE* f1 = fopen("c:\\trlevels.xml","wb");
    
    fwrite((LPCTSTR)strLevels.c_str(),1,strLevels.length(),f1);
    fwrite((LPCTSTR)strMvb.c_str(),1,strMvb.length(),f1);
    fwrite((LPCTSTR)strSpr.c_str(),1,strSpr.length(),f1);
    fwrite((LPCTSTR)strSnd.c_str(),1,strSnd.length(),f1);

    fclose(f1);

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vReadFromTextFile( const char* lpszTextFile, char* lpszTRLevel, char* lpszTRSaveLevel )
{
	char sz[512];
	long l;

    CTRFile trFile;

    if( !trFile.boLoad( lpszTRLevel ) )
		{
		AfxMessageBox( "Error reading the level !" );

		return;
		}

    TR_FILE& m_trFile = trFile.GetTRFile();
	TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

	FILE* ff = fopen( lpszTextFile, "rt" );

	fscanf( ff, "#######################################################" CRLF "Level %s" CRLF, sz );
	fscanf( ff, "Object textures:" CRLF CRLF );

 	for(DWORD dwObjText = 0; dwObjText < m_trFile.tr123File.dwNumObjectTextures; dwObjText++)
		{
		TR2_OBJECT_TEXTURE*	lpText = &m_trFile.tr123File.lpObjectTextures[ dwObjText ];

		fscanf( ff, "Obj text #%ld: ", &l );

		vReadObjText( lpText, ff, true );
		}

 	for(DWORD dwStatMesh = 0; dwStatMesh < m_trFile.tr123File.dwNumStaticMeshes; dwStatMesh++)
		{
		TR2_STATICMESH*	lpStatMsh = &m_trFile.tr123File.lpStaticMeshes[ dwStatMesh ];

		fscanf( ff, CRLF "Static Mesh id #%ld = mesh #%ld", &l, &l );
		}

	for(DWORD dwMesh = 0; dwMesh < lp->dwNumMeshPointers; dwMesh++)
		{
		TR123_MESH* lpMesh = lp->lplpMeshes[ lp->lpMeshPointers[ dwMesh ] ];

		fscanf( ff, CRLF "Mesh #%ld" CRLF CRLF, &l );

		vReadMesh( lpMesh, ff );
		}

    fclose( ff );

	trFile.boSave( lpszTRSaveLevel );
}


void vCheckAnimText(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

	std::string strDump;

	CString t;

	bool boOk = false;

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					DWORD	dwNumNoBump = m_trFile.tr123File.wNumTextNoBump;
					DWORD	dwNumInhab = m_trFile.tr123File.wNumTextInhab;
					DWORD	dwNumBump = m_trFile.tr123File.wNumTextBump;

					bitu16* lpsAnimatedTextures     = lp->lpAnimatedTextures;
					WORD    wNumAnimatedTextures    = *lpsAnimatedTextures++;

					strDump.reserve( 5000000 );

					strDump = CRLF "Animated textures:" CRLF CRLF;

					while(wNumAnimatedTextures--)
						{
						bitu16   wNum = *lpsAnimatedTextures++;  wNum++;
						while(wNum--)
							{
							bitu16  wNumObjText = *lpsAnimatedTextures;

							TR2_OBJECT_TEXTURE* lpObjText = &lp->lpObjectTextures[ wNumObjText & 0x7fff ];

							DWORD dwTile = lpObjText->wTile & 0x7fff;

							t.Format( "%.4d ", (long) wNumObjText );

							strDump += t;

							if( dwTile >= dwNumNoBump  &&  dwTile < dwNumNoBump + dwNumInhab )
								{
								strDump += " !! ";
								boOk = true;
								}

							lpsAnimatedTextures++;
							}
						strDump += CRLF;
						}

					char szz[123];
					sprintf(szz,"c:\\ts%s.txt",trFile.lpszGetShortLevelName());

					FILE* f6 = fopen(szz,"wb");
    				fwrite((LPCTSTR)strDump.c_str(),1,strDump.length(),f6);
					fclose(f6);
                   }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

	if( boOk )
		AfxMessageBox("The end !!!!!!!!!!!!!!!!!!!");
	else
		AfxMessageBox("The end");
}


void vGetListLevelsOfMvb(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

	CString t;

	typedef std::list<std::string>	LstStrings;

	LstStrings	MvbId[600];

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();
					TR123_FILE_STRUCT* lp = (TR123_FILE_STRUCT*) &m_trFile.tr123File;

                    dwNiv++;

					for(DWORD dwMvb = 0; dwMvb < m_trFile.tr123File.dwNumMoveables; ++dwMvb)
						{
						TR2_MOVEABLE* lpMvb = &m_trFile.tr123File.lpMoveables[dwMvb];

						MvbId[lpMvb->dwObjectId].push_back(trFile.lpszGetShortLevelName());
						}
                   }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

	std::string	str;

	str.reserve(1000000);

	for( DWORD dwMvb = 0; dwMvb < 600; ++dwMvb )
		{
		if( !MvbId[dwMvb].empty() )
			{
			t.Format( "\nMoveable #%.4d\n", dwMvb );

			str += t;
			for(LstStrings::iterator theIter = MvbId[dwMvb].begin(); theIter != MvbId[dwMvb].end(); ++theIter )
				{
				str += "\t";
				str += theIter->c_str();
				str += "\n";
				}
			}
		}

    FILE* f1 = fopen("c:\\trmvb.txt","wb");
    
    fwrite((LPCTSTR)str.c_str(),1,str.length(),f1);

    fclose(f1);

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void vConvertADPCMToPCM_( WAVEFORMATEX& wfxSrc, WAVEFORMATEX& wfxDst, DWORD dwCompressedSize, BYTE* lpbSrc, DWORD& dwUncompressedSize, BYTE* lpbDst )
{
	HACMSTREAM		has;

	MMRESULT	res = acmStreamOpen( &has, NULL, &wfxSrc, &wfxDst, NULL, NULL, NULL, 0 );

	if( res != 0 )
		{
		return;
		}

	ACMSTREAMHEADER	ash;

	ZeroMemory( &ash, sizeof( ash ) );

	ash.cbStruct = sizeof( ash );
	ash.pbSrc = lpbSrc;
	ash.cbSrcLength = dwCompressedSize;
	ash.pbDst = lpbDst;
	ash.cbDstLength = dwUncompressedSize;

	res = acmStreamPrepareHeader( has, &ash, 0 );

	if( res != 0 )
		{
		acmStreamClose( has, 0 );

		return;
		}

	res = acmStreamConvert( has, &ash, 0 );

	if( res != 0 )
		{
		acmStreamUnprepareHeader( has, &ash, 0 );
		acmStreamClose( has, 0 );

		return;
		}

	_ASSERTE( ash.cbSrcLength == ash.cbSrcLengthUsed );

	dwUncompressedSize = ash.cbDstLengthUsed;

	acmStreamUnprepareHeader( has, &ash, 0 );
	acmStreamClose( has, 0 );
}


void vExtractSounds(TCHAR* lpszBuf)
{
    WIN32_FIND_DATA fd;

    TCHAR   szCurDir[MAX_PATH];
    GetCurrentDirectory(MAX_PATH,szCurDir);

    SetCurrentDirectory(lpszBuf);

    TCHAR lpszFilter[] = _T("*.tr4");

    HANDLE  h = FindFirstFile(lpszFilter,&fd);

    if(h != INVALID_HANDLE_VALUE)
        {
        DWORD dwNiv = 0;

        do
            {
            if(!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
                {
                CTRFile trFile;

                if(trFile.boLoad(fd.cFileName))
                    {
                    TR_FILE& m_trFile = trFile.GetTRFile();

                    dwNiv++;

					DWORD	dwNumSounds = m_trFile.tr123File.dwNumSounds;

					for( DWORD dwI = 0; dwI < dwNumSounds; ++dwI )
						{
						TR4_SOUND* lpSound = &m_trFile.tr123File.lpSounds[ dwI ];

						DWORD	dwUnk = lpSound->dwUncompSoundLen;
						DWORD	dwSoundLen = lpSound->dwSoundLen;

						BYTE* lpb = lpSound->lpbData;

						WAVEFORMATEX*	lpwfxSrc =(WAVEFORMATEX*) ( lpb + 20 );
						WAVEFORMATEX	wfxDst = *lpwfxSrc;

						wfxDst.wFormatTag = 1;	// not compressed;
						wfxDst.nSamplesPerSec = 22050;
						wfxDst.wBitsPerSample = 16;
						wfxDst.nBlockAlign = wfxDst.nChannels * wfxDst.wBitsPerSample / 8;
						wfxDst.nAvgBytesPerSec = wfxDst.nSamplesPerSec * wfxDst.nBlockAlign;
						wfxDst.cbSize = 0;

						BYTE bHeader[] = {
									0x52, 0x49, 0x46, 0x46, 0, 0, 0, 0, 0x57, 0x41, 0x56, 0x45, 0x66, 0x6d, 0x74, 0x20, 0x10, 0, 0, 0,	// 20
									0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 16
									0x64, 0x61, 0x74, 0x61, 0, 0, 0, 0 // 8
									 };

						dwUnk *= 4;

						BYTE*	lpbSrc = lpb + 0x5A;
						DWORD	dwSrcSize = dwSoundLen - 0x5A;
						BYTE*	lpbDst = new BYTE [ dwUnk ];
						DWORD	dwDstSize = dwUnk - sizeof( bHeader );

						memcpy( lpbDst, bHeader, sizeof( bHeader ) );
						memcpy( lpbDst + 20, &wfxDst, sizeof( wfxDst ) - 2 );

						vConvertADPCMToPCM_( *lpwfxSrc, wfxDst, dwSrcSize, lpbSrc, dwDstSize, lpbDst + sizeof( bHeader ) );

						* ( (DWORD*) &lpbDst[ 4 ] ) = dwDstSize + sizeof( bHeader ) - 8;
						* ( (DWORD*) &lpbDst[ 40 ] ) = dwDstSize;

						_ASSERTE( dwSrcSize * 4 > dwDstSize );

						char sz[123];

						wsprintf(sz,"c:\\temp\\snd\\Snd#%.4d.wav",dwI);

						DWORD dwNumWritten;

						HANDLE ff = CreateFile(sz,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);

						_ASSERTE( ff != INVALID_HANDLE_VALUE );

						WriteFile( ff, lpb, dwSoundLen, &dwNumWritten, NULL );
						CloseHandle( ff );

						wsprintf(sz,"c:\\temp\\snd\\Snd#%.4d_.wav",dwI);
						
						ff = CreateFile(sz,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);

						_ASSERTE( ff != INVALID_HANDLE_VALUE );

						WriteFile( ff, lpbDst, dwDstSize + sizeof( bHeader ), &dwNumWritten, NULL );
						CloseHandle( ff );

						delete [] lpbDst;
						}

					break;
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Can't load %s (this file may not be a TR file)"),fd.cFileName);
                    AfxMessageBox(s);
                    }
                }
            }
        while(FindNextFile(h,&fd));

        FindClose(h);
        }

    SetCurrentDirectory(szCurDir);

    AfxMessageBox("The end");
}


void    vMiscDebug() 
{
    TCHAR szBuf[MAX_PATH];
//    TCHAR szBuf[MAX_PATH] = "D:\\TombRaider\\Data\\DataTR4\\bikebit.tr4";
//    TCHAR szBuf[MAX_PATH] = "D:\\TombRaider\\Data\\DataTR4\\library.tr4";

    if(!boChooseFolder(szBuf)) return;

    CWaitCursor dummy;

//    vCheckStaticMeshesFlags(szBuf);
//    vCountSfx(szBuf);
//    vExportAnmText(szBuf);
//    vExportSprSeq(szBuf);
//    vCountMeshes(szBuf);
//    vCheckAnimCommands(szBuf);
//    vCompareLevels("d:\\tombraider\\data\\datatr2\\wall.tr2","d:\\temp\\wall.tr2");
//    vCheckSounds(szBuf);
//    vGetNumSamplesPerSound(szBuf);
//	vCompareNumTiles(szBuf);
//	vCheckObjAttrFlag(szBuf);
//	vCheckAnimUnknown(szBuf);
//	vGenerateXML(szBuf);
//	vTestEuler2Quat();

//	vReadFromTextFile( "c:\\trlibrary.txt", szBuf, "D:\\Program Files\\Core Design\\trle\\data\\tut1.tr4" );

//	vCheckAnimText(szBuf);

//	vGetListLevelsOfMvb(szBuf);

//	vExtractSounds(szBuf);
	vGetStateId(szBuf);
}
