#include "stdafx.h"

#include "Resource.h"
#include "MainFrm.h"
#include "ChildFrm.h"
#include "FileTree.h"
#include "HandleBase.h"
#include "HandleStaticMeshes.h"
#include "HandleMesh.h"
#include "EditBase.h"
#include "EditMesh.h"


void    CHandleMesh::vEditSetInfo(TCHAR* lpszNodeName)
{
    CEditBase*  lpEdit = lpGetEditView();

    if( lpEdit != NULL ) dynamic_cast< CEditMesh* >( lpEdit )->vSetDataTypeHandler(this);
}


CMenu*  CHandleMesh::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    if(m_boStaticMesh) m_lpMenu->LoadMenu(IDR_MENU_STATICMESH); else m_lpMenu->LoadMenu(IDR_MENU_MESH);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleMesh::vReleasePopupupMenu()
{
    delete m_lpMenu;
    
    m_lpMenu = NULL;
}


void    CHandleMesh::vSetChanged(bool boReloadTextures)
{
    if(boReloadTextures)
        {
		CEditBase*  lpEdit = lpGetEditView();

		if( lpEdit != NULL ) dynamic_cast< CEditMesh* >( lpEdit )->vReloadTextures();
        }

    CHandleBase::vSetChanged();
}


void    CHandleMesh::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Delete the static mesh
        case ID_STATICMESH_DELETE:
            {
            vDelete();
            break;
            }
        // Export the mesh to a .3DS file
        case ID_MESH_EXPORTTO3DS:
        case ID_SMESH_EXPORTTO3DS:
            {
            vExportTo3DS();
            break;
            }
        // Load a 3ds file and replace the current mesh
        case ID_MESH_IMPORTFROM3DS:
        case ID_SMESH_IMPORTFROM3DS:
            {
            vImportFrom3DS();
            break;
            }
        // Replace the current mesh with an empty mesh
        case ID_MESH_CLEAR:
        case ID_SMESH_CLEAR:
            {
            vClearMesh();
            break;
            }
        // Export the mesh to a .TRMSH / .TRSMH file
        case ID_MESH_EXPORTASMSH:
        case ID_SMESH_EXPORTASMSH:
            {
            vExportAsMSH();
            break;
            }
        // Import the mesh from a .TRMSH / .TRSMH file
        case ID_MESH_IMPORTFROMMSH:
        case ID_SMESH_IMPORTFROMMSH:
            {
            vImportFromMSH();
            break;
            }
        // Export the texture to a .TGA file
        case ID_MESH_EXPORTTEXTUREASTGA:
        case ID_SMESH_EXPORTTEXTUREASTGA:
            {
            vExportTextureToTGA();
            break;
            }
        // Import the texture from a .TGA file
        case ID_MESH_IMPORTTEXTUREFROMTGA:
        case ID_SMESH_IMPORTTEXTUREFROMTGA:
            {
            vImportTextureFromTGA();
            break;
            }
        }
}


void    CHandleMesh::vExportTo3DS()
{
    // Get the filename
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("3ds"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    DWORD       dwNumMesh;
    CTrLstMsh*  lptrMSH = GetMesh(dwNumMesh);

    CString cstrFilePath = FileDialog.GetPathName();
    TCHAR*  lpszFilePath = (TCHAR*)(LPCTSTR)cstrFilePath;
    TCHAR   lpszObjName[_MAX_FNAME];
    TCHAR   szTextureName[9];

    _tsplitpath(lpszFilePath,NULL,NULL,lpszObjName,NULL);
    if(lstrlen(lpszObjName) <= 8)
        {
        lstrcpy(szTextureName, lpszObjName);
        }
    else
        {
        lstrcpy(szTextureName, &lpszObjName[lstrlen(lpszObjName) - 8]);
        }

    if(m_boStaticMesh)
		{
		boExportMSHTo3DS(*lptrMSH,dwNumMesh,lpszFilePath,szTextureName);
		}
	else
		{
		CTrLstMsh	Msh( *lptrMSH, dwNumMesh );

		boExportMSHTo3DS( Msh,0,lpszFilePath,szTextureName );
		}
}


void    CHandleMesh::vImportFrom3DS()
{
    BYTE* lpData = NULL;

    // Load the 3DS file
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    DWORD       dwNumMesh;
    CTrLstMsh*  lpMsh = GetMesh(dwNumMesh);

    DWORD dwObjId = lpMsh->dwGetObjectId();

	CString	cstrPathname = FileDialog.GetPathName();

    CTrLstMsh  Msh;
   
    if(lpTRFile->boImportMeshFrom3DS(&Msh,(LPTSTR)(LPCTSTR)cstrPathname,m_boStaticMesh,lpTRFile->GetTRType() == TR4_FILE))
        {
        if(m_boStaticMesh)
            {
			lpMsh->vRelease();
			lpMsh->vStealData(Msh);

            TR123_MESH* lpMesh = const_cast<TR123_MESH*>(lpMsh->lpGetMesh(0));

            lpMesh->lCollisionSize = 0;
            CTRFile::vComputeMeshExtents(lpMesh);

            CTrStatMsh* lpStatMsh = static_cast<CTrStatMsh*>(lpMsh);
            CBoundingBox    BB;

            lpStatMsh->vSetObjectId(0);
            lpStatMsh->vComputeBoundingBox(BB,0);
            lpStatMsh->vSetBBVisibility(BB);
            lpStatMsh->vSetBBCollision(BB);
            lpStatMsh->vSetFlags(TRSTATMSH_FLAGS_VISIBLE);
            }
		else
			{
			lpMsh->vReplaceMesh( Msh, 0, dwNumMesh );

            TR123_MESH* lpMesh = const_cast<TR123_MESH*>(lpMsh->lpGetMesh(0));

            lpMesh->lCollisionSize = 0;

            CTRFile::vComputeMeshExtents(lpMesh);
			}

        lpMsh->vSetObjectId(dwObjId);

        vSetChanged(true);
        vEditSetInfo(m_szNodeName);
        }
}


void    CHandleMesh::vClearMesh()
{
    DWORD dwNumMesh;
    CTrLstMsh* lpMesh = GetMesh(dwNumMesh);

    lpMesh->vClearMesh(dwNumMesh);

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Save as a .TRMSH
//
//
void    CHandleMesh::vExportAsMSH()
{
    // Get the filename
    CString strFilter(_T("TR Mesh files (*.trmsh)|*.trmsh||"));
    CString strExt(_T("trmsh"));

    if(m_boStaticMesh)
        {
        strFilter = _T("TR Static Mesh files (*.trsmh)|*.trsmh||");
        strExt = _T("trsmh");
        }

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,strExt,strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

	CString strFile = FileDialog.GetPathName();

	if( strFile.Find( strExt ) == -1 ) strFile.Replace( _T(".trm"), _T(".") + strExt );	// fix a bug with Win95 which truncate .trmsh to .trm

    CWaitCursor dummy;

    if(m_boStaticMesh)
        {
        DWORD dwNumMesh;
        CTrStatMsh*  lpMsh = static_cast<CTrStatMsh*>(GetMesh(dwNumMesh));

        lpMsh->boSave(strFile);
        }
    else
        {
        DWORD dwNumMesh;
        CTrLstMsh*  lpMsh = GetMesh(dwNumMesh);

		CTrLstMsh	Msh( *lpMsh, dwNumMesh );

        Msh.boSave(strFile);
        }
}


//
//
// Load a .TRMSH
//
//
void    CHandleMesh::vImportFromMSH(CTrLstMsh* lpMsh)
{
    CString cstrPath;

    // Load the MSH file
    if(lpMsh == NULL)
        {
        CString strFilter(_T("TR Mesh files (*.trmsh)|*.trmsh||"));

        if(m_boStaticMesh)
            {
            strFilter = _T("TR Static Mesh files (*.trsmh)|*.trsmh||");
            }

        CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

        if(FileDialog.DoModal() != IDOK) return;

        cstrPath = FileDialog.GetPathName();
        }

    CWaitCursor dummy;

    CTRFile*    lpTRFile = m_lpTree->lpGetTRFile();
    bool        boChanged = true;

    if(m_boStaticMesh)
        {
        DWORD dwNumMesh;
        CTrStatMsh* lpStaticMesh = static_cast<CTrStatMsh*>(GetMesh(dwNumMesh));

        DWORD dwObjId = lpStaticMesh->dwGetObjectId();

        // Build the TR static mesh
        if(lpMsh == NULL)
            {
            CTrStatMsh* Msh = CTrStatMsh::objLoad(cstrPath);

            if(Msh)
                {
                lpStaticMesh->vRelease();
                lpStaticMesh->vStealData(*Msh);

                delete Msh;
                }
            else
                {
                boChanged = false;
                AfxMessageBox(_T("Could not import this static mesh because of a version mismatch:\n\n - the static mesh was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
                }
            }
        else
            {
            lpStaticMesh->vRelease();
            lpStaticMesh->vStealData(*(static_cast<CTrStatMsh*>(lpMsh)));
            }

        lpStaticMesh->vSetObjectId(dwObjId);
        }
    else
        {
        DWORD dwNumMesh;
        CTrLstMsh* lpMesh = GetMesh(dwNumMesh);

        DWORD dwObjId = lpMesh->dwGetObjectId();

        // Build the TR mesh
        if(lpMsh == NULL)
            {
            CTrLstMsh* Msh = CTrLstMsh::objLoad(cstrPath);

			if( Msh )
				{
				lpMesh->vReplaceMesh( *Msh, 0, dwNumMesh );

				delete Msh;
				}
			else
				{
                boChanged = false;
                AfxMessageBox(_T("Could not import this mesh because of a version mismatch:\n\n - the mesh was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
				}
            }
        else
            {
			lpMesh->vRelease();
            lpMesh->vStealData(*lpMsh);
            }

        lpMesh->vSetObjectId(dwObjId);
        }

    if(boChanged) vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Export the texture to a .TGA file
//
//
void    CHandleMesh::vExportTextureToTGA()
{
    // Get the filename
    CString strFilter(_T("TGA files (*.tga)|*.tga||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("tga"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    // Save the texture
    DWORD       dwNumMesh;
    CTrLstMsh*  lpMesh = GetMesh(dwNumMesh);

    if(m_boStaticMesh)
		{
		lpMesh->vSaveTextureAsTGA((LPCTSTR)FileDialog.GetPathName());
		}
	else
		{
		CTrLstMsh	Msh( *lpMesh, dwNumMesh );

        Msh.vSaveTextureAsTGA((LPCTSTR)FileDialog.GetPathName());
		}
}


//
//
// Import the texture from a .TGA file
//
//
void    CHandleMesh::vImportTextureFromTGA()
{
    // Load the TGA file
    CString strFilter(_T("Gfx files (*.tga;*.bmp;*.gif)|*.tga;*.bmp;*.gif|Others (*.fli;*.flc;*.cel)|*.fli;*.flc;*.cel||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    // Load and replace the texture
    DWORD       dwNumMesh;
    CTrLstMsh*  lpMesh = GetMesh(dwNumMesh);

    if(m_boStaticMesh)
		{
		bool boOk = lpMesh->boLoadTextureFromTGA((LPCTSTR)FileDialog.GetPathName());

		if(!boOk)
			{
			CString s;
        
			s.Format(   _T("Could not replace the texture. Please, make sure that:\n\n")
						_T("    - your graphic file is a valid one\n")
						_T("    - your texture is %ldx%ld"),
						lpMesh->dwGetTextureWidth(), lpMesh->dwGetTextureHeight());

			AfxMessageBox(s,MB_OK | MB_ICONSTOP | MB_TOPMOST);

			return;
			}
		}
	else
		{
		CTrLstMsh	Msh( *lpMesh, dwNumMesh );

        if( !Msh.boLoadTextureFromTGA((LPCTSTR)FileDialog.GetPathName()) )
			{
			CString s;
        
			s.Format(   _T("Could not replace the texture. Please, make sure that:\n\n")
						_T("    - your graphic file is a valid one\n")
						_T("    - your texture is %ldx%ld"),
						lpMesh->dwGetTextureWidth(), lpMesh->dwGetTextureHeight());

			AfxMessageBox(s,MB_OK | MB_ICONSTOP | MB_TOPMOST);

			return;
			}

		lpMesh->vReplaceMesh( Msh, 0, dwNumMesh );
		}

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Delete the mesh
//
//
void    CHandleMesh::vDelete()
{
    if(AfxMessageBox(_T("Are you sure you want to delete this static mesh ?"), MB_YESNO | MB_ICONEXCLAMATION | MB_TOPMOST) != IDYES) return;

    DWORD       dwNumMesh;

    CTrStatMsh* lpStatMsh = static_cast<CTrStatMsh*>(GetMesh(dwNumMesh));

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    TRListStaticMesh* lpList = lpTRFile->GetTRFile().tr123File.lpListStaticMesh;

    for(TRListStaticMesh::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
        {
        if(*theIterator == lpStatMsh)
            {
            lpList->erase(theIterator);

			CEditBase*  lpEdit = lpGetEditView();

			if( lpEdit != NULL ) dynamic_cast< CEditMesh* >( lpEdit )->vRelease();

            delete lpStatMsh;

            vSetChanged();
            m_lpTree->m_lpHandleStaticMeshes->vExpandTree(true);

            break;
            }
        }
}
