#include "stdafx.h"

#include "Resource.h"
#include "MainFrm.h"
#include "ChildFrm.h"
#include "FileTree.h"
#include "HandleBase.h"
#include "HandleMoveables.h"
#include "HandleMoveable.h"
#include "HandleMesh.h"
#include "EditBase.h"
#include "EditMoveable.h"

#include "AnimData.h"
#include "AnimKey.h"
#include "AnimTrack.h"
#include "AnimMgr.h"
#include "TRAnimTrack.h"

#include "ImportData.h"
#include "Import3DS.h"
#include "ExportMvbl.h"

#include "ToolboxBase.h"
#include "ToolboxItfBase.h"
#include "TlboxMoveableItf.h"
#include "Toolbox2Moveable.h"


void    CHandleMoveable::vExpandTree(bool boAlreadyExpanded)
{
    if(boAlreadyExpanded) return;

    CString             s;
    DWORD               dwI;
    HTREEITEM           h;
    CTRFile*            lpTRFile    = m_lpTree->lpGetTRFile();
    enumTRType          trType      = lpTRFile->GetTRType();

    TCHAR               szMoveableName[256+1];
    
    lstrcpyn(szMoveableName,lpszGetTitle(),256+1);

    switch(trType)
        {
        //
        case TR1_FILE:
        case TR2_FILE:
        case TR3_FILE:
        case TR4_FILE:
            {
            CTrMvb* lpMvb = GetMoveable();
            for(dwI = 0; dwI < lpMvb->dwGetNumMeshes(); dwI++)
                {
                s.Format(_T("Mesh_%.3d (#%.4d)"),dwI,lpMvb->dwEdtGetStartingMesh() + dwI);
                INSERT(m_lpTree,h,m_hParentItem,new CHandleMesh(m_lpTree,m_lpChildFrm,(void*)lpMvb,false,dwI))
				m_lpTree->SetItemImage( h, 0, 0 );
                }
            break;
            }
        }
}


void    CHandleMoveable::vEditSetInfo(TCHAR* lpszNodeName)
{
    CEditBase*  lpEdit = lpGetEditView();

    if( lpEdit != NULL ) dynamic_cast< CEditMoveable* >( lpEdit )->vSetDataTypeHandler(this);
}


CMenu*  CHandleMoveable::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    m_lpMenu->LoadMenu(IDR_MENU_MOVEABLE);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleMoveable::vReleasePopupupMenu()
{
    delete m_lpMenu;
    
    m_lpMenu = NULL;
}


void    CHandleMoveable::vSetChanged(bool boReloadTextures)
{
    if(boReloadTextures)
        {
		CEditBase*  lpEdit = lpGetEditView();

		if( lpEdit != NULL ) dynamic_cast< CEditMoveable* >( lpEdit )->vReloadTextures();
        }

    CHandleBase::vSetChanged();
}


void    CHandleMoveable::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Delete the moveable
        case ID_MOVEABLE_DELETE:
            {
            vDelete();
            break;
            }
        // Export the moveable to a .3DS file
        case ID_MOVEABLE_EXPORTTO3DS:
            {
            vExportTo3DS();
            break;
            }
        // Load a 3ds file and replace the current moveable
        case ID_MOVEABLE_IMPORTFROM3DS:
            {
            vImportFrom3DS();
            break;
            }
        // Replace the meshes of the moveable with empty meshes
        case ID_MOVEABLE_CLEAR:
            {
            vClearMoveable();
            break;
            }
        // Export the moveable to a .TRMVB file
        case ID_MOVEABLE_EXPORTASMVB:
            {
            vExportAsMVB();
            break;
            }
        // Load a trmvb file and replace the current moveable
        case ID_MOVEABLE_IMPORTFROMMVB:
            {
            vImportFromMVB();
            break;
            }
        // Export the texture to a .TGA file
        case ID_MOVEABLE_EXPORTTEXTUREASTGA:
            {
            vExportTextureToTGA();
            break;
            }
        // Import the texture from a .TGA file
        case ID_MOVEABLE_LOADREPLACETEXTURE:
            {
            vImportTextureFromTGA();
            break;
            }
        // Remove all sound references into the moveable (in the anim commands)
        case ID_MOVEABLE_REMOVESOUNDS:
            {
            vRemoveSounds();
            break;
            }
        }
}


void    CHandleMoveable::vExportTo3DS()
{
    CTrMvb*  lptrMVB = GetMoveable();

	CToolbox2Moveable* lpToolbox = GET_TOOLBOX( CToolbox2Moveable );

    DWORD   dwCurTrack = lpToolbox->dwGetCurTrack();

    if( dwCurTrack ) dwCurTrack--;

    //
    CExportMvbl     ExportDlg( lptrMVB->lpGetAnimMgr(), dwCurTrack );

    if( ExportDlg.DoModal() != IDOK ) return;

    DWORD   dwNumTrack = ExportDlg.dwGetSelectedAnim();

    // Get the filename
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("3ds"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    CString cstrFilePath = FileDialog.GetPathName();
    TCHAR*  lpszFilePath = (TCHAR*)(LPCTSTR)cstrFilePath;
    TCHAR   lpszObjName[_MAX_FNAME];
    TCHAR   szTextureName[9];

    _tsplitpath(lpszFilePath,NULL,NULL,lpszObjName,NULL);
    if(lstrlen(lpszObjName) <= 8)
        {
        lstrcpy(szTextureName, lpszObjName);
        }
    else
        {
        lstrcpy(szTextureName, &lpszObjName[lstrlen(lpszObjName) - 8]);
        }

    boExportMoveableTo3DS(*lptrMVB,lpszFilePath,szTextureName,m_lpTree->lpGetTRFile(),dwNumTrack);
}


void    CHandleMoveable::vImportFrom3DS()
{
    BYTE* lpData = NULL;

    CTrMvb* lpMoveable = GetMoveable();

	CToolbox2Moveable* lpToolbox = GET_TOOLBOX( CToolbox2Moveable );

    DWORD   dwCurTrack = lpToolbox->dwGetCurTrack();

    if( dwCurTrack ) dwCurTrack--;

    //
    CImport3DS  ImportDlg( lpMoveable->lpGetAnimMgr(), dwCurTrack );

    if( ImportDlg.DoModal() != IDOK ) return;

    DWORD   dwNumTrack = ImportDlg.dwGetSelectedAnim();

    bool    boImportMeshes  = ImportDlg.boImportMeshes() ? true : false;
    bool    boReplaceAnim   = ImportDlg.boReplaceAnimation() ? true : false;
    bool    boDeleteAnims   = ImportDlg.boDeleteAnimations() ? true : false;
    bool    boAddAnim       = ImportDlg.boAddAnimation() ? true : false;
    bool    boDontOverwrite = ImportDlg.boDontOverwriteDummyMeshes() ? true : false;

    // Load the 3DS file
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    CTrMvb* lpNewMvb = new CTrMvb;

    if( !lpTRFile->boImportMoveableFrom3DS( lpNewMvb, (TCHAR*)(LPCTSTR)FileDialog.GetPathName(), false, lpTRFile->GetTRType() == TR4_FILE) )
        {
        }
    else
        {
        if( boDeleteAnims )
            {
            lpNewMvb->vReleaseAllAnims();
            lpMoveable->vReleaseAllAnims();
            }

        if( boImportMeshes )
            {
			lpNewMvb->vSetObjectId( lpMoveable->dwGetObjectId() );
			if( boDontOverwrite ) lpNewMvb->vSetDummyMeshes( *lpMoveable );
            lpMoveable->vRelease2(true,true,false);
            lpMoveable->vStealData( *lpNewMvb, true, false );
            }
        else if( !boDeleteAnims )
            {
            if( lpMoveable->dwGetNumMeshes() != lpNewMvb->dwGetNumMeshes() )
                {
                ::AfxMessageBox( _T("Imported 3DS file and current moveable must have the same number of meshes !"), MB_ICONSTOP | MB_OK | MB_TOPMOST );
                }
            else
                {
                if( boAddAnim )
                    {
                    lpMoveable->lpGetAnimMgr()->vStealTrack( -1, *lpNewMvb->lpGetAnimMgr(), 0 );
                    }
                else if( boReplaceAnim )
                    {
                    CAnimMgr*       lpAnimMgrOld = lpMoveable->lpGetAnimMgr();
                    CAnimMgr*       lpAnimMgrNew = lpNewMvb->lpGetAnimMgr();

                    DWORD           dwNextTrack, dwNextFrame;
                    
                    lpAnimMgrOld->vGetNextAnimInfos( dwNextTrack, dwNextFrame, dwNumTrack );

                    CTRAnimTrack*   lpTrackOld = static_cast<CTRAnimTrack*>( lpAnimMgrOld->lpGetTrack( dwNumTrack ) );
                    CTRAnimTrack*   lpTrackNew = static_cast<CTRAnimTrack*>( lpAnimMgrNew->lpGetTrack( 0 ) );

                    lpTrackNew->vStealData( *lpTrackOld );

                    lpAnimMgrOld->vStealTrack( dwNumTrack, *lpAnimMgrNew, 0 );
                    lpAnimMgrOld->vSetNextTrack( dwNumTrack, dwNextTrack, dwNextFrame );
                    }
                }
            }

        vSetChanged( true );
        vEditSetInfo( m_szNodeName );
        }

    delete lpNewMvb;
}


void    CHandleMoveable::vClearMoveable()
{
    CTrMvb*         lpMoveable = GetMoveable();

    for(DWORD dwI = 0; dwI < lpMoveable->dwGetNumMeshes(); dwI++)
        {
        lpMoveable->vClearMesh(dwI);
        }

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Save as a .TRMVB
//
//
void    CHandleMoveable::vExportAsMVB()
{
    // Get the filename
    CString strFilter(_T("TR Moveable files (*.trmvb)|*.trmvb||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("trmvb"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

	CString strFile = FileDialog.GetPathName();

	if( strFile.Find( _T(".trmvb") ) == -1 ) strFile.Replace( _T(".trm"), _T(".trmvb") );	// fix a bug with Win95 which truncate .trmvb to .trm

    CWaitCursor dummy;

    CTrMvb*  lptrMVB = GetMoveable();

    lptrMVB->boSave(strFile);
}


void    CHandleMoveable::vImportFromMVB(CTrMvb* lpMvb)
{
    CTrMvb* lpMoveable = GetMoveable();

    DWORD dwOldObjId = lpMoveable->dwGetObjectId();

    bool boImportMeshes = true;
    bool boImportAnimations = true;
	bool boDontOverwrite = false;

    // Load the MVB file
    if(lpMvb == NULL)
        {
        // What data to import ?
        CImportData Dlg;

        int iRes = Dlg.DoModal();

        if(iRes != IDOK) return;

        CString strFilter(_T("TR moveable files (*.trmvb)|*.trmvb||"));
        CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

        if(FileDialog.DoModal() != IDOK) return;

        boImportMeshes = Dlg.boImportMeshes();
        boImportAnimations = Dlg.boImportAnimations();
        boDontOverwrite = Dlg.boDontOverwriteDummyMeshes();

        // Build the TR moveable
        CTrMvb* Mvb = CTrMvb::objLoad(FileDialog.GetPathName());

        if(Mvb)
            {
            if(!boImportAnimations  &&  boImportMeshes  &&  Mvb->dwGetNumMeshes() != lpMoveable->dwGetNumMeshes())
                {
                AfxMessageBox( _T("Can't import only meshes from this moveable because it has a different number of meshes than the current one !"), MB_ICONSTOP | MB_TOPMOST);
                return;
                }

            if(boImportAnimations) m_lpTree->lpGetTRFile()->GetTRFile().tr123File.lpSndMgr->vMerge(lpMoveable->lpGetSndMgr(),true);

			Mvb->vEdtSetAssocMvbId( lpMoveable->dwEdtGetAssocMvbId() );
			Mvb->vEdtSetBodyMvbId( lpMoveable->dwEdtGetBodyMvbId() );
			Mvb->vEdtSetShowAssocMvb( lpMoveable->boEdtShowAssocMvb() );
			Mvb->vEdtSetShowRealMeshes( lpMoveable->boEdtShowRealMeshes() );

			if( boDontOverwrite ) Mvb->vSetDummyMeshes( *lpMoveable );

            lpMoveable->vRelease2(true,boImportMeshes,boImportAnimations);
            lpMoveable->vStealData(*Mvb,boImportMeshes,boImportAnimations);

            delete Mvb;

            if(!boImportAnimations) lpMoveable->vSetObjectId(dwOldObjId);
            }
        else
            {
            AfxMessageBox(_T("Could not import this moveable because of a version mismatch:\n\n - the moveable was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
            return;
            }
        }
    else
        {
        m_lpTree->lpGetTRFile()->GetTRFile().tr123File.lpSndMgr->vMerge(lpMoveable->lpGetSndMgr(),true);
        lpMoveable->vRelease();
        lpMoveable->vStealData(*lpMvb);
        }

    if(boImportAnimations)
        {
        m_lpTree->lpGetTRFile()->GetTRFile().tr123File.lpSndMgr->vMerge(lpMoveable->lpGetSndMgr());
        m_lpTree->lpGetTRFile()->vMakeSureSoundsAreAvailable();
        }

    CWaitCursor dummy;

    DWORD dwNewObjId = lpMoveable->dwGetObjectId();

    CTRFile*            lpTRFile = m_lpTree->lpGetTRFile();
    TR123_FILE_STRUCT&  trFile = lpTRFile->GetTRFile().tr123File;
    bool                boTR1 = lpTRFile->GetTRType() == TR1_FILE;

    // Replace the old object id in items list by the new one
    TRListItem* lpListItem = trFile.lpListItem;

    for(TRListItem::iterator theIterator4 = lpListItem->begin(); theIterator4 != lpListItem->end(); theIterator4++)
        {
        TR123_ITEM* lpItem= *theIterator4;

        if(boTR1)
            {
            if(lpItem->tr1.sObjectId == dwOldObjId) lpItem->tr1.sObjectId = dwNewObjId;
            }
        else
            {
            if(lpItem->tr2.sObjectId == dwOldObjId) lpItem->tr2.sObjectId = dwNewObjId;
            }
        }

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Export the texture to a .TGA file
//
//
void    CHandleMoveable::vExportTextureToTGA()
{
    // Get the filename
    CString strFilter(_T("TGA files (*.tga)|*.tga||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("tga"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    // Save the texture
    CTrMvb*     lpMvb = GetMoveable();

    lpMvb->vSaveTextureAsTGA((LPCTSTR)FileDialog.GetPathName());
}


//
//
// Import the texture from a .TGA file
//
//
void    CHandleMoveable::vImportTextureFromTGA()
{
    // Load the TGA file
    CString strFilter(_T("Gfx files (*.tga;*.bmp;*.gif)|*.tga;*.bmp;*.gif|Others (*.fli;*.flc;*.cel)|*.fli;*.flc;*.cel||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    // Load and replace the texture
    CTrMvb*     lpMvb = GetMoveable();

    bool boOk = lpMvb->boLoadTextureFromTGA((LPCTSTR)FileDialog.GetPathName());

    if(!boOk)
        {
        CString s;
        
        s.Format(   _T("Could not replace the texture. Please, make sure that:\n\n")
                    _T("    - your graphic file is a valid one\n")
                    _T("    - your texture is %ldx%ld"),
                    lpMvb->dwGetTextureWidth(), lpMvb->dwGetTextureHeight());

        AfxMessageBox(s,MB_OK | MB_ICONSTOP | MB_TOPMOST);

        return;
        }

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Remove all sound references from the anim commands
//
//
void    CHandleMoveable::vRemoveSounds()
{
    CTrMvb*     lpMvb       = GetMoveable();
    CAnimMgr*   lpAnimMgr   = lpMvb->lpGetAnimMgr();

    CTRFile*            lpTRFile = m_lpTree->lpGetTRFile();
    TR123_FILE_STRUCT&  trFile = lpTRFile->GetTRFile().tr123File;

    trFile.lpSndMgr->vMerge(lpMvb->lpGetSndMgr(),true);

    for(DWORD dwTrack = 0; dwTrack < lpAnimMgr->dwGetNumTracks(); dwTrack++)
        {
        CTRAnimTrack*   lpTrack     = static_cast<CTRAnimTrack*>(lpAnimMgr->lpGetTrack(dwTrack));

        DWORD dwI = lpTrack->dwGetNumAnimCommands();

        while(dwI--)
            {
            const CTRAnimCommand& AnimCmde = lpTrack->GetAnimCommand(dwI);

            bit16   sCmd = AnimCmde.sGetCommand().sValue;

            if( AnimCmde.boIsSound() )
                {
                if(sCmd == ANIMCMD_PLAYSOUNDONFRAME_2)
                    {
                    bit16   wSound = AnimCmde.sGetParam(1).sValue & 0x3FFF;
                    lpMvb->lpGetSndMgr()->vRemoveSound(wSound);
                    }
                lpTrack->vRemoveAnimCommand(dwI);
                dwI = lpTrack->dwGetNumAnimCommands();
                }
            }
        }

    lpTRFile->vMakeSureSoundsAreAvailable();

    vSetChanged(true);

    vEditSetInfo(m_szNodeName);
}


//
//
// Delete the moveable
//
//
void    CHandleMoveable::vDelete()
{
    if(AfxMessageBox(_T("Are you sure you want to delete this moveable ?"), MB_YESNO | MB_ICONEXCLAMATION | MB_TOPMOST) != IDYES) return;

    CTrMvb* lpMoveable = GetMoveable();
    DWORD   dwObjId = lpMoveable->dwGetObjectId();

    CTRFile*            lpTRFile = m_lpTree->lpGetTRFile();
    TR123_FILE_STRUCT&  trFile = lpTRFile->GetTRFile().tr123File;
    bool                boTR1 = lpTRFile->GetTRType() == TR1_FILE;

    trFile.lpSndMgr->vMerge(lpMoveable->lpGetSndMgr(),true);

    TRListMoveable* lpList = lpTRFile->GetTRFile().tr123File.lpListMoveable;

    for(TRListMoveable::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
        {
        if(*theIterator == lpMoveable)
            {
            // Remove from the items list
            TRListItem* lpListItem = trFile.lpListItem;

            for(TRListItem::iterator theIterator4 = lpListItem->begin(); theIterator4 != lpListItem->end(); theIterator4++)
                {
                TR123_ITEM* lpItem= *theIterator4;

                if(boTR1)
                    if(lpItem->tr1.sObjectId == dwObjId) { lpListItem->erase(theIterator4); delete lpItem; }
                else
                    if(lpItem->tr2.sObjectId == dwObjId) { lpListItem->erase(theIterator4); delete lpItem; }
                }

            //
            lpList->erase(theIterator);

			CEditBase*  lpEdit = lpGetEditView();

			if( lpEdit != NULL ) dynamic_cast< CEditMoveable* >( lpEdit )->vRelease();

            delete lpMoveable;

            vSetChanged();
            m_lpTree->m_lpHandleMoveables->vExpandTree(true);

            break;
            }
        }

    lpTRFile->vMakeSureSoundsAreAvailable();
}
