#include "stdafx.h"

#include "Resource.h"
#include "TrViewer.h"
#include "ChildFrm.h"
#include "FileTree.h"
#include "HandleBase.h"
#include "HandleEmpty.h"
#include "HandleMoveables.h"
#include "HandleMoveable.h"


void    CHandleMoveables::vExpandTree(bool boAlreadyExpanded)
{
    if(boAlreadyExpanded)
        {
        m_lpTree->m_boDeleting = true;
        m_lpTree->vReleaseNode(m_hParentItem,true,true);
        m_lpTree->m_boDeleting = false;
        }

    CString             s;
    HTREEITEM           h,hItem;
    CTRFile*            lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE             trFile      = lpTRFile->GetTRFile();
    enumTRType          trType      = lpTRFile->GetTRType();
	CTRViewerApp*		lpApp		= (CTRViewerApp*)AfxGetApp();

    switch(trType)
        {
        //
        case TR1_FILE:
        case TR2_FILE:
        case TR3_FILE:
        case TR4_FILE:
            {
            TRListMoveable* lpList = trFile.tr123File.lpListMoveable;

            for(TRListMoveable::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
                {
				CTrMvb*	lpMvb = *theIterator;
				CResMoveable*	lpMoveableInfo = lpMvb->lpGetResMgr();

				s = lpMoveableInfo->strGetName().c_str();

				lpMvb->vEdtSetBodyMvbId( lpMoveableInfo->dwGetBodyMvbId() );
				lpMvb->vEdtSetAssocMvbId( lpMoveableInfo->dwGetShowWithId() );
				lpMvb->vEdtSetShowRealMeshes( lpApp->boShowRealMeshes() );
				lpMvb->vEdtSetShowAssocMvb( lpApp->boShowAssociatedMoveable() );

                INSERT(m_lpTree,hItem,m_hParentItem,new CHandleMoveable(m_lpTree,m_lpChildFrm,*theIterator))

				m_lpTree->SetItemImage( hItem, lpMoveableInfo->dwGetIconId(), lpMoveableInfo->dwGetIconId() );

                INSERT(m_lpTree,h,hItem,new CHandleEmpty(m_lpTree,m_lpChildFrm))
                }

            break;
            }
        }
}


CMenu*  CHandleMoveables::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    m_lpMenu->LoadMenu(IDR_MENU_MOVEABLES);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleMoveables::vReleasePopupupMenu()
{
    delete m_lpMenu;
    
    m_lpMenu = NULL;
}


CHandleMoveable* CHandleMoveables::lpFindMoveable(DWORD dwObjId)
{
    HTREEITEM hNextItem;
    HTREEITEM hChildItem = m_lpTree->GetChildItem(m_hParentItem);

    while(hChildItem != NULL)
        {
        hNextItem = m_lpTree->GetNextItem(hChildItem, TVGN_NEXT);

        CHandleMoveable* lpNode = (CHandleMoveable*) m_lpTree->GetItemData(hChildItem);

        if(lpNode->GetMoveable()->dwGetObjectId() == dwObjId) return lpNode;

        hChildItem = hNextItem;
        }

    return NULL;
}


void    CHandleMoveables::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Load a 3ds file and add a new moveable
        case ID_MOVEABLES_IMPORTFROM3DS:
            {
            vImportFrom3DS();
            break;
            }
        // Add a moveable from a .TRMVB file
        case ID_MOVEABLES_IMPORTFROMTRMVB:
            {
            vImportFromMVB();
            break;
            }
        // Export all meshes to .MVB files
        case ID_MOVEABLES_EXPORTASMVB:
            {
            vExportAsMVB();
            break;
            }
        }
}


void    CHandleMoveables::vExportAsMVB()
{
    BROWSEINFO   bi;
    ZeroMemory(&bi,sizeof(BROWSEINFO));

    TCHAR szBuf[MAX_PATH];
    lstrcpy(szBuf, __TEXT("Choose a folder to save the TRMVB in:"));

    bi.hwndOwner  = NULL;
    bi.ulFlags    = 0;
    bi.lpszTitle  = static_cast<LPCSTR>(szBuf);
    bi.lpfn       = NULL;
    bi.lParam     = (long)this;

    LPITEMIDLIST  pidl = SHBrowseForFolder(&bi);

    if(!pidl) return;

    SHGetPathFromIDList(pidl,szBuf);

    CWaitCursor dummy;

    CTRFile*    lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE     trFile      = lpTRFile->GetTRFile();

    CString     s;
    TCHAR       szLevelName[20+1];
    DWORD       dwI = 0;
    
    lstrcpyn(szLevelName,lpTRFile->lpszGetShortLevelName(),20+1);

    TRListMoveable* lpList = trFile.tr123File.lpListMoveable;
    for(TRListMoveable::iterator theIterator = lpList->begin(); theIterator != lpList->end(); theIterator++)
        {
        CTrMvb*  lpMvb = *theIterator;

        DWORD   dwMVBSize;
        BYTE*   lpMVB = lpMvb->lpbGetData(dwMVBSize);

        s.Format(_T("%s\\%s_MVB%.3d.trmvb"),szBuf,szLevelName,dwI++);

        DWORD   dwNumWritten;
        HANDLE hFile = CreateFile(s,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);
    
        WriteFile(hFile,lpMVB,dwMVBSize,&dwNumWritten,NULL);
    
        CloseHandle(hFile);

        delete [] lpMVB;
        }
}


//
//
// Import a moveable from a .3DS
//
//
void    CHandleMoveables::vImportFrom3DS()
{
    // Load the 3DS file
    CString strFilter(_T("3DS files (*.3ds)|*.3ds||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    // NOT IMPLEMENTED
    CTrMvb* lpNewMvb = new CTrMvb;

    if(CTRFile::boImportMoveableFrom3DS(lpNewMvb,(TCHAR*)(LPCTSTR)FileDialog.GetPathName(),false,lpTRFile->GetTRType() == TR4_FILE))
        {
        lpTRFile->GetTRFile().tr123File.lpListMoveable->push_back(lpNewMvb);

        vSetChanged();

        if(m_boExpanded)
            {
            vExpandTree(true);
            }
        }
}


//
//
// Import a moveable from a .TRMVB
//
//
void    CHandleMoveables::vImportFromMVB(CTrMvb* lpMvbSrc)
{
    CTrMvb* lpMvb = new CTrMvb;

    // Load the MVB file
    if(lpMvbSrc == NULL)
        {
        CString strFilter = _T("TR Moveable files (*.trmvb)|*.trmvb||");

        CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

        if(FileDialog.DoModal() != IDOK)
            {
            delete lpMvb;
            return;
            }

        CWaitCursor dummy;

        CTrMvb* Mvb = CTrMvb::objLoad(FileDialog.GetPathName());

        if(Mvb)
            {
            lpMvb->vRelease();
            lpMvb->vStealData(*Mvb);
            delete Mvb;
            }
        else
            {
            delete lpMvb;
            AfxMessageBox(_T("Could not import this moveable because of a version mismatch:\n\n - the moveable was produced with an old TRViewer release and its format is not supported anymore !"),MB_ICONSTOP | MB_TOPMOST);
            return;
            }
        }
    else
        {
        CWaitCursor dummy;

        lpMvb->vRelease();
        lpMvb->vStealData(*lpMvbSrc);
        }

    CTRFile* lpTRFile = m_lpTree->lpGetTRFile();

    lpTRFile->GetTRFile().tr123File.lpSndMgr->vMerge(lpMvb->lpGetSndMgr());

    lpTRFile->GetTRFile().tr123File.lpListMoveable->push_back(lpMvb);

    vSetChanged();

    if(m_boExpanded)
        {
        vExpandTree(true);
        }
}
