#include "stdafx.h"

#include "TRViewer.h"
#include "ChildFrm.h"
#include "FileTree.h"
#include "HandleBase.h"
#include "HandleRootTR123.h"
#include "HandleEmpty.h"
#include "HandleStaticMeshes.h"
#include "HandleMesh.h"
#include "HandleMoveables.h"
#include "HandleMoveable.h"
#include "HandleSprites.h"
#include "HandleSprite.h"
#include "HandleSounds.h"
#include "HandleTile.h"


void    CHandleRootTR123::vExpandTree(bool boAlreadyExpanded)
{
    if(boAlreadyExpanded) return;

    CString             s;
    HTREEITEM           hItem, h;
    CTRFile*            lpTRFile    = m_lpTree->lpGetTRFile();
    TR_FILE&            trFile      = lpTRFile->GetTRFile();
    enumTRType          trType      = lpTRFile->GetTRType();
    const CResMgr&		ResMgr		= ((CTRViewerApp*)AfxGetApp())->GetResMgr();
	const char*			lpszLevName	= lpTRFile->lpszGetShortLevelName();
	BYTE				bTRResId	= ResMgr.bGetResId( trType );

	CResCategory	MoveableInfo = ResMgr.GetCategoryInfo( bTRResId, lpszLevName, "moveables" );

    m_lpTree->m_lpHandleMoveables = new CHandleMoveables(m_lpTree,m_lpChildFrm);

    s.Format(MoveableInfo.strGetName().c_str());
    INSERT(m_lpTree,hItem,m_hParentItem,m_lpTree->m_lpHandleMoveables)
	m_lpTree->SetItemImage( hItem, MoveableInfo.dwGetIconId(), MoveableInfo.dwGetIconId() );
    INSERT(m_lpTree,h,hItem,new CHandleEmpty(m_lpTree,m_lpChildFrm))  // Child beneath
    m_lpTree->m_lpHandleMoveables->vExpandTree_(hItem);

	CResCategory	StaticMeshesInfo = ResMgr.GetCategoryInfo( bTRResId, lpszLevName, "static_meshes" );

    m_lpTree->m_lpHandleStaticMeshes = new CHandleStaticMeshes(m_lpTree,m_lpChildFrm);

    s.Format(StaticMeshesInfo.strGetName().c_str());
    INSERT(m_lpTree,hItem,m_hParentItem,m_lpTree->m_lpHandleStaticMeshes)
	m_lpTree->SetItemImage( hItem, StaticMeshesInfo.dwGetIconId(), StaticMeshesInfo.dwGetIconId() );
    INSERT(m_lpTree,h,hItem,new CHandleEmpty(m_lpTree,m_lpChildFrm))  // Child beneath
    m_lpTree->m_lpHandleStaticMeshes->vExpandTree_(hItem);

	CResCategory	SpritesInfo = ResMgr.GetCategoryInfo( bTRResId, lpszLevName, "sprites" );

    m_lpTree->m_lpHandleSprites = new CHandleSprites(m_lpTree,m_lpChildFrm);

    s.Format(SpritesInfo.strGetName().c_str());
    INSERT(m_lpTree,hItem,m_hParentItem,m_lpTree->m_lpHandleSprites)
	m_lpTree->SetItemImage( hItem, SpritesInfo.dwGetIconId(), SpritesInfo.dwGetIconId() );
    INSERT(m_lpTree,h,hItem,new CHandleEmpty(m_lpTree,m_lpChildFrm))  // Child beneath
    m_lpTree->m_lpHandleSprites->vExpandTree_(hItem);

	if( trType >= TR4_FILE )
		{
		s.Format(_T("Font & sky"));
		INSERT(m_lpTree,hItem,m_hParentItem,new CHandleTile(m_lpTree,m_lpChildFrm,&trFile.tr123File.Texture3,TILE_256xY_32))
		m_lpTree->SetItemImage( hItem, ResMgr.dwGetIconId( "Sky" ), ResMgr.dwGetIconId( "Sky" ) );
		}

	CResCategory	SoundsInfo = ResMgr.GetCategoryInfo( bTRResId, lpszLevName, "sounds" );

    s.Format(SoundsInfo.strGetName().c_str());
    INSERT(m_lpTree,hItem,m_hParentItem,new CHandleSounds(m_lpTree,m_lpChildFrm,NULL))
	m_lpTree->SetItemImage( hItem, SoundsInfo.dwGetIconId(), SoundsInfo.dwGetIconId() );
}


CMenu*  CHandleRootTR123::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    const CResMgr& ResMgr = ((CTRViewerApp*)AfxGetApp())->GetResMgr();

    m_lpMenu = const_cast<CMenu*>(ResMgr.lpGetMenu(ResMgr.bGetResId(m_lpTree->lpGetTRFile()->GetTRType())));

    return m_lpMenu;
}


void    CHandleRootTR123::vReleasePopupupMenu()
{
    m_lpMenu = NULL;
}


void    CHandleRootTR123::vHandleMenu(WORD wMenuId)
{
    const CResMgr& ResMgr = ((CTRViewerApp*)AfxGetApp())->GetResMgr();

    BYTE        bResId = ResMgr.bGetResId(m_lpTree->lpGetTRFile()->GetTRType());

    const CResItem* lpResItem = ResMgr.lpGetResItem(bResId,wMenuId);

    if(!lpResItem) return;

    CTRFile File;

    CString cstrLevelPath(ResMgr.cstrGetDataPath(bResId));

    cstrLevelPath += lpResItem->lpszGetLevelName();

    if(!File.boLoad((LPSTR)(LPCTSTR)cstrLevelPath))
        {
        CString s;
        s.Format(_T("Can't load the level %s to get the data"),cstrLevelPath);
        AfxMessageBox(s,MB_ICONSTOP | MB_TOPMOST);
        return;
        }

    CString cstrErr(_T(""));

    for(DWORD dwI = 0; dwI < lpResItem->dwGetNumObjects(); dwI++)
        {
        const CResObject& ResObject = lpResItem->GetResObject(dwI);

        const enumTrObjType ObjType = ResObject.GetObjType();
        const DWORD         dwObjId = ResObject.dwGetObjId();

        switch(ObjType)
            {
            case TrObj_Type_Moveable:
                {
                CTrMvb* lpMvb = File.lpGetMoveable(dwObjId);
                if(lpMvb)
                    {
                    CTrMvb* lpMvbDst = m_lpTree->lpGetTRFile()->lpGetMoveable(dwObjId);
                    if(!lpMvbDst) m_lpTree->m_lpHandleMoveables->vImportFromMVB(lpMvb);
                    else
                        {
                        m_lpTree->m_lpHandleMoveables->lpFindMoveable(dwObjId)->vImportFromMVB(lpMvb);
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Could not find moveable #%ld in level %s\n"),dwObjId,lpResItem->lpszGetLevelName());
                    cstrErr += s;
                    }
                break;
                }
            case TrObj_Type_Sprite:
                {
                CTrLstSpr* lpSpr = File.lpGetSprite(dwObjId);
                if(lpSpr)
                    {
                    CTrLstSpr* lpSprDst = m_lpTree->lpGetTRFile()->lpGetSprite(dwObjId);
                    if(!lpSprDst) m_lpTree->m_lpHandleSprites->vImportFromSPR(lpSpr);
                    else
                        {
                        m_lpTree->m_lpHandleSprites->lpFindSprite(dwObjId)->vImportFromSPR(lpSpr);
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Could not find sprite #%ld in level %s\n"),dwObjId,lpResItem->lpszGetLevelName());
                    cstrErr += s;
                    }
                break;
                }
            case TrObj_Type_StaticMesh:
                {
                CTrStatMsh* lpMsh = File.lpGetStaticMesh(dwObjId);
                if(lpMsh)
                    {
                    CTrStatMsh* lpMshDst = m_lpTree->lpGetTRFile()->lpGetStaticMesh(dwObjId);
                    if(!lpMshDst) m_lpTree->m_lpHandleStaticMeshes->vImportFromMSH(lpMsh);
                    else
                        {
                        m_lpTree->m_lpHandleStaticMeshes->lpFindStaticMesh(dwObjId)->vImportFromMSH(lpMsh);
                        }
                    }
                else
                    {
                    CString s;
                    s.Format(_T("Could not find static mesh #%ld in level %s\n"),dwObjId,lpResItem->lpszGetLevelName());
                    cstrErr += s;
                    }
                break;
                }
            }
        }

    if(cstrErr == _T(""))
        {
        AfxMessageBox("Resource(s) successfully added !",MB_OK  | MB_TOPMOST);
        }
    else
        AfxMessageBox(cstrErr,MB_ICONSTOP | MB_TOPMOST);
}
