#include "stdafx.h"

#include "Resource.h"
#include "ChildFrm.h"
#include "HandleBase.h"
#include "HandleTile.h"
#include "EditBase.h"
#include "EditTile.h"

#include "TextureTools.h"
#include "TGAFile.h"


void    CHandleTile::vEditSetInfo(TCHAR* lpszNodeName)
{
    CEditTile*  lpEdit = dynamic_cast< CEditTile* >( lpGetEditView() );

    if( lpEdit != NULL ) lpEdit->vSetDataTypeHandler(this);
}


CMenu*  CHandleTile::lpGetPopupupMenu(HTREEITEM hSelectedItem)
{
    m_lpMenu = new CMenu;
    m_lpMenu->LoadMenu(IDR_MENU_TILE);

    return m_lpMenu->GetSubMenu(0);
}


void    CHandleTile::vReleasePopupupMenu()
{
    if(m_lpMenu) delete m_lpMenu;
    m_lpMenu = NULL;
}


void    CHandleTile::vHandleMenu(WORD wMenuId)
{
    switch(wMenuId)
        {
        // Export the texture to a .TGA file
        case ID_TILE_EXPORT:
            {
            vExportToTGA();
            break;
            }
        // Load a gfx file and replace the current tile
        case ID_TILE_REPLACE:
            {
            vReplaceTile();
            break;
            }
        }
}


void    CHandleTile::vExportToTGA()
{
    // Get the filename
    CString strFilter(_T("TGA files (*.tga)|*.tga||"));

	CString strName( m_szNodeName ); strName.Replace("/","-");

    CFileDialog FileDialog(FALSE,_T("tga"),strName,OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    // Save the tile
    switch(m_TileType)
        {
        case TILE_256x256_16:
            {
            BYTE* lpDataDest = new BYTE[256 * 256 * 4];
            vPromoteTo32Bits((BYTE*)GetTile(),256,256,16,NULL,lpDataDest);
            vMakeTGA32Bits((LPCTSTR)FileDialog.GetPathName(),lpDataDest,256,256,4,false);
            delete [] lpDataDest;
            break;
            }
        case TILE_256xY_32:
            {
            DWORD   dwHeight = 0;
            BYTE*   lpabDataUncomp = NULL;
            vUncompressTexture((TR4_COMPR_TEXTURE*)GetTile(),true,&lpabDataUncomp,&dwHeight);
            BYTE*   lpDataDest = new BYTE[256 * dwHeight * 4];
            vPromoteTo32Bits(lpabDataUncomp,256,dwHeight,32,NULL,lpDataDest);
            vMakeTGA32Bits((LPCTSTR)FileDialog.GetPathName(),lpDataDest,256,dwHeight,4,false);
            delete [] lpabDataUncomp;
            delete [] lpDataDest;
            break;
            }
        case TILE_256xY_16:
            {
            DWORD   dwHeight = 0;
            BYTE*   lpabDataUncomp = NULL;
            vUncompressTexture((TR4_COMPR_TEXTURE*)GetTile(),false,&lpabDataUncomp,&dwHeight);
            BYTE*   lpDataDest = new BYTE[256 * dwHeight * 4];
            vPromoteTo32Bits(lpabDataUncomp,256,dwHeight,16,NULL,lpDataDest);
            vMakeTGA32Bits((LPCTSTR)FileDialog.GetPathName(),lpDataDest,256,dwHeight,4,false);
            delete [] lpabDataUncomp;
            delete [] lpDataDest;
            break;
            }
        }
}


void    CHandleTile::vReplaceTile()
{
    BYTE* lpData = NULL;

    // Load the TGA file
    CString strFilter(_T("Gfx files (*.tga;*.bmp;*.gif)|*.tga;*.bmp;*.gif|Others (*.fli;*.flc;*.cel)|*.fli;*.flc;*.cel||"));
    CFileDialog FileDialog(TRUE,_T(""),_T(""),OFN_HIDEREADONLY,strFilter);

    if(FileDialog.DoModal() != IDOK) return;

    CWaitCursor dummy;

    DWORD   dwWidth,dwHeight;
    DWORD   dwRes = dwLoadImage((LPCTSTR)FileDialog.GetPathName(),&dwWidth,&dwHeight,&lpData);

    if(dwRes)
        {
        if(lpData) delete [] lpData;
        CString s;
        if(dwRes == -1)
            s.Format(_T("Can't instantiate the ReadPic component: try to register it typing \"regsvr32 ReadPic.dll\" in the folder where ReadPic.dll can be found."));
        else
            s.Format(_T("Error %ld loading the file (maybe an unknown format)."),dwRes);
        AfxMessageBox(s,MB_OK | MB_ICONSTOP | MB_TOPMOST);
        return;
        }

    // Check the image size
    if((m_TileType == TILE_256xY_32  ||  m_TileType == TILE_256xY_16)  &&  (dwWidth != 256))
        {
        if(lpData) delete [] lpData;
        AfxMessageBox(_T("Invalid width. It must be 256 pixels."),MB_OK | MB_ICONSTOP | MB_TOPMOST);
        return;
        }

    if(m_TileType == TILE_256x256_16)
        {
        BYTE* lpbSaveData = lpData;

        lpData = lpbEnlarge32BitsBitmap(lpData,dwWidth,dwHeight,256,256);

        delete [] lpbSaveData;
        }

    // Replace the node with the new tile
    switch(m_TileType)
        {
        case TILE_256x256_16:
            {
            vConvertABGRToARGB(256,256,false,lpData,(BYTE*)GetTile());
            vEditSetInfo(m_szNodeName);
            break;
            }
        case TILE_256xY_16:
            {
            vConvertABGRToARGB(256,dwHeight,false,lpData,lpData);
            TR4_COMPR_TEXTURE* lpText = (TR4_COMPR_TEXTURE*)GetTile();
            delete [] lpText->lpbTextData;
            vCompressTexture(lpData,dwWidth * dwHeight * 2,lpText);
            vEditSetInfo(m_szNodeName);
            break;
            }
        case TILE_256xY_32:
            {
            vConvertABGRToARGB(256,dwHeight,true,lpData,lpData);
            TR4_COMPR_TEXTURE* lpText = (TR4_COMPR_TEXTURE*)GetTile();
            delete [] lpText->lpbTextData;
            vCompressTexture(lpData,dwWidth * dwHeight * 4,lpText);
            vEditSetInfo(m_szNodeName);
            break;
            }
        }

    if(lpData)
		{
        vSetChanged();
		delete [] lpData;
		}
}


void    CHandleTile::vCopyTo(CHandleBase* lpDst,bool boCopyInSameLevel)
{
    if(AfxMessageBox(_T("Copy the tile ?"),MB_YESNO | MB_ICONINFORMATION | MB_TOPMOST) != IDYES) return;

    CHandleTile* lpDest = static_cast<CHandleTile*>(lpDst);

    if(GetTileType()  !=  lpDest->GetTileType())
        {
        AfxMessageBox(_T("Invalid operation !"),MB_OK | MB_TOPMOST | MB_ICONSTOP);
        return;
        }

    void* lpDataSrc = GetTile();
    void* lpDataDst = lpDest->GetTile();

    switch(GetTileType())
        {
        case TILE_256x256_16:   *((TR2_TEXTILE16*)lpDataDst) = *((TR2_TEXTILE16*)lpDataSrc); break;
        case TILE_256xY_16:
        case TILE_256xY_32:
            {
            delete [] ((TR4_COMPR_TEXTURE*)lpDataDst)->lpbTextData;
            *((TR4_COMPR_TEXTURE*) lpDataDst) = *((TR4_COMPR_TEXTURE*) lpDataSrc);
            ((TR4_COMPR_TEXTURE*)lpDataDst)->lpbTextData = new BYTE [((TR4_COMPR_TEXTURE*)lpDataDst)->dwTextCompSize];
            memcpy(((TR4_COMPR_TEXTURE*)lpDataDst)->lpbTextData,((TR4_COMPR_TEXTURE*)lpDataSrc)->lpbTextData,((TR4_COMPR_TEXTURE*)lpDataDst)->dwTextCompSize);
            break;
            }
        }

    lpDest->vSetChanged();
}
