#if !defined(AFX_OPENGLWND1_H__D75C4CE2_F1E2_11D3_AA63_00105A49000B__INCLUDED_)
#define AFX_OPENGLWND1_H__D75C4CE2_F1E2_11D3_AA63_00105A49000B__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000
// OpenGLWnd1.h : header file
//

/////////////////////////////////////////////////////////////////////////////
// COpenGLWnd window

#include <GL/gl.h>     // OpenGL interface
#include <GL/glu.h>    // OpenGL utility Library interface

#include "Camera.h"

class CToolboxMoveable;


typedef struct sGfxInfo
{
	GLuint*			LstTextId;
	DWORD			dwNumTextures;
	enumTrObjType   trObjType;
	CTrObjGfx*      lpTrGfx;
	bool			boShow;

	sGfxInfo() : LstTextId( NULL ), dwNumTextures( 0 ), trObjType( TrObj_Type_Mesh ), lpTrGfx( NULL ), boShow( false ) { }
	sGfxInfo( enumTrObjType trObjType_, CTrObjGfx* lpTrGfx_, bool boShow_ ) : 
		LstTextId( NULL ), dwNumTextures( 0 ), trObjType( trObjType_ ), lpTrGfx( lpTrGfx_ ), boShow( boShow_ ) { }
} sGfxInfo;

typedef struct sGfxDisplay
{
	bool		boShowDataInToolbox;
	sGfxInfo	AnimatedGfx;
	sGfxInfo	DisplayedGfx;

	sGfxDisplay() : AnimatedGfx(), DisplayedGfx(), boShowDataInToolbox( false ) { }

	sGfxDisplay( bool boShowDataInToolbox_, enumTrObjType trObjType, CTrObjGfx* lpTrGfxMain, CTrObjGfx* lpTrGfxSecond = NULL, bool boShowFirst = true, bool boShowSecond = false ) :
		boShowDataInToolbox( boShowDataInToolbox_ ), AnimatedGfx( trObjType, lpTrGfxMain, boShowFirst ), DisplayedGfx( trObjType, lpTrGfxSecond, boShowSecond ) { }
} sGfxDisplay;


class COpenGLWnd : public CWnd
{
// Construction
public:
	COpenGLWnd();

// Attributes
public:

// Operations
public:

// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(COpenGLWnd)
	//}}AFX_VIRTUAL

// Implementation
public:
	virtual ~COpenGLWnd();
    void    vSetGfxAddr(sGfxDisplay* lpLstGfx, DWORD dwNumGfx,bool boDontReleaseTextures = false,DWORD dwNumMeshToRender = 0);
    void    vSetSprSeqAddr(sGfxDisplay* lpLstGfx, DWORD dwNumGfx,bool boDontReleaseTextures = false);
	void	vSetGfxVisiblity( DWORD dwNumGfx, bool boAnimatedVisible, bool boDisplayedVisible );
    void    vSetMoveableAnimation(DWORD dwNumAnim);
    void    vActivateWnd();
    void    vCheckKeys();
    void    vSetAutomaticCamera(bool boOk);
    void    vSetShowAxis(bool boOk);
    void    vSetBkgColor(COLORREF BkgColor);
    void    vSetShowCollSphere(bool boOk) { m_boShowCollSphere = boOk; InvalidateRect(NULL); }
    void    vSetShowVertices(bool boOk) { m_boShowVertices = boOk; InvalidateRect(NULL); }
    void    vSetShowStaticBB1(bool boOk) { m_boShowStaticBB1 = boOk; InvalidateRect(NULL); }
    void    vSetStaticBB1IsSel(bool boOk) { m_boStaticBB1IsSel = boOk; InvalidateRect(NULL); }
    void    vSetShowStaticBB2(bool boOk) { m_boShowStaticBB2 = boOk; InvalidateRect(NULL); }
    void    vSetStaticBB2IsSel(bool boOk) { m_boStaticBB2IsSel = boOk; InvalidateRect(NULL); }
    void    vSetShowMoveableBB(bool boOk) { m_boShowMoveableBB = boOk; InvalidateRect(NULL); }
    void    vReleaseTextures();
    void    vPick(DWORD dwPosX,DWORD dwPosY);
    FLOAT   fGetFrameRate() { return m_fFrameRate; }
    void    vModifyIntensOfSelectedVertices(bit16 sStep);
    void    vSetAnimationType(bool boChainAnimations,bool boResetMoveable);
    void    vSetAutomaticAnim(bool boOk) { m_boAutomaticAnim = boOk; }
    void    vSetAnimOverallSpeed(FLOAT fSpeed);
    void    vSetAnimCurFrame(DWORD dwNumFrame);
    void    vSetAnimCurKey(DWORD dwNumKey);
    void    vSetAnimInterp(bool boOk);
    void    vChangeAnimInterp(int iDir);
    void    vSetAnimInterp(FLOAT fInterp);
    void    vSetSprSeqOverallSpeed(FLOAT fSpeed) { m_fSprSeqSpeed = fSpeed; }
    void    vSetSprSeqShowBorder(bool boShow) {  m_boSprSeqShowBorder = boShow; }
    void    vSetSprSeqTypeAnimated(bool boAnimated) {  m_boSprSeqTypeAnimated = boAnimated; }
    void    vSetSprSeqListWidth(FLOAT fListWidth) {  m_fSprSeqListWidth = fListWidth; }

	void    OnLButtonDown(UINT nFlags, CPoint point);
	void    OnLButtonUp(UINT nFlags, CPoint point);
	void    OnRButtonDown(UINT nFlags, CPoint point);
	void    OnRButtonUp(UINT nFlags, CPoint point);
	void    OnLButtonDblClk(UINT nFlags, CPoint point);
	void    OnMouseWheel(UINT nFlags, short zDelta, CPoint pt);
	void    OnMouseMove(UINT nFlags, CPoint point);

	void	vZoom(FLOAT fValue);
	void	vSetTextToDisplay(const char* pszText);

protected:
	// Generated message map functions
	//{{AFX_MSG(COpenGLWnd)
	afx_msg void OnPaint();
	afx_msg void OnSize(UINT nType, int cx, int cy);
    afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

private:

	typedef std::vector<void*>  ListFacePtr;

    void    vChangeSize(int cx, int cy);
	void	vRenderAndAnimate( sGfxDisplay& Gfx );
    void    vDraw(HDC hDC);
    void    vGenerateTextures( sGfxDisplay& Gfx );
	void	vGenerateTextures_( sGfxInfo& Gfx );
    void    vRenderMesh( sGfxInfo& Gfx, DWORD dwNumMesh,bool boShowCollisionSphere = false,bool boWireframe = false,bool boShowVertices = false,bool boDontShowDummies = false);
    void    vRenderMoveable( sGfxDisplay& Gfx );
    void    vRenderSprite( sGfxInfo& Gfx, const WORLD_POSITIONNING* lpWorldP,const TR2_OBJECT_TEXTURE* lpText,bool boAlign = false );
    void    vRenderSprSeq( sGfxInfo& Gfx );
    void    vInitiateCamera(FLOAT fMinX,FLOAT fMaxX,FLOAT fMinY,FLOAT fMaxY,FLOAT fMinZ,FLOAT fMaxZ, bool boCenterCamera = false);
    void    vRenderMeshVertices(const TR123_MESH* lpMesh);

	void    vDrawBoundingBox(CBoundingBox& BB, CVect3D vecColor = CVect3D((FLOAT)1.0,(FLOAT)1.0,(FLOAT)1.0));
	void	vRenderFace( sGfxInfo& Gfx, TR2_VERTEX* lpVertices,const TR2_OBJECT_TEXTURE* lpOText,bitu16* lpFace,WORD wTextureIndex,bit16* lpLstLighting,bool boInverseIntens,bool boQuad,bool boWireframe);
	void	vRenderFaceArray( sGfxInfo& Gfx, DWORD dwNumFaces,const TR123_MESH* lpMesh,const TR2_OBJECT_TEXTURE* lpOText,void* lpArray,bool boInversIntens,bool boQuad,bool boWireframe,ListFacePtr& ChromaFaces, ListFacePtr& AlphaFaces);
	void	vRenderFaceList( sGfxInfo& Gfx, const TR123_MESH* lpMesh,const TR2_OBJECT_TEXTURE* lpOText,ListFacePtr& ListFaces,bool boInversIntens,bool boQuad,bool boWireframe);
	void	vRenderText();

	void	vSetProjMode();
	void	vSetOrthoMode();

	void	vReleaseTextures_( sGfxInfo& Gfx );

    HGLRC               m_hRC;
    HDC                 m_hDC;
    bool                m_boMakeCurrentOk;
    GLdouble            m_gldAspect;
    bool                m_boShowAxis;
    COLORREF            m_BkgColor;

	long				m_lOldTextureBinded;
	DWORD				m_dwNumGfx;
	sGfxDisplay			m_Gfx[ 2 ];

    bool                m_boShowCollSphere;
    bool                m_boShowVertices;
    bool                m_boShowStaticBB1;
    bool                m_boStaticBB1IsSel;
    bool                m_boShowStaticBB2;
    bool                m_boStaticBB2IsSel;
    DWORD               m_dwNumMeshToRender;
    FLOAT               m_fMax;
	bool				m_boShowMoveableBB;

    DWORD               m_dwOldTime;
    LARGE_INTEGER       m_OldTime;
    FLOAT               m_fFrameRate;

    CAMERA              m_Camera;
    bool                m_boAutomaticCamera;

    bool                m_boPickingMode;
    DWORD               m_dwShowNumVertex;

    bool                m_boAutomaticAnim;
    bool                m_boAnimShow;
    bool                m_boChainAnimations;
    bool                m_boAnimInterp;
    bool                m_boRecalcInterp;
    DWORD               m_dwAnimCurAnim;
    FLOAT               m_fAnimOverallSpeed;

    FLOAT               m_fSprSeqSpeed;
    bool                m_boSprSeqShowBorder;
    bool                m_boSprSeqTypeAnimated;
    FLOAT               m_fSprSeqListWidth;

    bool                m_boRotWithMouse;
    CPoint              m_RotMousePos;
    bool                m_boTransWithMouse;
    CPoint              m_TransMousePos;

	int					m_iWidth;
	int					m_iHeight;
	char				m_szTextToDisplay[256];
};

/////////////////////////////////////////////////////////////////////////////

//{{AFX_INSERT_LOCATION}}
// Microsoft Visual C++ will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_OPENGLWND1_H__D75C4CE2_F1E2_11D3_AA63_00105A49000B__INCLUDED_)
