#pragma once

#include "MarkupSTL.h"


class CResObject
{
public:

    CResObject(enumTrObjType ObjType, DWORD dwObjId) : m_ObjType(ObjType), m_dwObjId(dwObjId) { }

    const enumTrObjType GetObjType() const { return m_ObjType; }
    const DWORD         dwGetObjId() const { return m_dwObjId; }

private:

    enumTrObjType   m_ObjType;
    DWORD           m_dwObjId;
};


class CResLevel : public CResObject
{
public:

	CResLevel( const std::string& strShortName, const std::string& strLongName, bool boCutScene, DWORD dwIconId ) : 
		CResObject( TrObj_Type_Level, 0 ),
		m_strShortName( strShortName ), m_strLongName( strLongName ), m_boCutscene( boCutScene ), m_dwIconId( dwIconId )
	{ }

	const std::string&	strGetShortName()	const	{ return m_strShortName; }
	const std::string&	strGetLongName()	const	{ return m_strLongName; }
	bool				boIsCutScene()		const	{ return m_boCutscene; }
	DWORD				dwGetIconId()		const	{ return m_dwIconId; }

private:

	std::string		m_strShortName;
	std::string		m_strLongName;
	bool			m_boCutscene;
	DWORD			m_dwIconId;
};


class CResMoveable : public CResObject
{
public:

	typedef std::map< DWORD, std::string >	MapStateId2Descr;
	typedef MapStateId2Descr::iterator		MapStateId2DescrIter;

	CResMoveable( DWORD dwObjId, const std::string& strName, DWORD dwIconId, DWORD dwBodyMvbId, DWORD dwShowWithId, MapStateId2Descr* lpLstStates ) :
		CResObject( TrObj_Type_Moveable, dwObjId ),
		m_strName( strName ), m_dwIconId( dwIconId ), m_dwBodyMvbId( dwBodyMvbId ), m_dwShowWithId( dwShowWithId ), m_lpLstStates( lpLstStates )
	{ }

	~CResMoveable()
	{
		if( m_lpLstStates ) m_lpLstStates->clear();

		delete m_lpLstStates;
	}

	const std::string&		strGetName()		const	{ return m_strName; }
	DWORD					dwGetIconId()		const	{ return m_dwIconId; }
	DWORD					dwGetBodyMvbId()	const	{ return m_dwBodyMvbId; }
	DWORD					dwGetShowWithId()	const	{ return m_dwShowWithId; }
	MapStateId2Descr* const	pmapGetLstStates()	const	{ return m_lpLstStates; }

private:

	std::string			m_strName;
	DWORD				m_dwIconId;
	DWORD				m_dwBodyMvbId;
	DWORD				m_dwShowWithId;
	MapStateId2Descr*	m_lpLstStates;
};


class CResStaticMesh : public CResObject
{
public:

	CResStaticMesh( DWORD dwObjId, const std::string& strName, DWORD dwIconId ) :
		CResObject( TrObj_Type_StaticMesh, dwObjId ),
		m_strName( strName ), m_dwIconId( dwIconId )
	{ }

	const std::string&	strGetName()		const	{ return m_strName; }
	DWORD				dwGetIconId()		const	{ return m_dwIconId; }

private:

	std::string		m_strName;
	DWORD			m_dwIconId;
};


class CResSprite : public CResObject
{
public:

	CResSprite( DWORD dwObjId, const std::string& strName, DWORD dwIconId, bool boIsAnimated, float fZoom ) :
		CResObject( TrObj_Type_Sprite, dwObjId ),
		m_strName( strName ), m_dwIconId( dwIconId ), m_boIsAnimated( boIsAnimated ), m_fZoom( fZoom )
	{ }

	const std::string&	strGetName()		const	{ return m_strName; }
	DWORD				dwGetIconId()		const	{ return m_dwIconId; }
	bool				boIsAnimated()		const	{ return m_boIsAnimated; }
	float				fGetZoom()			const	{ return m_fZoom; }

private:

	std::string		m_strName;
	DWORD			m_dwIconId;
	bool			m_boIsAnimated;
	float			m_fZoom;
};


class CResCategory
{
public:

	CResCategory( const std::string& strName, DWORD dwIconId ) :
		m_strName( strName ), m_dwIconId( dwIconId )
	{ }

	const std::string&	strGetName()		const	{ return m_strName; }
	DWORD				dwGetIconId()		const	{ return m_dwIconId; }

private:

	std::string		m_strName;
	DWORD			m_dwIconId;
};


class CResItem
{
public:

    CResItem(TCHAR* lpszName, TCHAR* lpszLevelName);

    void    vAddObject(enumTrObjType ObjType, DWORD dwObjId);

    const DWORD dwGetNumObjects() const					{ return m_ListObjects.size(); }
    const CResObject& GetResObject(DWORD dwNum) const	{ return m_ListObjects[dwNum]; }
    const TCHAR* lpszGetLevelName() const				{ return m_szLevelName; }

private:

	typedef std::vector<CResObject> ListResObject;

    TCHAR           m_szName[128 + 1];
    TCHAR           m_szLevelName[128 + 1];
    ListResObject   m_ListObjects;
};


class CResSFX
{
public:

    CResSFX(TCHAR* lpszName, DWORD dwNumSamples, DWORD dwFirstSoundIndex) : m_cstrName(lpszName), m_dwNumSamples(dwNumSamples), m_dwFirstSoundIndex(dwFirstSoundIndex) { }

    CString     m_cstrName;
    DWORD       m_dwNumSamples;
    DWORD       m_dwFirstSoundIndex;
};


typedef std::vector<CResItem*>  ListResItem;
typedef std::vector<CResSFX*>   ListSfx;


#define     __RESMGR_TR1__          0
#define     __RESMGR_TR1GOLD__      1
#define     __RESMGR_TR2__          2
#define     __RESMGR_TR2GOLD__      3
#define     __RESMGR_TR3__          4
#define     __RESMGR_TR3GOLD__      5
#define     __RESMGR_TR4__			6
#define     __RESMGR_TRUNK__        255

#define     __RESMGR_MAXTR__        7

class CResMgr
{
public:

    CResMgr();
    ~CResMgr();
    void            vInitialize(TCHAR* lpszPath);
    void            vRelease();

    const CMenu*    lpGetMenu(BYTE bTRResId) const                      { return &m_Menu[bTRResId]; }
    const CResItem* lpGetResItem(BYTE bTRResId, DWORD dwNum) const      { return m_List[bTRResId][dwNum]; }

    const CString&  cstrGetDataPath(BYTE bTRResId) const                { return m_DataPath[bTRResId]; }
    void            vSetDataPath(BYTE bTRResId, CString& cstrDataPath);

    CString         cstrGetSfxName(BYTE bTRResId, DWORD dwInternalIndex) const;
    DWORD           dwGetSfxNumSamples(BYTE bTRResId, DWORD dwInternalIndex) const;
    DWORD           dwGetSfxFirstSoundIndex(BYTE bTRResId, DWORD dwInternalIndex) const;

    BYTE            bGetResId(enumTRType trType) const;

	const CBitmap*	lpGetIconsImage() const { return m_lpIconsImage; }

	CResLevel		GetLevelInfo( BYTE bTRResId, const std::string& strLevelShortName ) const;
	CResCategory	GetCategoryInfo( BYTE bTRResId, const std::string& strLevelShortName, const std::string& strCategoryName ) const;
	CResMoveable*	lpGetMoveableInfo( BYTE bTRResId, const std::string& strLevelShortName, DWORD dwObjId ) const;
	CResStaticMesh	GetStaticMeshInfo( BYTE bTRResId, const std::string& strLevelShortName, DWORD dwObjId ) const;
	CResSprite		GetSpriteInfo( BYTE bTRResId, const std::string& strLevelShortName, DWORD dwObjId ) const;

	DWORD			dwGetIconId( const std::string& strIconName ) const;

private:

	typedef struct
	{
		void*	lpPosRoot;
		void*	lpPosLevels;
		void*	lpPosMoveables;
		void*	lpPosSprites;
		void*	lpPosSounds;
	} sXMLTr;

	CString	cstrGetTRAbrevFromResId( BYTE bTRResId );
	BYTE	bGetResIdFromTRAbrev( const CString& cstrTRAbrev );

    void    vParseFile_MenuResource(CMenu& Menu, TCHAR* lpszFilename, enumTRType trType, ListResItem& List);
    void    vParseFile_SFX(CString& cstrFilename, ListSfx& ListSFX);

	void	vCreateIconsImage();

	bool		boFindXMLNode( BYTE bTRResId, const std::string& strLevelShortName, const std::string& strCategName, const std::string& strNodeName, const std::string& strIdToFind ) const;
	bool		boFindXMLNode( const std::string& strNodeName, const std::string& strIdToFind ) const;

    ListResItem             m_List[__RESMGR_MAXTR__];
    CMenu                   m_Menu[__RESMGR_MAXTR__];

    ListSfx                 m_ListSFX[__RESMGR_MAXTR__];

    CString                 m_DataPath[__RESMGR_MAXTR__];

    TCHAR                   m_szPath[MAX_PATH + 1];

	CBitmap*				m_lpIconsImage;

	mutable CMarkupSTL		m_XMLRes;

	void*					m_XMLIconsPos;
	sXMLTr					m_XMLTr[__RESMGR_MAXTR__];
};

