#include "stdafx.h"

#include "Split.h"


//----------------------------------------------------------------------------
//
//  Function:   dwNbOccurrence
//
//  Synopsis:   Count the occurrences of a string in another
//
//  Arguments:  [lpsz1] -- string to parse
//              [lpsz2] -- string to look for
//
//  Return:     num occurrences of lpsz2 in lpsz1
//
//----------------------------------------------------------------------------
DWORD   dwNbOccurrence(char* lpsz1,char* lpsz2)
{
    if(!lpsz1  ||  !lpsz2  ||  !*lpsz2) return 0;

    DWORD dwLen2  = strlen(lpsz2);
    DWORD dwCount = 0;

    do
        {
        lpsz1 = strstr(lpsz1,lpsz2);
        if(lpsz1)
            {
            dwCount++;
            lpsz1 += dwLen2;
            }
        }
    while(lpsz1);

    return dwCount;
}


//----------------------------------------------------------------------------
//
//  Function:   CSplit constructor
//
//  Synopsis:   Split the string and build internal arrays
//
//  Arguments:  [lpszString] -- string to split
//              [lpszDelimiter] -- delimiter string for the split
//              [boDestructive] -- if TRUE, lpszString is lost (use less
//               memory because avoid some new[], but lpszString is written in)
//
//  Return:     -
//
//----------------------------------------------------------------------------
CSplit::CSplit(char* lpszString,char* lpszDelimiter,BOOL boDestructive)
{
    m_boDestructive = boDestructive;
    m_lpArray       = NULL;
    m_dwNumEntries  = 0;

    if(lpszString  &&  *lpszString  &&  lpszDelimiter  &&  *lpszDelimiter)
        {
        DWORD   dwDelimiterLen  = strlen(lpszDelimiter);
        DWORD   dwNumEntries    = dwNbOccurrence(lpszString,lpszDelimiter) + 1;

        m_lpArray = new char*[dwNumEntries];

        char* lpsz;

        do
            {
            lpsz = strstr(lpszString,lpszDelimiter);

            if(lpsz)
                {
                if(!m_boDestructive)
                    {
                    DWORD   dwStringLen = lpsz - lpszString;

                    m_lpArray[m_dwNumEntries] = new char[dwStringLen + 1];
                    memcpy(m_lpArray[m_dwNumEntries],lpszString,dwStringLen);
                    m_lpArray[m_dwNumEntries][dwStringLen] = 0;
                    }
                else
                    {
                    *lpsz = 0;
                    m_lpArray[m_dwNumEntries] = lpszString;
                    }
                m_dwNumEntries++;
                lpszString = lpsz + dwDelimiterLen;
                }
            }
        while(lpsz);

        if(!m_boDestructive)
            {
            m_lpArray[m_dwNumEntries] = new char[strlen(lpszString) + 1];
            strcpy(m_lpArray[m_dwNumEntries],lpszString);
            }
        else
            {
            m_lpArray[m_dwNumEntries] = lpszString;
            }

        m_dwNumEntries++;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   CSplit destructor
//
//  Synopsis:   Release the memory
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
CSplit::~CSplit()
{
    if(!m_boDestructive  &&  m_lpArray) for(DWORD dwI = 0; dwI < m_dwNumEntries; dwI++) delete m_lpArray[dwI];
    if(m_lpArray) delete [] m_lpArray;

    m_lpArray       = NULL;
    m_dwNumEntries  = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   Operator[] for the CSplit class
//
//  Synopsis:   Allow access to the splitted string
//
//  Arguments:  [lIndex] -- index of the string to get
//
//  Return:     pointer on the string if index is valid, else NULL
//
//----------------------------------------------------------------------------
char*   CSplit::operator [] (long lIndex)
{
    if(lIndex + 1 > (long)m_dwNumEntries) return NULL;

    return m_lpArray[lIndex];
}
