#include "stdafx.h"

//#define __EXACT_TRANGLE__


//----------------------------------------------------------------------------
//
//  Function:   vComputeMoveableExtents
//
//  Synopsis:   Compute the bounding box of a moveable
//
//  Arguments:  [lpMoveable] -- Moveable concerned
//              [lpfMinx,lpfMaxX,lpfMinY,lpfMaxY,lpfMinZ,lpfMaxZ] -- at
//               return, holds the bounding box extents
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vComputeMoveableExtents(TR2_MOVEABLE* lpMoveable,float* lpfMinX,float* lpfMaxX,float* lpfMinY,float* lpfMaxY,float* lpfMinZ,float* lpfMaxZ)
{
    *lpfMinX = 1e20f; *lpfMaxX = -1e20f;
    *lpfMinY = 1e20f; *lpfMaxY = -1e20f;
    *lpfMinZ = 1e20f; *lpfMaxZ = -1e20f;

    #define STACK_DEPTHMAX 10

    float   Stack[3 * STACK_DEPTHMAX];
    DWORD   dwStackPointer = 0;
    float   fTransX = 0, fTransY = 0, fTransZ = 0;

    TR2_MESHTREE*   lpMeshTree = &m_trFile.tr123File.lpMeshTrees[lpMoveable->dwMeshTree];
    TR123_MESH*     lpMesh = m_trFile.tr123File.lplpMeshes[m_trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + 0]];

    if(lpMesh)
        {
        vComputeMeshExtents(lpMesh);
        *lpfMinX = lpMesh->fMinX;
        *lpfMaxX = lpMesh->fMaxX;
        *lpfMinY = lpMesh->fMinY;
        *lpfMaxY = lpMesh->fMaxY;
        *lpfMinZ = lpMesh->fMinZ;
        *lpfMaxZ = lpMesh->fMaxZ;
        }

   for(DWORD dwI = 1; dwI < lpMoveable->wNumMeshes; dwI++)
        {
        DWORD   dwNumMesh = m_trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + dwI];
        lpMesh = m_trFile.tr123File.lplpMeshes[dwNumMesh];

        bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
        bit32   lTransX = lpMeshTree->lCoord; lpMeshTree++;
        bit32   lTransY = lpMeshTree->lCoord; lpMeshTree++;
        bit32   lTransZ = lpMeshTree->lCoord; lpMeshTree++;

        if(dwStack & 1)
            {
            if(dwStackPointer == 0)
                {
                //!AfxMessageBox("Moveable bounding box calculation: stack pointer = 0 with pop operation !",MB_OK | MB_ICONSTOP | MB_TOPMOST);
				//! Have seen this condition occure for mvb #418 in TR4 levels
				//! However, it seems that the moveable is rendered correctly... Go figure !
                }
            else
                {
                dwStackPointer -= 3;
                fTransX = Stack[dwStackPointer + 0];
                fTransY = Stack[dwStackPointer + 1];
                fTransZ = Stack[dwStackPointer + 2];
                }
            }

        if(dwStack & 2)
            {
            if(dwStackPointer + 3 > STACK_DEPTHMAX)
                {
                AfxMessageBox("Moveable bounding box calculation: stack overflow !",MB_OK | MB_ICONSTOP | MB_TOPMOST);
                }
            else
                {
                Stack[dwStackPointer + 0] = fTransX;
                Stack[dwStackPointer + 1] = fTransY;
                Stack[dwStackPointer + 2] = fTransZ;
                dwStackPointer += 3;
                }
            }

        fTransX +=  (float)lTransX;
        fTransY += -(float)lTransY;
        fTransZ += -(float)lTransZ;

        if(lpMesh)
            {
            vComputeMeshExtents(lpMesh);

            if(lpMesh->fMinX + fTransX < *lpfMinX) *lpfMinX = lpMesh->fMinX + fTransX;
            if(lpMesh->fMinY + fTransY < *lpfMinY) *lpfMinY = lpMesh->fMinY + fTransY;
            if(lpMesh->fMinZ + fTransZ < *lpfMinZ) *lpfMinZ = lpMesh->fMinZ + fTransZ;
            if(lpMesh->fMaxX + fTransX > *lpfMaxX) *lpfMaxX = lpMesh->fMaxX + fTransX;
            if(lpMesh->fMaxY + fTransY > *lpfMaxY) *lpfMaxY = lpMesh->fMaxY + fTransY;
            if(lpMesh->fMaxZ + fTransZ > *lpfMaxZ) *lpfMaxZ = lpMesh->fMaxZ + fTransZ;
            }
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vTranslateMeshInMoveable
//
//  Synopsis:   Translate a mesh of a moveable
//
//  Arguments:  [lpMoveable] -- moveable
//              [dwNumMeshToTranslate] -- n of mesh to translate
//              [sX,sY,sZ] -- translation on X / Y / Z axis to apply
//              [boTranslateChildren] -- true if the translation of the mesh
//               translates its children too
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vTranslateMeshInMoveable(TR2_MOVEABLE* lpMoveable,DWORD dwNumMeshToTranslate,bit16 sX,bit16 sY,bit16 sZ,bool boTranslateChildren)
{
    TR2_MESHTREE* lpMeshTree = &m_trFile.tr123File.lpMeshTrees[lpMoveable->dwMeshTree];

    if(boTranslateChildren)
        {
        for(DWORD dwI = 1; dwI < lpMoveable->wNumMeshes; dwI++)
            {
            DWORD dwNumMesh = m_trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + dwI];

            bool boOk = (dwNumMesh == dwNumMeshToTranslate);

            bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
            bit32   lTransX = lpMeshTree->lCoord; if(boOk) lpMeshTree->lCoord += sX; lpMeshTree++;
            bit32   lTransY = lpMeshTree->lCoord; if(boOk) lpMeshTree->lCoord += sY; lpMeshTree++;
            bit32   lTransZ = lpMeshTree->lCoord; if(boOk) lpMeshTree->lCoord += sZ; lpMeshTree++;
            }
        return;
        }

    #define STACK_DEPTHMAX 10

    DWORD   Stack[STACK_DEPTHMAX];
    DWORD   dwStackPointer = 0;
    DWORD   dwLevel = 0, dwMeshLevel = 255;

    for(DWORD dwI = 1; dwI < lpMoveable->wNumMeshes; dwI++)
        {
        DWORD dwNumMesh = m_trFile.tr123File.lpMeshPointers[lpMoveable->wStartingMesh + dwI];

        bool boOk = (dwNumMesh == dwNumMeshToTranslate);

        bitu32  dwStack = lpMeshTree->lCoord; lpMeshTree++;
        
        if(dwStack & 1)
            {
            if(dwStackPointer > 0)
                {
                dwStackPointer--;
                dwLevel = Stack[dwStackPointer];
                if(dwLevel <= dwMeshLevel) dwMeshLevel = 255;
                }
            }

        if(boOk) dwMeshLevel = dwLevel;

        bit32   lTransX = lpMeshTree->lCoord;
        if(boOk) lpMeshTree->lCoord += sX;
        if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord -= sX;
        lpMeshTree++;

        bit32   lTransY = lpMeshTree->lCoord;
        if(boOk) lpMeshTree->lCoord += sY;
        if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord -= sY;
        lpMeshTree++;

        bit32   lTransZ = lpMeshTree->lCoord;
        if(boOk) lpMeshTree->lCoord += sZ;
        if(dwLevel == dwMeshLevel+1) lpMeshTree->lCoord -= sZ;
        lpMeshTree++;

        if(dwStack & 2)
            {
            if(dwStackPointer + 1 <= STACK_DEPTHMAX)
                {
                Stack[dwStackPointer] = dwLevel;
                dwStackPointer++;
                }
            }

        dwLevel++;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vCopyMoveable
//
//  Synopsis:   Copy a moveable in another moveable
//
//  Arguments:  [lpTRFileSrc] -- pointer on the CTRFile where the source
//               moveable comes from
//              [lpMoveableSrc] -- Source moveable
//              [lpTRFileDst] -- pointer on the CTRFile where the destination
//               moveable comes from
//              [lpMoveableDst] -- Destination moveable
//              [boKeepMesh0] -- true to not replace the dest mesh if it is
//               the mesh#0
//              [boDontCopyMesh0] -- true to not replace the dest mesh if the
//               source mesh is mesh#0
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vCopyMoveable(CTRFile* lpTRFileSrc,TR2_MOVEABLE* lpMoveableSrc,CTRFile* lpTRFileDst,TR2_MOVEABLE* lpMoveableDst,bool boKeepMesh0,bool boDontCopyMesh0)
{
    if(lpMoveableDst->wNumMeshes != lpMoveableSrc->wNumMeshes) return;

    TR_FILE&        trFileSrc = lpTRFileSrc->GetTRFile();
    TR2_MESHTREE*   lpMeshTreeSrc = &trFileSrc.tr123File.lpMeshTrees[lpMoveableSrc->dwMeshTree];
    DWORD*          lpMeshPointersSrc = &trFileSrc.tr123File.lpMeshPointers[lpMoveableSrc->wStartingMesh + 0];

    TR_FILE&        trFileDst = lpTRFileDst->GetTRFile();
    TR2_MESHTREE*   lpMeshTreeDst = &trFileDst.tr123File.lpMeshTrees[lpMoveableDst->dwMeshTree];
    DWORD*          lpMeshPointersDst = &trFileDst.tr123File.lpMeshPointers[lpMoveableDst->wStartingMesh + 0];

    memcpy(lpMeshTreeDst,lpMeshTreeSrc,(lpMoveableSrc->wNumMeshes - 1) * 4 * sizeof(TR2_MESHTREE));

    for(WORD wI = 0; wI < lpMoveableSrc->wNumMeshes; wI++)
        {
        DWORD dwNumMeshSrc = lpMeshPointersSrc[wI];
        DWORD dwNumMeshDst = lpMeshPointersDst[wI];

        if(boKeepMesh0  &&  !dwNumMeshDst) ;
        else if(boDontCopyMesh0  &&  !dwNumMeshSrc) ;
        else lpMeshPointersDst[wI] = dwNumMeshSrc;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   dwGetNumAnimsForMoveable
//
//  Synopsis:   Get the number of animations of a moveable
//
//  Arguments:  [lpMoveable] -- moveable to get the number of animations
//
//  Return:     number of animations for this moveable
//
//----------------------------------------------------------------------------
DWORD   CTRFile::dwGetNumAnimsForMoveable(TR2_MOVEABLE* lpMoveable)
{
    // That's ugly code ! But it seems there's not this info (num animations)
    // available in the moveable, so to calculate this info, we:
    //  - get the "# starting anim" for the next moveable (->N)
    //  - substract the "# starting anim" for lpMoveable to N
    // Doing this, we assume that the next moveable has its animations following
    // the animations of the current moveable (seems right for all tested
    // levels, but...)
    // We have to deal with the fact that the next moveable could have "# starting anim" == -1
    // (ie. anim handled by the engine, like the ponytail anim). If it's the case, we skip the moveable
    // and use the next moveable for our computation

    WORD    wStartAnim = lpMoveable->wAnimation;
    WORD    wNextStartAnim;

    DWORD   dwNumMoveables = m_trFile.tr123File.dwNumMoveables;
    TR2_MOVEABLE*   lpLastMoveable = &m_trFile.tr123File.lpMoveables[dwNumMoveables-1];

    while(lpMoveable != lpLastMoveable)
        {
        TR2_MOVEABLE* lpNextMoveable = lpMoveable + 1;
        wNextStartAnim = lpNextMoveable->wAnimation;
        if(wNextStartAnim != 0xFFFF) break;
        lpMoveable = lpNextMoveable++;
        }

    if(lpMoveable == lpLastMoveable)
        {
        wNextStartAnim = m_trFile.tr123File.dwNumAnimations;
        }

    WORD    wNumAnim = (wStartAnim != 0xFFFF) ? wNextStartAnim - wStartAnim : 0;

    return wNumAnim;
}


//----------------------------------------------------------------------------
//
//  Function:   vGetRotAngleForFrame
//
//  Synopsis:   Get angles of rotation from TR frame data
//
//  Arguments:  [trType] -- type of TR (TR1 / TR2 / TR3)
//              [lplpFrame] -- data of the frame
//
//  Return:     [fAngleX, fAngleY, fAngleZ] -- angles of rotation
//              *lplpFrame is positionned on the data for the next mesh
//
//----------------------------------------------------------------------------
/*static*/ void    CTRFile::vGetRotAngleForFrame(enumTRType trType,bitu16** lplpFrame,FLOAT& fAngleX,FLOAT& fAngleY,FLOAT& fAngleZ, WORD& wTRAngle1, WORD& wTRAngle2)
{
    bitu16  wFrameData = **lplpFrame; (*lplpFrame)++;

#ifndef __EXACT_TRANGLE__
    wTRAngle1 = 0;
    wTRAngle2 = 0;
#else
    wTRAngle1 = wFrameData;
    wTRAngle2 = 0x1234;
#endif

    if((wFrameData & 0xC000)  &&  (trType != TR1_FILE)) // single axis of rotation
        {
        WORD  wAngle = trType == TR4_FILE ? (wFrameData & 0xFFF) >> 2 : wFrameData & 0x3FF;
        FLOAT fAngle = wAngle;

		fAngle *= 360.0 / 1024.0;

        switch(wFrameData & 0xC000)
            {
            case 0x4000:
#ifndef __EXACT_TRANGLE__
                wTRAngle1 = wAngle << 4;
#endif
                fAngleX = fAngle;
                return;
            case 0x8000:
#ifndef __EXACT_TRANGLE__
                wTRAngle1 = (wAngle & 0x03C0) >> 6;
                wTRAngle2 = (wAngle & 0x003F) << 10;
#endif
                fAngleY = fAngle;
                return;
            case 0xC000:
#ifndef __EXACT_TRANGLE__
                wTRAngle2 = wAngle;
#endif
                fAngleZ = fAngle;
                return;
            }
        }
    else    // 3 axis of rotation
        {
        WORD    wFrameData2 = **lplpFrame; (*lplpFrame)++;

        if(trType == TR1_FILE)
            {
            WORD wTemp  = wFrameData;
            wFrameData  = wFrameData2;
            wFrameData2 = wTemp;
            }

        wTRAngle1 = wFrameData;
        wTRAngle2 = wFrameData2;

#ifdef __EXACT_TRANGLE__
        ASSERT(wTRAngle2 != 0x1234);
#endif

        fAngleX = (wFrameData >> 4) & 0x03FF;
        fAngleX *= 360.0 / 1024.0;

        fAngleY = (wFrameData << 6) & 0x03C0;
        fAngleY += (wFrameData2 >> 10) & 0x003F;
        fAngleY *= 360.0 / 1024.0;

        fAngleZ = wFrameData2 & 0x3FF;
        fAngleZ *= 360.0 / 1024.0;
        }

    return;
}


//----------------------------------------------------------------------------
//
//  Function:   lpGetMoveable
//
//  Synopsis:   Look for a given moveable in the level
//
//  Arguments:  [dwObjId] -- id of the moveable
//
//  Return:     pointer on the moveable if found, else NULL
//
//----------------------------------------------------------------------------
CTrMvb* CTRFile::lpGetMoveable(DWORD dwObjId)
{
    TRListMoveable*     lpListMoveable = m_trFile.tr123File.lpListMoveable;

    for(TRListMoveable::iterator theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++)
        {
        CTrMvb* lpMvb = *theIterator;

        if(lpMvb->dwGetObjectId() == dwObjId) return lpMvb;
        }

    return NULL;
}


//----------------------------------------------------------------------------
//
//  Function:   vMakeSureSoundsAreAvailable
//
//  Synopsis:   Scan all the moveables and make sure that sounds used are
//              available in the level
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTRFile::vMakeSureSoundsAreAvailable()
{
    TRListMoveable*     lpListMoveable = m_trFile.tr123File.lpListMoveable;

    for(TRListMoveable::iterator theIterator = lpListMoveable->begin(); theIterator != lpListMoveable->end(); theIterator++)
        {
        CTrMvb* lpMvb = *theIterator;

        m_trFile.tr123File.lpSndMgr->vMerge(lpMvb->lpGetSndMgr());
        }
}
