#include "stdafx.h"

#include "zlib.h"
#include "TextureTools.h"

#import "ReadPic\ReadPic.tlb" no_namespace


// Blit a bitmap to the given DC
void    vDrawBitmap(CDC* pdc,HBITMAP hBitmap,int xStart,int yStart)
{
    BITMAP  bm;
    CDC     hdcMem;
    POINT   ptSize,ptOrg;

    hdcMem.CreateCompatibleDC(pdc);
    hdcMem.SelectObject(hBitmap);
    SetMapMode(hdcMem,GetMapMode(pdc->m_hDC));

    GetObject(hBitmap,sizeof(BITMAP),(LPVOID)&bm);
    ptSize.x = bm.bmWidth;
    ptSize.y = bm.bmHeight;
    DPtoLP(pdc->m_hDC,&ptSize,1);

    ptOrg.x = 0;
    ptOrg.y = 0;
    DPtoLP(hdcMem,&ptOrg,1);

    pdc->BitBlt(xStart,yStart,ptSize.x,ptSize.y,&hdcMem,ptOrg.x,ptOrg.y,SRCCOPY);
}


/* ===========================================================================
     Decompresses the source buffer into the destination buffer.  sourceLen is
   the byte length of the source buffer. Upon entry, destLen is the total
   size of the destination buffer, which must be large enough to hold the
   entire uncompressed data. (The size of the uncompressed data must have
   been saved previously by the compressor and transmitted to the decompressor
   by some mechanism outside the scope of this compression library.)
   Upon exit, destLen is the actual size of the compressed buffer.
     This function can be used to decompress a whole file at once if the
   input file is mmap'ed.

     uncompress returns Z_OK if success, Z_MEM_ERROR if there was not
   enough memory, Z_BUF_ERROR if there was not enough room in the output
   buffer, or Z_DATA_ERROR if the input data was corrupted.
*/
int uncompress (BYTE* dest, DWORD* destLen, BYTE* source, DWORD sourceLen)
{
    z_stream stream;
    int err;

    stream.next_in = (Bytef*)source;
    stream.avail_in = (uInt)sourceLen;

    stream.next_out = dest;
    stream.avail_out = (uInt)*destLen;

    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;

    err = inflateInit(&stream);
    if (err != Z_OK) return err;

    err = inflate(&stream, Z_FINISH);
    if (err != Z_STREAM_END) {
        inflateEnd(&stream);
        return err;
    }
    *destLen = stream.total_out;

    err = inflateEnd(&stream);
    return err;
}


/* ===========================================================================
     Compresses the source buffer into the destination buffer.  sourceLen is
   the byte length of the source buffer. Upon entry, destLen is the total
   size of the destination buffer, which must be at least 0.1% larger than
   sourceLen plus 8 bytes. Upon exit, destLen is the actual size of the
   compressed buffer.
     This function can be used to compress a whole file at once if the
   input file is mmap'ed.
     compress returns Z_OK if success, Z_MEM_ERROR if there was not
   enough memory, Z_BUF_ERROR if there was not enough room in the output
   buffer.
*/
int compress (BYTE* dest, DWORD* destLen, BYTE* source, DWORD sourceLen)
{
    z_stream stream;
    int err;

    stream.next_in = source;
    stream.avail_in = sourceLen;
    stream.next_out = dest;
    stream.avail_out = *destLen;

    stream.zalloc = (alloc_func)0;
    stream.zfree = (free_func)0;
    stream.opaque = (voidpf)0;

    err = deflateInit(&stream, Z_DEFAULT_COMPRESSION);
    if (err != Z_OK) return err;

    err = deflate(&stream, Z_FINISH);
    if (err != Z_STREAM_END) {
        deflateEnd(&stream);
        return err == Z_OK ? Z_BUF_ERROR : err;
    }
    *destLen = stream.total_out;

    err = deflateEnd(&stream);
    return err;
}


// Convert a 8 / 16 or 32 bits source buffer to a 32 bits one and flip the image vertically
// For the 16 and 32 bits source buffer, colors must be ARGB format (B beeing the least significant bits, A the most)
// 16 bits: R=G=B=5 bits and A=1 bit (if set, pixel not transparent)
// 32 bits: R=G=B=A=8 bits (if A=255, pixel not transparent, if A=0 pixel fully transparent)
// The dest buffer is ARGB format, with R=G=B=A=8 bits
void    vPromoteTo32Bits(BYTE* lpDataSrc,DWORD dwWidth,DWORD dwHeight,DWORD dwDepth,BYTE* lpbPalette,BYTE* lpDataDest,BOOL boReverse,DWORD dwSrcWidth,DWORD dwSrcHeight)
{
    if(dwSrcWidth == -1)    dwSrcWidth = dwWidth;
    if(dwSrcHeight == -1)   dwSrcHeight = dwHeight;

    switch(dwDepth)
        {
        // Convert 8 to 32 bits
        case 8:
            {
            for(DWORD dwJ = 0; dwJ < dwSrcHeight; dwJ++)
                {
                for(DWORD dwI = 0; dwI < dwSrcWidth; dwI++)
                    {
                    BYTE bIndex;
                    
                    if(boReverse) bIndex = *((BYTE*)(lpDataSrc + dwI + (dwSrcHeight - 1 - dwJ) * dwSrcWidth));
                    else bIndex = *((BYTE*)(lpDataSrc + dwI + dwJ * dwSrcWidth));

                    *lpDataDest++ = *((BYTE*)(lpbPalette + 3 * bIndex + 2)) * 4;
                    *lpDataDest++ = *((BYTE*)(lpbPalette + 3 * bIndex + 1)) * 4;
                    *lpDataDest++ = *((BYTE*)(lpbPalette + 3 * bIndex + 0)) * 4;
                    *lpDataDest++ = bIndex ? 255 : 0;
                    }
                lpDataDest += dwWidth - dwSrcWidth;
                }
            break;
            }
        // Convert 16 to 32 bits
        case 16:
            {
            for(DWORD dwJ = 0; dwJ < dwSrcHeight; dwJ++)
                for(DWORD dwI = 0; dwI < dwSrcWidth; dwI++)
                    {
                    BYTE    bARGB[4];
                    WORD    wData = *((WORD*)(lpDataSrc + (dwI + dwJ * dwSrcWidth) * 2));
                    bARGB[0] = ((wData >> 0) & 31) << 3;
                    bARGB[1] = ((wData >> 5) & 31) << 3;
                    bARGB[2] = ((wData >> 10) & 31) << 3;
                    bARGB[3] = wData >> 15; if(bARGB[3]) bARGB[3] = 255;
                    if(boReverse) memcpy(lpDataDest + (dwI + (dwHeight - 1 - dwJ) * dwWidth) * 4,bARGB,4);
                    else memcpy(lpDataDest + (dwI + dwJ * dwWidth) * 4,bARGB,4);
                    }
            break;
            }
        // Keep 32 bits
        case 32:
            {
            for(DWORD dwJ = 0; dwJ < dwSrcHeight; dwJ++)
                {
                if(boReverse) memcpy(lpDataDest + (dwHeight - 1 - dwJ) * dwWidth * 4,lpDataSrc + dwJ * dwSrcWidth * 4,dwSrcWidth * 4);
                else memcpy(lpDataDest + dwJ * dwWidth * 4,lpDataSrc + dwJ * dwSrcWidth * 4,dwSrcWidth * 4);
                }
            break;
            }
        }
}


// Expand a TR4 texture
void    vUncompressTexture(TR4_COMPR_TEXTURE* lpComprTexture,bool bo32Bits,BYTE** lplpDataUncomp,DWORD* lpdwHeight)
{
    DWORD   dwUncompressedSize  = lpComprTexture->dwTextUncompSize;
    DWORD   dwCompressedSize    = lpComprTexture->dwTextCompSize;
    
    *lplpDataUncomp      = new BYTE[dwUncompressedSize];
    
    if(bo32Bits) *lpdwHeight = dwUncompressedSize / 4 / 256; else *lpdwHeight = dwUncompressedSize / 2 / 256;

    uncompress(*lplpDataUncomp,&dwUncompressedSize,lpComprTexture->lpbTextData,dwCompressedSize);
}


// Compress a texture to the format known by TR4
void    vCompressTexture(BYTE* lpDataUncomp,DWORD dwDataUncompSize,TR4_COMPR_TEXTURE* lpComprTexture)
{
    lpComprTexture->dwTextCompSize = dwDataUncompSize + (dwDataUncompSize/10) + 8;
    BYTE*   lpBuf = new BYTE[lpComprTexture->dwTextCompSize];
    lpComprTexture->dwTextUncompSize = dwDataUncompSize;
    compress (lpBuf,&lpComprTexture->dwTextCompSize,lpDataUncomp,dwDataUncompSize);
    lpComprTexture->lpbTextData = new BYTE[lpComprTexture->dwTextCompSize];
    memcpy(lpComprTexture->lpbTextData,lpBuf,lpComprTexture->dwTextCompSize);
    delete [] lpBuf;
}


// Load an image into a 32 bits buffer. The gfx format handled are:
//   TGA,BMP,GIF,FLI,FLC and CEL
DWORD   dwLoadImage(const TCHAR* lpszFileName,DWORD* lpdwWidth,DWORD* lpdwHeight,BYTE** lplpData)
{
    CoInitialize(NULL);

    DWORD           dwError = 0;
    long            dwWidth = 0, dwHeight = 0;
    BYTE*           lpData = NULL;
    IImageBasePtr   ReadPic;
    
    ReadPic.CreateInstance(__uuidof(ImageBase));

    if(ReadPic == NULL) return -1;

    _bstr_t bsttImageName(lpszFileName);

    DWORD   dwImgHandle = ReadPic->ReadImage(bsttImageName);

    // The file could be read ?
    if(dwImgHandle != 0)
        {
        // Yes, get informations about the image
        ReadPic->CurrentImageHandle = dwImgHandle;
        
        bool bo32Bits  = ReadPic->ImageDepthBits == 24 ? false : true;
        
        DWORD dwNbBytesPerTexel = 4; // Final images are always 32 bits

        dwWidth      = ReadPic->ImageWidth;
        dwHeight     = ReadPic->ImageHeight;
        DWORD dwNbBytesPerLine = dwWidth * dwNbBytesPerTexel;
        DWORD dwDataSize    = dwHeight * dwNbBytesPerLine;
        lpData       = new BYTE[dwDataSize];

        if(bo32Bits)
            {
            // Source image is already 32 bits, so make a raw copy
            memcpy(lpData,ReadPic->ImageData,dwDataSize);
            }
        else
            {
            // Convert from 24 bits to 32 bits
            BYTE*   lpSrcData = ReadPic->ImageData;
            memset(lpData,255,dwDataSize);
            for(long lV = 0; lV < dwHeight; lV++)
                {
                for(long lU = 0; lU < dwWidth; lU++)
                    {
                    memcpy(lpData + lV * dwNbBytesPerLine + lU * dwNbBytesPerTexel,lpSrcData,3);
                    lpSrcData += 3;
                    }
                }
            }
        
        // Release resources
        ReadPic->ReleaseImage(dwImgHandle);
        }
    else dwError = ReadPic->LastError;

    ReadPic.Release();

    CoUninitialize();

    *lpdwWidth  = dwWidth;
    *lpdwHeight = dwHeight;
    *lplpData   = lpData;

    return dwError;
}


// Convert a ABGR buffer to a ARGB one (dwLoadImage() produces a ABGR buffer)
void    vConvertABGRToARGB(DWORD dwWidth,DWORD dwHeight,bool bo32Bits,BYTE* lpSrc,BYTE* lpDst)
{
    for(DWORD dwJ = 0; dwJ < dwHeight; dwJ++)
        {
        for(DWORD dwI = 0; dwI < dwWidth; dwI++)
            {
            BYTE bR = *lpSrc;
            BYTE bG = *(lpSrc+1);
            BYTE bB = *(lpSrc+2);
            BYTE bA = *(lpSrc+3);

            if(bo32Bits)
                {
                *lpDst = bB;
                *(lpDst+1) = bG;
                *(lpDst+2) = bR;
                *(lpDst+3) = bA;
                lpDst += 4;
                }
            else
                {
                WORD wPix = ((bA >> 7) << 15) + ((bR >> 3) << 10) + ((bG >> 3) << 5) + ((bB >> 3) << 0);
            
                *((WORD*)lpDst) = wPix;
                lpDst += 2;
                }
            
            lpSrc += 4;
            }
        }
}


// Add right / bottom borders to a source bitmap (32 bits) in order to have (dwDstWidth,dwDstHeight) dimensions
BYTE*   lpbEnlarge32BitsBitmap(BYTE* lpbSrc,DWORD dwSrcWidth,DWORD dwSrcHeight,DWORD dwDstWidth,DWORD dwDstHeight)
{
    if(dwSrcWidth > dwDstWidth  ||  dwSrcHeight > dwDstHeight) return NULL;

    DWORD   dwDataLen = dwDstWidth * dwDstHeight * 4;
    BYTE*   lpbDst = new BYTE [dwDataLen];

    ZeroMemory(lpbDst, dwDataLen);

    lpbSrc += 4 * dwSrcWidth * (dwSrcHeight - 1);

    while(dwSrcHeight--)
        {
        memcpy(lpbDst + dwSrcHeight * 4 * dwDstWidth, lpbSrc, 4 * dwSrcWidth);
        lpbSrc -= 4 * dwSrcWidth;
        }

    return lpbDst;
}


//
BYTE    bFindColor(BYTE* lpbPalette,BYTE bR,BYTE bG,BYTE bB)
{
    DWORD       dwMin = 10000000;
    BYTE        bIdx = 0;

    for(DWORD dwI = 0; dwI < 256; dwI++)
        {
        DWORD   dwDiff;

        TR2_COLOR col = ((TR2_PALETTE8*)lpbPalette)->aRGB[dwI];

        long    lRDiff = (DWORD)bR - col.bRed     * 4; if(lRDiff < 0) lRDiff = -lRDiff;
        long    lGDiff = (DWORD)bG - col.bGreen   * 4; if(lGDiff < 0) lGDiff = -lGDiff;
        long    lBDiff = (DWORD)bB - col.bBlue    * 4; if(lBDiff < 0) lBDiff = -lBDiff;

        dwDiff = lRDiff + lGDiff + lBDiff;

        if(dwDiff < dwMin)
            {
            dwMin = dwDiff;
            bIdx = (BYTE)dwI;
            if(dwMin == 0) break;
            }
        }

    return bIdx;
}


// Convert a 16 bits buffer to a 8 bits one
void    vConvert16BitsTo8Bits(BYTE* lpbSrc,DWORD dwSrcWidth,DWORD dwSrcHeight,BYTE* lpbDst,BYTE* lpMap)
{
    for(DWORD dwJ = 0; dwJ < dwSrcHeight; dwJ++)
        {
        for(DWORD dwI = 0; dwI < dwSrcWidth; dwI++)
            {
            BYTE    bIdx;
            WORD    wData = *((WORD*)(lpbSrc + (dwI + dwJ * dwSrcWidth) * 2));
            DWORD   dwB = ((wData >> 0) & 31);
            DWORD   dwG = ((wData >> 5) & 31);
            DWORD   dwR = ((wData >> 10) & 31);
            BYTE    bA  = wData >> 15;

            if(bA == 0) bIdx = 0; else bIdx = lpMap[dwB + dwG * 32 + dwR * 32 * 32];

            lpbDst[dwI + dwJ * dwSrcWidth] = bIdx;
            }
        }
}
