// TlboxMoveableMain.cpp : implementation file
//

#include "stdafx.h"

#include "trviewer.h"
#include "ChildFrm.h"
#include "HandleBase.h"
#include "HandleMoveable.h"
#include "EditBase.h"
#include "EditMoveable.h"

#include "ToolboxBase.h"
#include "ToolboxItfBase.h"
#include "ToolboxSheet.h"

#include "TlboxMoveableItf.h"
#include "TlboxMoveableMain.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CTlboxMoveableMain property page

IMPLEMENT_DYNCREATE(CTlboxMoveableMain, CPropertyPage)

CTlboxMoveableMain::CTlboxMoveableMain( CToolboxBase* lpParent ) : CToolboxItfBase(CTlboxMoveableMain::IDD, lpParent)
{
	//{{AFX_DATA_INIT(CTlboxMoveableMain)
	m_cstrObjId = _T("");
	m_cstrMvblExtents = _T("");
	m_boShowRealMeshes = FALSE;
	m_boShowAssocMvb = FALSE;
	m_boShowMoveableBB = FALSE;
	//}}AFX_DATA_INIT
}

CTlboxMoveableMain::~CTlboxMoveableMain()
{
}

void CTlboxMoveableMain::DoDataExchange(CDataExchange* pDX)
{
	CPropertyPage::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CTlboxMoveableMain)
	DDX_Text(pDX, IDC_OBJID, m_cstrObjId);
	DDX_Text(pDX, IDC_MVBLEXTENTS, m_cstrMvblExtents);
	DDX_Check(pDX, IDC_SHOWREALMESHES, m_boShowRealMeshes);
	DDX_Check(pDX, IDC_SHOWASSOCMVB, m_boShowAssocMvb);
	DDX_Check(pDX, IDC_SHOWMOVEABLEBB, m_boShowMoveableBB);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CTlboxMoveableMain, CPropertyPage)
	//{{AFX_MSG_MAP(CTlboxMoveableMain)
	ON_EN_KILLFOCUS(IDC_OBJID, OnKillfocusObjId)
	ON_BN_CLICKED(IDC_SHOWREALMESHES, OnShowRealMeshes)
	ON_BN_CLICKED(IDC_SHOWASSOCMVB, OnShowAssocMvb)
	ON_BN_CLICKED(IDC_SHOWMOVEABLEBB, OnShowMoveableBB)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CTlboxMoveableMain message handlers

void CTlboxMoveableMain::vInit( bool boNewMoveable )
{
	GET_TLBOXHANDLER( CHandleMoveable );

    CString s(_T(""));
    CTrMvb* lpMoveable = lpHandler ? lpHandler->GetMoveable() : NULL;
    DWORD   dwNumTris = 0, dwNumQuads = 0;

    // The new moveable exists ?
    if(lpMoveable)
        {
        // Yes, set some general infos about the moveable in the toolbox
        for(DWORD dwI = 0; dwI < lpMoveable->dwGetNumMeshes(); dwI++)
            {
            const TR123_MESH* lpMesh = lpMoveable->lpGetMesh(dwI);

            if(lpMesh)
                {
                dwNumTris   += lpMesh->sNumColoredTriangles + lpMesh->sNumTexturedTriangles;
                dwNumQuads  += lpMesh->sNumColoredRectangles + lpMesh->sNumTexturedRectangles;
                }
            }

        // Num meshes
        s.Format("%ld (%ld tris, %ld quads)",lpMoveable->dwGetNumMeshes(),dwNumTris,dwNumQuads);
        GetDlgItem(IDC_NUMMESHES)->SetWindowText(s);

        // Object Id
        s.Format(_T("%ld"),lpMoveable->dwGetObjectId());
        m_cstrObjId = s;
        }
    else
        {
        // No, reset infos
        GetDlgItem(IDC_NUMMESHES)->SetWindowText(s);
        m_cstrObjId = s;
        }

    vRefreshExtents( false );

    GetDlgItem(IDC_MOVEABLENAME)->SetWindowText(lpHandler ? lpHandler->lpszGetTitle() : _T(""));

	m_boShowRealMeshes = lpMoveable ? lpMoveable->boEdtShowRealMeshes() : false;
	m_boShowAssocMvb = lpMoveable ? lpMoveable->boEdtShowAssocMvb() : false;

	((CButton*)GetDlgItem(IDC_SHOWREALMESHES))->EnableWindow( FALSE );
	((CButton*)GetDlgItem(IDC_SHOWASSOCMVB))->EnableWindow( FALSE );

	if( lpMoveable  &&  lpMoveable->dwEdtGetBodyMvbId() != -1 )
		{
		((CButton*)GetDlgItem(IDC_SHOWREALMESHES))->EnableWindow( TRUE );
		}

	if( lpMoveable  &&  lpMoveable->dwEdtGetAssocMvbId() != -1 )
		{
		((CButton*)GetDlgItem(IDC_SHOWASSOCMVB))->EnableWindow( TRUE );
		}

    UpdateData(FALSE);
}


// Display the bounding box extents of the moveable
void    CTlboxMoveableMain::vRefreshExtents( bool boUpdateData /*= true*/ )
{
	GET_TLBOXHANDLER( CHandleMoveable );

    CTrMvb* lpMoveable = lpHandler ? lpHandler->GetMoveable() : NULL;

    CString s( _T("") );

    if( lpMoveable )
        {
        CBoundingBox BoundingBox;

        lpMoveable->vComputeBoundingBox( BoundingBox );

        CVect3D vec = BoundingBox.GetBBHi() - BoundingBox.GetBBLo();

        s.Format( _T("%ld,%ld,%ld"), (long) vec[0], (long) vec[1], (long) vec[2] );
        }

    m_cstrMvblExtents = s;

    if( boUpdateData ) UpdateData( FALSE );
}


void	CTlboxMoveableMain::OnKillfocusObjId() 
{
	GET_TLBOXHANDLER( CHandleMoveable );

    CTrMvb*   lpMoveable = lpHandler ? lpHandler->GetMoveable() : NULL;

    if(!lpMoveable) return;

	UpdateData(TRUE);

    lpMoveable->vSetObjectId(_ttol(m_cstrObjId));
}


void	CTlboxMoveableMain::OnShowRealMeshes()
{
	GET_TLBOXHANDLER( CHandleMoveable );
	GET_TLBOXEDIT( CEditMoveable );

	if( !lpHandler  ||  !lpEdit ) return;

	UpdateData( TRUE );

	lpHandler->GetMoveable()->vEdtSetShowRealMeshes( m_boShowRealMeshes ? true : false );
	lpEdit->vSetGfxVisiblity( 0, true, m_boShowRealMeshes ? true : false );
}

void	CTlboxMoveableMain::OnShowAssocMvb()
{
	GET_TLBOXHANDLER( CHandleMoveable );
	GET_TLBOXEDIT( CEditMoveable );

	if( !lpHandler  ||  !lpEdit ) return;

	UpdateData( TRUE );

	lpHandler->GetMoveable()->vEdtSetShowAssocMvb( m_boShowAssocMvb ? true : false );
	lpEdit->vSetGfxVisiblity( 1, m_boShowAssocMvb ? true : false, m_boShowAssocMvb ? true : false );
}

void	CTlboxMoveableMain::OnShowMoveableBB()
{
	GET_TLBOXEDIT( CEditMoveable );

	if( !lpEdit ) return;

	UpdateData( TRUE );

	lpEdit->vSetShowMoveableBB( m_boShowMoveableBB ? true : false );
}
