#include "stdafx.h"

#include "Fragment.h"


//----------------------------------------------------------------------------
//
//  Function:   vInitialize
//
//  Synopsis:   Reset the members of the class
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vInitialize()
{
    m_lpLstMeshes       = NULL;
    m_dwNumMeshes       = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   vRelease
//
//  Synopsis:   Release memory / Reset the members of the class
//
//  Arguments:  [boReleaseParent] -- true to release the parent class
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vRelease(bool boReleaseParent /*= true*/)
{
    while(m_dwNumMeshes--)
        {
        CTRFile::vReleaseMeshMemory(&m_lpLstMeshes[m_dwNumMeshes]);
        }

    delete [] m_lpLstMeshes;

    vInitialize();

    if(boReleaseParent) CTrObjGfx::vRelease(true);
}


//----------------------------------------------------------------------------
//
//  Function:   vStealData
//
//  Synopsis:   Transfer data from an object to the current one: after the
//              operation, the source object has lost his data member !
//
//  Arguments:  [obj] -- object to get the data from
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vStealData(CTrLstMsh& obj)
{
    CTrObjGfx::vStealData(obj);

    m_dwNumMeshes       = obj.m_dwNumMeshes;
    m_lpLstMeshes       = obj.m_lpLstMeshes;

    obj.vInitialize();
}


void	CTrLstMsh::vCopy( const CTrLstMsh& o )
{
	CTrObjGfx::vCopy( o );

	TR123_MESH** lplpLstMeshes = new TR123_MESH* [ o.m_dwNumMeshes ];

	for( DWORD dwI = 0; dwI < o.m_dwNumMeshes; ++dwI )
		{
		lplpLstMeshes[ dwI ] = &o.m_lpLstMeshes[ dwI ];
		}

	vCreateMeshList( o.m_dwNumMeshes, lplpLstMeshes );

	delete [] lplpLstMeshes;
}


//----------------------------------------------------------------------------
//
//  Function:   Constructor of class CTrLstMsh
//
//  Synopsis:   Construct a CTrLstMsh from a TR mesh
//
//  Arguments:  [lpMesh] -- mesh to convert
//              [lpTRFile] -- 
//              [dwDesiredTextureWidth] -- desired width of the texture
//                (0 if no matter)
//              [dwDesiredTextureHeight] -- desired height of the texture
//                (0 if no matter)
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrLstMsh::CTrLstMsh(TR123_MESH* lpMesh, CTRFile* lpTRFile, DWORD dwDesiredTextureWidth /*= 0*/, DWORD dwDesiredTextureHeight /*= 0*/) :
        CTrObjGfx(VERSION_TRMESH,TrObj_Type_Mesh,(enumTrVersion)lpTRFile->GetTRType())
{
    vInitialize();

    TR123_MESH* LstMeshes[] = { lpMesh };

    // Construct CTrLstMsh
    vConstruct(lpTRFile,1,LstMeshes,dwDesiredTextureWidth,dwDesiredTextureHeight);
}


//----------------------------------------------------------------------------
//
//  Function:   Constructor of class CTrLstMsh
//
//  Synopsis:   Construct a CTrLstMsh from TR meshes
//
//  Arguments:  [lplpMeshes] -- list of meshes to convert
//              [dwNumMeshes] -- num meshes to convert
//              [lpTRFile] -- 
//              [dwDesiredTextureWidth] -- desired width of the texture
//                (0 if no matter)
//              [dwDesiredTextureHeight] -- desired height of the texture
//                (0 if no matter)
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrLstMsh::CTrLstMsh(TR123_MESH** lplpMeshes, DWORD dwNumMeshes, CTRFile* lpTRFile, DWORD dwDesiredTextureWidth /*= 0*/, DWORD dwDesiredTextureHeight /*= 0*/) :
        CTrObjGfx(VERSION_TRMESH,TrObj_Type_Mesh,(enumTrVersion)lpTRFile->GetTRType())
{
    vInitialize();

    // Construct CTrLstMsh
    vConstruct(lpTRFile,dwNumMeshes,lplpMeshes,dwDesiredTextureWidth,dwDesiredTextureHeight);
}


CTrLstMsh::CTrLstMsh( const CTrLstMsh& MshSrc, DWORD dwNumMesh, DWORD dwDesiredTextureWidth /*= 0*/, DWORD dwDesiredTextureHeight /*= 0*/ )
{
    vInitialize();

    // Construct CTrLstMsh
    vConstruct( MshSrc, dwNumMesh, dwDesiredTextureWidth, dwDesiredTextureHeight );
}


//----------------------------------------------------------------------------
//
//  Function:   Constructor of class CTrLstMsh
//
//  Synopsis:   Construct a CTrLstMsh from a block in memory
//
//  Arguments:  [lpMSH] -- memory block
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrLstMsh::CTrLstMsh(BYTE* lpMSH, bool boDeleteData) : CTrObjGfx(lpMSH)
{
    vInitialize();

    if(lpMSH)
        {
        //if(GetVersion() != VERSION_TRMESH) throw((DWORD)__LOADERR_MISMATCHVERSION__);

		BYTE* lpbData = lpMSH + CTrObjGfx::dwGetDataSize();

        vConstruct(lpbData);

        if(boDeleteData) delete [] lpMSH;
        }
}


CTrLstMsh::CTrLstMsh( const CTrLstMsh& o )
{
	vCopy( o );
}


//----------------------------------------------------------------------------
//
//  Function:   dwGetDataSize
//
//  Synopsis:   Get the size of the CTrLstMsh when put in a continuous block
//              of memory
//
//  Arguments:  -
//
//  Return:     size
//
//----------------------------------------------------------------------------
DWORD   CTrLstMsh::dwGetDataSize()
{
    DWORD   dwSize = 
                CTrObjGfx::dwGetDataSize() +
                sizeof(DWORD);                              /* m_dwNumMeshes */

    for(DWORD dwI = 0; dwI < m_dwNumMeshes; dwI++)
        {
        dwSize += CTRFile::dwCalcMeshDataSize(&m_lpLstMeshes[dwI],0,true);
        }

    return dwSize;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbGetData
//
//  Synopsis:   Convert a CTrLstMsh to a continuous block of memory
//
//  Arguments:  [dwDataSize] -- on return, hold the size of the block
//
//  Return:     pointer to the memory block
//
//  Caution:    you must make a delete [] on the pointer returned when you
//              have finished working with it !
//----------------------------------------------------------------------------
BYTE*   CTrLstMsh::lpbGetData(DWORD& dwDataSize)
{
    BYTE*   lpbData = CTrObjGfx::lpbGetData(dwDataSize);
    DWORD   dwTrObjGfxDataSize = CTrObjGfx::dwGetDataSize();
    BYTE*   lpbDataSave = lpbData;

    lpbData += dwTrObjGfxDataSize;

    // Meshes
    *(DWORD*)lpbData = m_dwNumMeshes;  lpbData += sizeof(DWORD);

    for(DWORD dwI = 0; dwI < m_dwNumMeshes; dwI++)
        {
        lpbData = CTRFile::lpbWriteMesh(&m_lpLstMeshes[dwI], lpbData, lpbData, true);
        }

    return lpbDataSave;
}


//----------------------------------------------------------------------------
//
//  Function:   boLoad
//
//  Synopsis:   Load a CTrLstMsh from disk
//
//  Arguments:  [cstrFileName] -- name of the file on disk
//
//  Return:     the object loaded & initialized
//
//----------------------------------------------------------------------------
/*static*/ CTrLstMsh*  CTrLstMsh::objLoad(CString& cstrFileName)
{
    BYTE*   lpbData = CTrObjBase::lpbLoad(cstrFileName);

    CTrLstMsh* lpMsh = NULL;

try
    {
    lpMsh = new CTrLstMsh(lpbData,true);
    }
catch(DWORD)
    {
    delete lpMsh;
    lpMsh = NULL;
    }

    return lpMsh;
}


//----------------------------------------------------------------------------
//
//  Function:   vConstruct
//
//  Synopsis:   Construct a CTrLstMsh from a block in memory
//
//  Arguments:  [lpObjMsh] -- memory block
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vConstruct(BYTE* lpObjMsh)
{
    BYTE* lpbData = lpObjMsh;

    // Meshes
    m_dwNumMeshes = *(DWORD*)lpbData;     lpbData += sizeof(DWORD);
    m_lpLstMeshes   = new TR123_MESH [ m_dwNumMeshes ];

    for(DWORD dwI = 0; dwI < m_dwNumMeshes; dwI++)
        {
        lpbData = CTRFile::lpbReadMesh(&m_lpLstMeshes[dwI],lpbData,lpbData, true);
        CTRFile::vComputeMeshExtents(&m_lpLstMeshes[dwI]);
        }
}


void	CTrLstMsh::vCreateMeshList( DWORD dwNumMeshes, TR123_MESH** lplpLstMeshes )
{
    // Copy the list of meshes
    TR123_MESH* lpLstMeshes = new TR123_MESH [ dwNumMeshes ];

    for(DWORD dwI = 0; dwI < dwNumMeshes; dwI++)
        {
        TR123_MESH* lpMesh = &lpLstMeshes[dwI];

        if(lplpLstMeshes[dwI])  // if NULL, the mesh is a dummy mesh
            {
            CTRFile::vCopyMesh(lplpLstMeshes[dwI],lpMesh);
            CTRFile::vComputeMeshExtents(lpMesh);
            lpMesh->boIsDummy = false;
            }
        else
            {
            // dummy mesh
            CTRFile::vInitializeMesh(lpMesh,0,0,0,false,true);
            CTRFile::vComputeMeshCollisionSphere(lpMesh);
            lpMesh->boIsDummy = true;
            }
        }

    m_dwNumMeshes   = dwNumMeshes;
    m_lpLstMeshes   = lpLstMeshes;
}


void	CTrLstMsh::vConvertLighting()
{
    for(DWORD dwI = 0; dwI < m_dwNumMeshes; dwI++)
        {
        TR123_MESH* lpMesh = &m_lpLstMeshes[dwI];

        // Convert lighting to TR3's lighting
        if( GetTrVersion() < Tr3_File  &&  lpMesh->Lighting.LightingType == MESH_LIGHTING_INTERNAL )
            {
            for(DWORD dwJ = 0; dwJ < lpMesh->Lighting.sNumData; dwJ++)
                {
                lpMesh->Lighting.lpLights[dwJ] = 8192 - lpMesh->Lighting.lpLights[dwJ];
                }
            }
        }
}


//----------------------------------------------------------------------------
//
//  Function:   vConstruct
//
//  Synopsis:   Construct a CTrLstMsh from a list of meshes
//
//  Arguments:  [lpTRFile] -- 
//              [dwNumMeshes] -- number of meshes
//              [lplpLstMeshes] -- list of meshes
//              [dwDesiredTextureWidth] -- desired width of the texture
//                (0 if no matter)
//              [dwDesiredTextureHeight] -- desired height of the texture
//                (0 if no matter)
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vConstruct(CTRFile* lpTRFile,DWORD dwNumMeshes,TR123_MESH** lplpLstMeshes,DWORD dwDesiredTextureWidth, DWORD dwDesiredTextureHeight)
{
	vCreateMeshList( dwNumMeshes, lplpLstMeshes );

    // Collect fragments
    DWORD   dwNumFragments = 0;

    CFragment*  lpLstFragments = NULL;

    CFragment::vCollectFragments(lpTRFile,m_lpLstMeshes,&lpLstFragments,&dwNumFragments,dwNumMeshes);

    // Construct CTrObjGfx
    CTrObjGfx::vConstruct(lpLstFragments,dwNumFragments,dwDesiredTextureWidth,dwDesiredTextureHeight);

    delete [] lpLstFragments;

	vConvertLighting();
}


void    CTrLstMsh::vConstruct(MATNAME2INFO* lpMapMat, DWORD dwNumMeshes, TR123_MESH* lpLstMeshes,DWORD dwDesiredTextureWidth, DWORD dwDesiredTextureHeight)
{
    // Collect fragments
    DWORD   dwNumFragments = 0;

    CFragment*  lpLstFragments;

    CFragment::vCollectFragments(lpMapMat,&lpLstFragments,&dwNumFragments);

    // Construct CTrObjGfx
    CTrObjGfx::vConstruct(lpLstFragments,dwNumFragments,dwDesiredTextureWidth,dwDesiredTextureHeight);

    delete [] lpLstFragments;

    m_dwNumMeshes   = dwNumMeshes;
    m_lpLstMeshes   = lpLstMeshes;
}


void    CTrLstMsh::vConstruct( const CTrLstMsh& MshSrc, DWORD dwNumMesh, DWORD dwDesiredTextureWidth, DWORD dwDesiredTextureHeight )
{
	CTrLstMsh	Mesh( MshSrc );

    // Collect fragments
    DWORD   dwNumFragments = 0;

    CFragment*  lpLstFragments = new CFragment[ Mesh.dwGetNumFragments() ];

    CFragment::vCollectFragments( &Mesh, lpLstFragments, dwNumFragments, dwNumMesh );

    // Construct CTrObjGfx
    CTrObjGfx::vConstruct(lpLstFragments,dwNumFragments,dwDesiredTextureWidth,dwDesiredTextureHeight);

    TR123_MESH** lplpLstMeshes = new TR123_MESH* [ 1 ];

	lplpLstMeshes[ 0 ] = const_cast<TR123_MESH*>( Mesh.lpGetMesh( dwNumMesh ) );

	vCreateMeshList( 1, lplpLstMeshes );

	delete [] lplpLstMeshes;

    delete [] lpLstFragments;
}


void    CTrLstMsh::vReplaceMesh( const CTrLstMsh& MshSrc, DWORD dwNumMesh, DWORD dwNumMeshReplaced )
{
	_ASSERTE( dwNumMeshReplaced < m_dwNumMeshes );

	if( dwNumMeshReplaced >= m_dwNumMeshes ) return;

	CTRFile::vReleaseMeshMemory( &m_lpLstMeshes[dwNumMeshReplaced] );

    // Collect fragments
	CTrLstMsh	Mesh( MshSrc );

    DWORD   dwNumFragments	= 0;
	DWORD	dwNumMaxFrag	= Mesh.dwGetNumFragments() + dwGetNumFragments();

    CFragment*  lpLstFragments = new CFragment[ dwNumMaxFrag ];

    CFragment::vCollectFragments( &Mesh, lpLstFragments, dwNumFragments, dwNumMesh );
    CFragment::vCollectFragments( this, lpLstFragments, dwNumFragments, dwNumMeshReplaced, true );

    // Construct CTrObjGfx
    CTrObjGfx	ObjGfx( lpLstFragments, dwNumFragments );
	
	CTrObjGfx::vStealData( ObjGfx );

    CTRFile::vCopyMesh( const_cast<TR123_MESH*>( Mesh.lpGetMesh( dwNumMesh ) ), &m_lpLstMeshes[dwNumMeshReplaced] );
    CTRFile::vComputeMeshExtents( &m_lpLstMeshes[dwNumMeshReplaced] );
}


//----------------------------------------------------------------------------
//
//  Function:   vComputeBoundingBox
//
//  Synopsis:   Compute the bounding box of a mesh
//
//  Arguments:  [dwNumMesh] -- Mesh to compute the bounding box. If -1, take
//               into account all the meshes
//
//  Return:     [BoundingBox] -- at return, holds the bounding box extents
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vComputeBoundingBox(DWORD dwNumMesh, CBoundingBox& BoundingBox, CMat4x4* lpMat)
{
    ASSERT(dwNumMesh == -1  ||  dwNumMesh < m_dwNumMeshes);

    BoundingBox.vSetBB(CVect3D((FLOAT)1e20,(FLOAT)1e20,(FLOAT)1e20),CVect3D((FLOAT)-1e20,(FLOAT)-1e20,(FLOAT)-1e20));

    if(dwNumMesh >= m_dwNumMeshes  &&  dwNumMesh != -1) return;
        
    float fMinX = 1e20f, fMaxX = -1e20f;
    float fMinY = 1e20f, fMaxY = -1e20f;
    float fMinZ = 1e20f, fMaxZ = -1e20f;

    for(DWORD dwI = 0; dwI < m_dwNumMeshes; dwI++)
        {
        if(dwNumMesh == -1  ||  dwNumMesh == dwI)
            {
            TR123_MESH* lpMesh = &m_lpLstMeshes[dwI];

			if( !lpMesh->boIsDummy )
				{
				short sNumVertices = lpMesh->sNumVertices;

				while(sNumVertices--)
					{
					TR2_VERTEX* lpVertex = &lpMesh->lpVertices[sNumVertices];
        
					float   fX =  (float)lpVertex->sX;
					float   fY = -(float)lpVertex->sY;
					float   fZ = -(float)lpVertex->sZ;

					if( lpMat )
						{
						CVect4D vec( fX, fY, fZ, -1.0 );

						vec = (*lpMat) * vec;

						fX = vec[0]; fY = vec[1]; fZ = vec[2];
						}

					if(fX < fMinX) fMinX = fX; else if(fX > fMaxX) fMaxX = fX;
					if(fY < fMinY) fMinY = fY; else if(fY > fMaxY) fMaxY = fY;
					if(fZ < fMinZ) fMinZ = fZ; else if(fZ > fMaxZ) fMaxZ = fZ;
					}
				}
			else
				{
				float   fX = 0;
				float   fY = 0;
				float   fZ = 0;

				if( lpMat )
					{
					CVect4D vec( fX, fY, fZ, -1.0 );

					vec = (*lpMat) * vec;

					fX = vec[0]; fY = vec[1]; fZ = vec[2];
					}

				if(fX < fMinX) fMinX = fX; else if(fX > fMaxX) fMaxX = fX;
				if(fY < fMinY) fMinY = fY; else if(fY > fMaxY) fMaxY = fY;
				if(fZ < fMinZ) fMinZ = fZ; else if(fZ > fMaxZ) fMaxZ = fZ;
				}
            }
        }

    BoundingBox.vSetBB(CVect3D(fMinX,fMinY,fMinZ),CVect3D(fMaxX,fMaxY,fMaxZ));
}


//----------------------------------------------------------------------------
//
//  Function:   vClearMesh
//
//  Synopsis:   Clear a mesh (ie. make an empty mesh)
//
//  Arguments:  [dwNumMesh] -- Mesh to clear
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vClearMesh(DWORD dwNumMesh)
{
    ASSERT(dwNumMesh < m_dwNumMeshes);

    if(dwNumMesh >= m_dwNumMeshes) return;

    TR123_MESH* lpMesh = &m_lpLstMeshes[dwNumMesh];

    CTRFile::vReleaseMeshMemory(lpMesh);

    lpMesh->Center.sX = lpMesh->Center.sY = lpMesh->Center.sZ = 0;
    lpMesh->lCollisionSize = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   vSetMeshes
//
//  Synopsis:   Replace the meshes with the given list
//
//  Arguments:  [dwNumMeshes] -- Number of meshes in the list
//              [lpLstMeshes] -- list of meshes
//              [lpMapMat] -- texturing infos for the meshes
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstMsh::vSetMeshes(DWORD dwNumMeshes, TR123_MESH* lpLstMeshes,MATNAME2INFO* lpMapMat)
{
    CTrLstMsh::vRelease();

    vConstruct(lpMapMat,dwNumMeshes,lpLstMeshes);
}


void	CTrLstMsh::vSetDummyMeshes( const CTrLstMsh& obj )
{
	for( DWORD dwI = 0; dwI < m_dwNumMeshes; ++dwI )
		{
		if( dwI < obj.m_dwNumMeshes  &&  obj.m_lpLstMeshes[ dwI ].boIsDummy )
			{
			CTRFile::vReleaseMeshMemory( &m_lpLstMeshes[ dwI ] );

            CTRFile::vInitializeMesh(&m_lpLstMeshes[ dwI ],0,0,0,false,true);

            CTRFile::vComputeMeshCollisionSphere(&m_lpLstMeshes[ dwI ]);

            m_lpLstMeshes[ dwI ].boIsDummy = true;
			}
		}
}

