#include "stdafx.h"

#include "Fragment.h"


//----------------------------------------------------------------------------
//
//  Function:   vInitialize
//
//  Synopsis:   Reset the members of the class
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstSpr::vInitialize()
{
    m_lpFragWorldPos = NULL;
}


//----------------------------------------------------------------------------
//
//  Function:   vRelease
//
//  Synopsis:   Release memory / Reset the members of the class
//
//  Arguments:  [boReleaseParent] -- true to release the parent class
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstSpr::vRelease(bool boReleaseParent /*= true*/)
{
    delete [] m_lpFragWorldPos;

    vInitialize();

    if(boReleaseParent) CTrObjGfx::vRelease(true);
}


//----------------------------------------------------------------------------
//
//  Function:   vStealData
//
//  Synopsis:   Transfer data from an object to the current one: after the
//              operation, the source object has lost his data member !
//
//  Arguments:  [obj] -- object to get the data from
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstSpr::vStealData(CTrLstSpr& obj)
{
    CTrObjGfx::vStealData(obj);

    m_lpFragWorldPos    = obj.m_lpFragWorldPos;

    obj.vInitialize();
}


//----------------------------------------------------------------------------
//
//  Function:   Constructor of class CTrLstSpr
//
//  Synopsis:   Construct a CTrLstSpr from a TR sprite sequence
//
//  Arguments:  [lpSprSeq] -- sprite sequence to convert
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrLstSpr::CTrLstSpr(TR2_SPRITE_SEQUENCE* lpSprSeq, CTRFile* lpTRFile) : CTrObjGfx(VERSION_TRSPRITE,TrObj_Type_Sprite,(enumTrVersion)lpTRFile->GetTRType())
{
    vInitialize();

    // Construct CTrLstSpr
    vConstruct(lpTRFile,lpSprSeq);
}


//----------------------------------------------------------------------------
//
//  Function:   Constructor of class CTrLstSpr
//
//  Synopsis:   Construct a CTrLstSpr from a block in memory
//
//  Arguments:  [lpSMsh] -- memory block
//              [boDeleteData] -- if true, delete the lpMVB pointer at the end
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrLstSpr::CTrLstSpr(BYTE* lpSpr,bool boDeleteData) throw(DWORD) : CTrObjGfx(lpSpr)
{
    vInitialize();

    if(lpSpr)
        {
        if(GetVersion() != VERSION_TRSPRITE) throw((DWORD)__LOADERR_MISMATCHVERSION__);

        BYTE* lpbData = lpSpr + CTrObjGfx::dwGetDataSize();

        vConstruct(lpbData);

        if(boDeleteData) delete [] lpSpr;
        }
}


//----------------------------------------------------------------------------
//
//  Function:   dwGetDataSize
//
//  Synopsis:   Get the size of the CTrLstSpr when put in a continuous block
//              of memory
//
//  Arguments:  -
//
//  Return:     size
//
//----------------------------------------------------------------------------
DWORD   CTrLstSpr::dwGetDataSize()
{
    DWORD   dwSize =    CTrObjGfx::dwGetDataSize() +
                        sizeof(WORLD_POSITIONNING) * m_dwNumFragText;   /* world positions */

    return  dwSize;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbGetData
//
//  Synopsis:   Convert a CTrLstSpr to a continuous block of memory
//
//  Arguments:  [dwDataSize] -- on return, hold the size of the block
//
//  Return:     pointer to the memory block
//
//  Caution:    you must make a delete [] on the pointer returned when you
//              have finished working with it !
//----------------------------------------------------------------------------
BYTE*   CTrLstSpr::lpbGetData(DWORD& dwDataSize)
{
    BYTE*   lpbData = CTrObjGfx::lpbGetData(dwDataSize);
    DWORD   dwTrLstMshDataSize = CTrObjGfx::dwGetDataSize();
    BYTE*   lpbDataSave = lpbData;

    lpbData += dwTrLstMshDataSize;

    // World positions
    memcpy(lpbData, m_lpFragWorldPos, sizeof(WORLD_POSITIONNING) * m_dwNumFragText);
    lpbData += sizeof(WORLD_POSITIONNING) * m_dwNumFragText;

    return lpbDataSave;
}


//----------------------------------------------------------------------------
//
//  Function:   boLoad
//
//  Synopsis:   Load a CTrLstSpr from disk
//
//  Arguments:  [cstrFileName] -- name of the file on disk
//
//  Return:     the object loaded & initialized or NULL if problems
//
//----------------------------------------------------------------------------
/*static*/ CTrLstSpr*  CTrLstSpr::objLoad(CString& cstrFileName)
{
    BYTE*   lpbData = CTrObjBase::lpbLoad(cstrFileName);

    CTrLstSpr* lpSpr = NULL;

try
    {
    lpSpr = new CTrLstSpr(lpbData,true);
    }
catch(DWORD)
    {
    delete lpSpr;
    lpSpr = NULL;
    }

    return lpSpr;
}


//----------------------------------------------------------------------------
//
//  Function:   vConstruct
//
//  Synopsis:   Construct a CTrLstSpr from a block in memory
//
//  Arguments:  [lpObjSpr] -- memory block
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstSpr::vConstruct(BYTE* lpObjSpr)
{
    BYTE* lpbData = lpObjSpr;

    // World positions
    m_lpFragWorldPos = new WORLD_POSITIONNING [ m_dwNumFragText ];

    memcpy(m_lpFragWorldPos, lpbData, sizeof(WORLD_POSITIONNING) * m_dwNumFragText);
    lpbData += sizeof(WORLD_POSITIONNING) * m_dwNumFragText;
}


//----------------------------------------------------------------------------
//
//  Function:   vConstruct
//
//  Synopsis:   Construct a CTrLstSpr from a TR sprite sequence
//
//  Arguments:  [lpSprSeq] -- sprite sequence to convert
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrLstSpr::vConstruct(CTRFile* lpTRFile, TR2_SPRITE_SEQUENCE* lpSprSeq)
{
    // Object Id
    vSetObjectId(lpSprSeq->lObjectId);

    // Collect sprite pictures
    DWORD   dwNumFragments = 0;

    TR_FILE& trFile = lpTRFile->GetTRFile();

    CFragment*  lpLstFragments = new CFragment [ -lpSprSeq->sNegativeLength ];

    TR2_SPRITE_TEXTURE* lpSprText = new TR2_SPRITE_TEXTURE [ -lpSprSeq->sNegativeLength ];

    for(DWORD dwI = 0; dwI < -lpSprSeq->sNegativeLength; dwI++)
        {
        TR2_SPRITE_TEXTURE* lpSprText_ = &lpTRFile->GetTRFile().tr123File.lpSpriteTextures[lpSprSeq->sOffset + dwI];

        lpSprText[dwI] = *lpSprText_;

        WORD wTile      = lpSprText[dwI].wTile;
        BYTE bX         = lpSprText[dwI].bX;
        BYTE bY         = lpSprText[dwI].bY;
        WORD wWidth     = (lpSprText[dwI].wWidth - 255) / 256;
        WORD wHeight    = (lpSprText[dwI].wHeight - 255) / 256;

		TR2_TEXTILE16*	lpTile = trFile.tr123File.lplpTextiles16[wTile];

        BYTE*   lpbTextureData = (BYTE*) lpTile->awTile;

        TR2_OBJECT_TEXTURE  ObjText;

		ZeroMemory( &ObjText, sizeof( ObjText ) );

        ObjText.wTile       = wTile;
        ObjText.wAttribute  = 0;

		if( lpTRFile->GetTRType() != TR4_FILE )
			{
			ObjText.Vertices[0].bXPixel = bX;           ObjText.Vertices[0].cXCoordinate = 1;
			ObjText.Vertices[0].bYPixel = bY;           ObjText.Vertices[0].cYCoordinate = 1;
			ObjText.Vertices[1].bXPixel = bX + wWidth;  ObjText.Vertices[1].cXCoordinate = -1;
			ObjText.Vertices[1].bYPixel = bY;           ObjText.Vertices[1].cYCoordinate = 1;
			ObjText.Vertices[2].bXPixel = bX + wWidth;  ObjText.Vertices[2].cXCoordinate = -1;
			ObjText.Vertices[2].bYPixel = bY + wHeight; ObjText.Vertices[2].cYCoordinate = -1;
			ObjText.Vertices[3].bXPixel = bX;           ObjText.Vertices[3].cXCoordinate = 1;
			ObjText.Vertices[3].bYPixel = bY + wHeight; ObjText.Vertices[3].cYCoordinate = -1;
			}
		else
			{
			ObjText.Vertices[0].bXPixel = lpSprText[dwI].sLeftSide;		ObjText.Vertices[0].cXCoordinate = 1;
			ObjText.Vertices[0].bYPixel = lpSprText[dwI].sTopSide;		ObjText.Vertices[0].cYCoordinate = 1;
			ObjText.Vertices[1].bXPixel = lpSprText[dwI].sRightSide-1;	ObjText.Vertices[1].cXCoordinate = -1;
			ObjText.Vertices[1].bYPixel = lpSprText[dwI].sTopSide;		ObjText.Vertices[1].cYCoordinate = 1;
			ObjText.Vertices[2].bXPixel = lpSprText[dwI].sRightSide-1;	ObjText.Vertices[2].cXCoordinate = -1;
			ObjText.Vertices[2].bYPixel = lpSprText[dwI].sBottomSide-1;	ObjText.Vertices[2].cYCoordinate = -1;
			ObjText.Vertices[3].bXPixel = lpSprText[dwI].sLeftSide;		ObjText.Vertices[3].cXCoordinate = 1;
			ObjText.Vertices[3].bYPixel = lpSprText[dwI].sBottomSide-1;	ObjText.Vertices[3].cYCoordinate = -1;
			}

        CFragment::vAddFragment(lpLstFragments, dwNumFragments, &ObjText, 4, OBJTEXT_TYPE_SPRITE, (void*)&lpSprText[dwI], false, lpbTextureData);
        }

    // Construct CTrObjGfx
    CTrObjGfx::vConstruct(lpLstFragments,dwNumFragments,0,0,false);

    // Construct the world positions array
    m_lpFragWorldPos = new WORLD_POSITIONNING [ -lpSprSeq->sNegativeLength ];

    for(dwI = 0; dwI < m_dwNumFragText; dwI++)
        {
        TR2_SPRITE_TEXTURE* lpSprText_ = (TR2_SPRITE_TEXTURE*) lpLstFragments[dwI].m_LstTextCoord.front().lpUserData;
        
        for(DWORD dwSpriteIndex = 0; dwSpriteIndex < -lpSprSeq->sNegativeLength; dwSpriteIndex++)
            {
            if(&lpSprText[dwSpriteIndex] == lpSprText_) break;
            }

		if( lpTRFile->GetTRType() != TR4_FILE )
			{
			m_lpFragWorldPos[dwSpriteIndex].sBottomSide	= lpSprText_->sBottomSide;
			m_lpFragWorldPos[dwSpriteIndex].sLeftSide   = lpSprText_->sLeftSide;
			m_lpFragWorldPos[dwSpriteIndex].sRightSide  = lpSprText_->sRightSide;
			m_lpFragWorldPos[dwSpriteIndex].sTopSide    = lpSprText_->sTopSide;
			m_lpFragWorldPos[dwSpriteIndex].wFragIndex  = dwI;
			m_lpFragWorldPos[dwSpriteIndex].bX			= 0;
			m_lpFragWorldPos[dwSpriteIndex].bY			= 0;
			}
		else
			{
			bit16	sWidth  = lpSprText_->sRightSide - lpSprText_->sLeftSide;
			bit16	sHeight = lpSprText_->sBottomSide - lpSprText_->sTopSide;

			m_lpFragWorldPos[dwSpriteIndex].sBottomSide = (sHeight / 2) *2;
			m_lpFragWorldPos[dwSpriteIndex].sLeftSide   = (-sWidth / 2) *2;
			m_lpFragWorldPos[dwSpriteIndex].sRightSide  = (sWidth / 2) *2;
			m_lpFragWorldPos[dwSpriteIndex].sTopSide    = (-sHeight / 2) *2;
			m_lpFragWorldPos[dwSpriteIndex].wFragIndex  = dwI;
			m_lpFragWorldPos[dwSpriteIndex].bX			= lpSprText_->bX;
			m_lpFragWorldPos[dwSpriteIndex].bY			= lpSprText_->bY;
			}
        }

    delete [] lpSprText;
    delete [] lpLstFragments;
}

