#include "stdafx.h"


//----------------------------------------------------------------------------
//
//  Function:   vInitialize
//
//  Synopsis:   Reset the members of the class
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
/*virtual*/ void    CTrObjBase::vInitialize()
{
    m_bVersion      = 0;
    m_trType        = TrObj_NoObj;
    m_trVersion     = Tr_UnknownVersion;
    m_bDummy        = 0;
    m_dwDirectory   = 0;
}


//----------------------------------------------------------------------------
//
//  Function:   vRelease
//
//  Synopsis:   Release memory / Reset the members of the class
//
//  Arguments:  [boReleaseParent] -- true to release the parent class
//
//  Return:     -
//
//----------------------------------------------------------------------------
/*virtual*/ void    CTrObjBase::vRelease(bool boReleaseParent /*= true*/)
{
    vInitialize();
}


//----------------------------------------------------------------------------
//
//  Function:   vStealData
//
//  Synopsis:   Transfer data from an object to the current one: after the
//              operation, the source object has lost his data member !
//
//  Arguments:  [obj] -- object to get the data from
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrObjBase::vStealData(CTrObjBase& obj)
{
    m_bVersion      = obj.m_bVersion;
    m_trType        = obj.m_trType;
    m_trVersion     = obj.m_trVersion;
    m_bDummy        = obj.m_bDummy;
    m_dwDirectory   = obj.m_dwDirectory;

    obj.vInitialize();
}


void	CTrObjBase::vCopy( const CTrObjBase& o )
{
    m_bVersion      = o.m_bVersion;
    m_trType        = o.m_trType;
    m_trVersion     = o.m_trVersion;
    m_bDummy        = o.m_bDummy;
    m_dwDirectory   = o.m_dwDirectory;
}


//----------------------------------------------------------------------------
//
//  Function:   Default constructor of class CTrObjBase
//
//  Synopsis:   Reset the members of the class
//
//  Arguments:  -
//
//  Return:     -
//
//----------------------------------------------------------------------------
CTrObjBase::CTrObjBase(BYTE* lpObj)
{
    vInitialize();

    if(lpObj) vConstruct(lpObj);
}


CTrObjBase::CTrObjBase( const CTrObjBase& o )
{
	vCopy( o );
}


//----------------------------------------------------------------------------
//
//  Function:   dwGetDataSize
//
//  Synopsis:   Get the size of the CTrObjBase when put in a continuous block
//              of memory
//
//  Arguments:  -
//
//  Return:     size
//
//----------------------------------------------------------------------------
/*virtual*/ DWORD   CTrObjBase::dwGetDataSize()
{
    DWORD   dwSize =    sizeof(BYTE) +              /* m_bVersion */
                        sizeof(BYTE) +              /* m_trType */
                        sizeof(BYTE) +              /* m_trVersion */
                        sizeof(BYTE) +              /* m_bDummy */
                        sizeof(DWORD);              /* m_dwDirectory */

    return dwSize;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbGetData
//
//  Synopsis:   Convert a CTrObjBase to a continuous block of memory
//
//  Arguments:  [dwDataSize] -- on return, hold the size of the block
//
//  Return:     pointer to the memory block
//
//  Caution:    you must make a delete [] on the pointer returned when you
//              have finished working with it !
//----------------------------------------------------------------------------
/*virtual*/BYTE*   CTrObjBase::lpbGetData(DWORD& dwDataSize)
{
    dwDataSize = dwGetDataSize();

    BYTE*   lpbData = new BYTE [ dwDataSize ];
    BYTE*   lpbDataSave = lpbData;

    *lpbData            = m_bVersion;           lpbData += sizeof(BYTE);
    *lpbData            = (BYTE)m_trType;       lpbData += sizeof(BYTE);
    *lpbData            = (BYTE)m_trVersion;    lpbData += sizeof(BYTE);
    *lpbData            = 0;                    lpbData += sizeof(BYTE);
    *(DWORD*)lpbData    = m_dwDirectory;        lpbData += sizeof(DWORD);

    return lpbDataSave;
}


//----------------------------------------------------------------------------
//
//  Function:   boSave
//
//  Synopsis:   Save the CTrObjBase to disk
//
//  Arguments:  [cstrFileName] -- name of the file on disk
//
//  Return:     -
//
//----------------------------------------------------------------------------
bool    CTrObjBase::boSave(CString& cstrFileName)
{
    DWORD   dwDataSize;
    BYTE*   lpbData = lpbGetData(dwDataSize);

    DWORD   dwNumWritten;

    HANDLE hFile = CreateFile(cstrFileName,GENERIC_WRITE,0,NULL,CREATE_ALWAYS,FILE_ATTRIBUTE_ARCHIVE,NULL);
    WriteFile(hFile, lpbData, dwDataSize, &dwNumWritten, NULL);
    CloseHandle(hFile);

    delete [] lpbData;

    return true;
}


//----------------------------------------------------------------------------
//
//  Function:   lpbLoad
//
//  Synopsis:   Load a CTrObjBase from disk
//
//  Arguments:  [cstrFileName] -- name of the file on disk
//
//  Return:     pointer on the loaded data
//
//----------------------------------------------------------------------------
/*static*/ BYTE*   CTrObjBase::lpbLoad(CString& cstrFileName)
{
    DWORD   dwNumRead;
    HANDLE hFile = CreateFile(cstrFileName,GENERIC_READ,0,NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

    DWORD dwDataSize = GetFileSize(hFile,NULL);
    BYTE* lpData = new BYTE[dwDataSize];

    ReadFile(hFile,lpData,dwDataSize,&dwNumRead,NULL);

    CloseHandle(hFile);

    return lpData;
}


//----------------------------------------------------------------------------
//
//  Function:  vConstruct
//
//  Synopsis:   Construct a CTrObjBase from a block in memory
//
//  Arguments:  [lpObjBase] -- memory block
//
//  Return:     -
//
//----------------------------------------------------------------------------
void    CTrObjBase::vConstruct(BYTE* lpObjBase)
{
    m_bVersion     = *lpObjBase;                    lpObjBase += sizeof(BYTE);
    m_trType       = (enumTrObjType)(*lpObjBase);   lpObjBase += sizeof(BYTE);
    m_trVersion    = (enumTrVersion)(*lpObjBase);   lpObjBase += sizeof(BYTE);
    m_bDummy       = 0;                             lpObjBase += sizeof(BYTE);
    m_dwDirectory  = *(DWORD*)lpObjBase;            lpObjBase += sizeof(DWORD);
}
