#include "StdAfx.h"

#include "TrSndMgr.h"


CTrSndMgr::CTrSnd::CTrSnd(TR2_SOUND_DETAILS* lpSoundDetail, DWORD dwFirstSoundIndex, DWORD dwInternalIndex)
{
    m_dwInternalIndex	= dwInternalIndex;
    m_dwFirstSoundIndex = dwFirstSoundIndex;
    m_lVolume           = lpSoundDetail->sVolume;
    m_wNumSounds        = (lpSoundDetail->sUnknown2 & 0x00FC) >> 2;
    m_sUnknown1         = lpSoundDetail->sUnknown1;
    m_sUnknown2         = lpSoundDetail->sUnknown2;
	m_sSample			= lpSoundDetail->sSample;
}


CTrSndMgr::CTrSnd::CTrSnd(BYTE* lpbData,DWORD dwInternalIndex)
{
    m_dwInternalIndex   = dwInternalIndex;
    m_dwFirstSoundIndex = * (DWORD*)lpbData;            lpbData += sizeof(DWORD);
    m_lVolume           = * (long*) lpbData;            lpbData += sizeof(long);
    m_wNumSounds        = * (WORD*) lpbData;            lpbData += sizeof(WORD);
    m_sUnknown1         = * (short*)lpbData;            lpbData += sizeof(short);
    m_sUnknown2         = * (short*)lpbData;            lpbData += sizeof(short);
	m_sSample			= 0;
}


CTrSndMgr::CTrSnd::CTrSnd(CTrSnd* lpSnd)
{
    m_dwInternalIndex   = lpSnd->m_dwInternalIndex;
    m_dwFirstSoundIndex = lpSnd->m_dwFirstSoundIndex;
    m_wNumSounds        = lpSnd->m_wNumSounds;
    m_lVolume           = lpSnd->m_lVolume;
    m_sUnknown1         = lpSnd->m_sUnknown1;
    m_sUnknown2         = lpSnd->m_sUnknown2;
	m_sSample			= lpSnd->m_sSample;
}


DWORD   CTrSndMgr::CTrSnd::dwGetDataSize()
{
    return  sizeof(DWORD) +         // FirstSoundIndex
            sizeof(long) +          // Volume
            sizeof(WORD) +          // Num sounds
            sizeof(short) +         // Unknown1
            sizeof(short);          // Unknown2
}


void    CTrSndMgr::CTrSnd::vSerialize(BYTE* lpbData)
{
    *(DWORD*)lpbData = m_dwFirstSoundIndex;				lpbData += sizeof(DWORD);
    *(long*) lpbData = m_lVolume;                       lpbData += sizeof(long);
    *(WORD*) lpbData = m_wNumSounds;                    lpbData += sizeof(WORD);
    *(short*)lpbData = m_sUnknown1;                     lpbData += sizeof(short);
    *(short*)lpbData = m_sUnknown2;                     lpbData += sizeof(short);
}


void    CTrSndMgr::CTrSnd::vConvertToTR(TR2_SOUND_DETAILS* lpSoundDetail, DWORD dwSampleIndice)
{
    lpSoundDetail->sSample      = dwSampleIndice;
    lpSoundDetail->sUnknown1    = m_sUnknown1;
    lpSoundDetail->sUnknown2    = m_sUnknown2;
    lpSoundDetail->sVolume      = m_lVolume;
}


CTrSndMgr::CTrSndMgr(CTRFile* lpTRFile) : m_SoundMap()
{
    TR_FILE& trFile = lpTRFile->GetTRFile();

    DWORD   dwMaxInternalIndex = (lpTRFile->GetTRType() == TR1_FILE) ? sizeof(TR1_SOUNDMAP) / sizeof(bit16) : sizeof(TR2_SOUNDMAP) / sizeof(bit16);

    // Build the internal map according to sounds found in the level
    for(DWORD dwInternalIndex = 0; dwInternalIndex < dwMaxInternalIndex; dwInternalIndex++)
        {
        bit16   sSoundDetail = trFile.tr123File.SoundMap.tr2.aData[dwInternalIndex];

        if(sSoundDetail == -1) continue;

        TR2_SOUND_DETAILS*  lpSoundDetail = &trFile.tr123File.lpSoundDetails[sSoundDetail];

        m_SoundMap[dwInternalIndex] = new CTrSnd(lpSoundDetail, trFile.tr123File.lpSampleIndices[lpSoundDetail->sSample],dwInternalIndex);
        }
}


CTrSndMgr::CTrSndMgr(BYTE* lpbData) : m_SoundMap()
{
    DWORD   dwNumSounds = *(DWORD*) lpbData;            lpbData += sizeof(DWORD);

    for(DWORD dwSound = 0; dwSound < dwNumSounds; dwSound++)
        {
        DWORD dwInternalIndex = *(DWORD*) lpbData;      lpbData += sizeof(DWORD);

        m_SoundMap[dwInternalIndex] = new CTrSnd(lpbData,dwInternalIndex);

        lpbData += CTrSnd::dwGetDataSize();
        }
}


DWORD   CTrSndMgr::dwGetDataSize()
{
    return (CTrSnd::dwGetDataSize() + sizeof(DWORD)) * m_SoundMap.size() + sizeof(DWORD);
}


BYTE*   CTrSndMgr::lpbGetData(DWORD& dwDataSize)
{
    dwDataSize = dwGetDataSize();

    BYTE*   lpbData = new BYTE [ dwDataSize ];
    BYTE*   lpbDataSave = lpbData;

    *(DWORD*) lpbData = m_SoundMap.size();                  lpbData += sizeof(DWORD);

    long    lPrevInternalIndex = -1;

    for(MapIdx2Snd::iterator theIter = m_SoundMap.begin(); theIter != m_SoundMap.end(); theIter++)
        {
        DWORD   dwInternalIndex = theIter->first;
        CTrSnd* lpSnd = theIter->second;

        ASSERT((long)dwInternalIndex > lPrevInternalIndex);

        *(DWORD*) lpbData = dwInternalIndex;                lpbData += sizeof(DWORD);
        lpSnd->vSerialize(lpbData);

        lpbData += lpSnd->dwGetDataSize();
        lPrevInternalIndex = dwInternalIndex;
        }

    return lpbDataSave;
}


CTrSndMgr::~CTrSndMgr()
{
    // Release the memory used by the map
    for(MapIdx2Snd::iterator theIter = m_SoundMap.begin(); theIter != m_SoundMap.end(); theIter++)
        {
        delete theIter->second;
        }

    m_SoundMap.clear();
}


bool    CTrSndMgr::boHaveSound(DWORD dwInternalIndex)
{
    MapIdx2Snd::iterator theIter = m_SoundMap.find(dwInternalIndex);

    if(theIter != m_SoundMap.end()) return true;

    return false;
}


bool    CTrSndMgr::boAddSound(DWORD dwInternalIndex, TR2_SOUND_DETAILS* lpSoundDetail, DWORD dwFirstSoundIndex)
{
    MapIdx2Snd::iterator theIter = m_SoundMap.find(dwInternalIndex);

    // This sound is already in the map ?
    if(theIter != m_SoundMap.end()) return false;

    // Add the sound to the map
    m_SoundMap[dwInternalIndex] = new CTrSnd(lpSoundDetail, dwFirstSoundIndex, dwInternalIndex);

    return true;
}


bool    CTrSndMgr::boAddSound(CTrSnd* lpSnd)
{
    MapIdx2Snd::iterator theIter = m_SoundMap.find(lpSnd->m_dwInternalIndex);

    // This sound is already in the map ?
    if(theIter != m_SoundMap.end()) return false;

    // Add the sound to the map
    m_SoundMap[lpSnd->m_dwInternalIndex] = lpSnd;

    return true;
}


bool    CTrSndMgr::boCopySound(CTrSndMgr* lpSndMgrSrc, DWORD dwInternalIndex)
{
    MapIdx2Snd::iterator theIter = lpSndMgrSrc->m_SoundMap.find(dwInternalIndex);

    // This sound is already in the map ?
    if(theIter == m_SoundMap.end()) return false;

    return boAddSound(new CTrSnd(theIter->second));
}


void    CTrSndMgr::vRemoveSound(DWORD dwInternalIndex)
{
    MapIdx2Snd::iterator theIter = m_SoundMap.find(dwInternalIndex);

    // This sound is in the map ?
    if(theIter == m_SoundMap.end()) return;

    delete theIter->second;

    m_SoundMap.erase(theIter);
}


void    CTrSndMgr::vMerge(CTrSndMgr* lpSndMgr,bool boDelete)
{
    for(MapIdx2Snd::iterator theIter = lpSndMgr->m_SoundMap.begin(); theIter != lpSndMgr->m_SoundMap.end(); theIter++)
        {
        DWORD   dwInternalIndex = theIter->first;
        CTrSnd* lpSnd = theIter->second;

        if(!boDelete)
            {
            MapIdx2Snd::iterator theIter = m_SoundMap.find(dwInternalIndex);

            if(theIter != m_SoundMap.end()) continue;

            m_SoundMap[dwInternalIndex] = new CTrSnd(lpSnd);
            }
        else
            {
            vRemoveSound(dwInternalIndex);
            }
        }
}


void    CTrSndMgr::vConvertToTR(CTRFile* lpTRFile)
{
    TR_FILE& trFile = lpTRFile->GetTRFile();

    delete [] trFile.tr123File.lpSampleIndices;
    delete [] trFile.tr123File.lpSoundDetails;

    DWORD   dwNumSampleIndices  = 0;
    DWORD   dwNumSoundDetails   = m_SoundMap.size();

    TR2_SOUND_DETAILS*  lpSoundDetails = new TR2_SOUND_DETAILS [ dwNumSoundDetails ];

    DWORD   dwMaxInternalIndex = (lpTRFile->GetTRType() == TR1_FILE) ? sizeof(TR1_SOUNDMAP) / sizeof(bit16) : sizeof(TR2_SOUNDMAP) / sizeof(bit16);

    // Clear the sound map
    for(DWORD dwSound = 0; dwSound < dwMaxInternalIndex; dwSound++)
        {
        trFile.tr123File.SoundMap.tr2.aData[dwSound] = -1;
        }

    // Calculate the number of samples indices
    for(MapIdx2Snd::iterator theIter = m_SoundMap.begin(); theIter != m_SoundMap.end(); theIter++)
        {
        CTrSnd* lpSnd = theIter->second;

        dwNumSampleIndices += lpSnd->dwGetNumSounds();
        }

    bitu32* lpSampleIndices = new bitu32 [ dwNumSampleIndices ];

    // Convert to TR structures
    DWORD   dwSoundDetail   = 0;
    DWORD   dwSampleIndice  = 0;

    long    lPrevInternalIndex = -1;

    for(theIter = m_SoundMap.begin(); theIter != m_SoundMap.end(); theIter++, dwSoundDetail++)
        {
        DWORD   dwInternalIndex = theIter->first;
        CTrSnd* lpSnd = theIter->second;

        ASSERT((long)dwInternalIndex > lPrevInternalIndex);
        ASSERT(dwInternalIndex < dwMaxInternalIndex);

        TR2_SOUND_DETAILS*  lpSoundDetail = &lpSoundDetails[dwSoundDetail];

        trFile.tr123File.SoundMap.tr2.aData[dwInternalIndex] = dwSoundDetail;

        lpSnd->vConvertToTR(lpSoundDetail, dwSampleIndice);

        for(DWORD dwIndex = 0; dwIndex < lpSnd->dwGetNumSounds(); dwIndex++)
            {
            lpSampleIndices[ dwSampleIndice++ ] = lpSnd->dwGetFirstSoundIndex() + dwIndex;
            }

        lPrevInternalIndex = dwInternalIndex;
        }

    ASSERT(dwSampleIndice == dwNumSampleIndices);
    ASSERT(dwSoundDetail == dwNumSoundDetails);

    // Set the new data members for the level
    trFile.tr123File.dwNumSoundDetails  = dwNumSoundDetails;
    trFile.tr123File.lpSoundDetails     = lpSoundDetails;

    trFile.tr123File.dwNumSampleIndices = dwNumSampleIndices;
    trFile.tr123File.lpSampleIndices    = lpSampleIndices;
}
