/**********************************************************************************************************************/
/*                                                VECTORS & MATRICES CLASSES                                          */
/* 17/09/97, M : creation of the file                                                                                 */
/* 25/12/97, M : some functions added from BMATH                                                                      */
/* 17/02/98, M : MIN/MAX macros                                                                                       */
/* 18/02/98, M : v3ModifyAxis                                                                                         */
/* 14/12/98, M : inlining bugfixed !                                                                                  */
/* 14/10/99, P : Set fct added for optims                                                                             */
/* 25/11/99, M : rewriting for operators speed optims                                                                 */
/* 09/06/00, M : include optim, const                                                                                 */
/**********************************************************************************************************************/

#ifndef   _CVECTMAT_H_
#define   _CVECTMAT_H_

/************************************************ INCLUDE *************************************************************/

#include  <strstrea.h>                                      // ostrstream

/************************************************ DEFINE VALUES *******************************************************/

#define   _DANGEROUS_CAST_ 1                                // allow dangerous casts
#undef    _DANGEROUS_CAST_                                  // ...not

enum      { _X_,_Y_,_Z_,_W_ };                              // 4D axis
enum      { _R_,_G_,_B_,_A_ };                              // 4D color
enum      { _U_,_V_,_Q_ };                                  // 3D tex coord

#define   _PI_                          3.141592654
#define   _EPS_NORMALIZE_               0.001

typedef unsigned char SBYTE;

#define FLOAT double

/************************************************ MACROS **************************************************************/

#define   _RAD2DEG_(fAngleRad)          (fAngleRad*180.0/_PI_)
#define   _DEG2RAD_(fAngleDeg)          (fAngleDeg*_PI_/180.0)

#define   _V2MIN_(v1,v2)                { if(v2[_X_] < v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] < v1[_Y_]) v1[_Y_] = v2[_Y_]; }
#define   _V2MAX_(v1,v2)                { if(v2[_X_] > v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] > v1[_Y_]) v1[_Y_] = v2[_Y_]; }

#define   _V3MIN_(v1,v2)                { if(v2[_X_] < v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] < v1[_Y_]) v1[_Y_] = v2[_Y_]; \
                                          if(v2[_Z_] < v1[_Z_]) v1[_Z_] = v2[_Z_]; }
#define   _V3MAX_(v1,v2)                { if(v2[_X_] > v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] > v1[_Y_]) v1[_Y_] = v2[_Y_]; \
                                          if(v2[_Z_] > v1[_Z_]) v1[_Z_] = v2[_Z_]; }

#define   _V4MIN_(v1,v2)                { if(v2[_X_] < v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] < v1[_Y_]) v1[_Y_] = v2[_Y_]; \
                                          if(v2[_Z_] < v1[_Z_]) v1[_Z_] = v2[_Z_]; \
                                          if(v2[_W_] < v1[_W_]) v1[_W_] = v2[_W_]; }
#define   _V4MAX_(v1,v2)                { if(v2[_X_] > v1[_X_]) v1[_X_] = v2[_X_]; \
                                          if(v2[_Y_] > v1[_Y_]) v1[_Y_] = v2[_Y_]; \
                                          if(v2[_Z_] > v1[_Z_]) v1[_Z_] = v2[_Z_]; \
                                          if(v2[_W_] > v1[_W_]) v1[_W_] = v2[_W_]; }

/************************************************ CLASSES *************************************************************/

class     CVect2D;                                          // 2D vector
class     CVect3D;                                          // 3D vector
class     CVect4D;                                          // 4D vector
class     CMat3x3;                                          // 3*3 matrix
class     CMat4x4;                                          // 4*4 matrix

/**********************************************************************************************************************/
/*                                                2D VECTOR                                                           */
/**********************************************************************************************************************/

class     CVect2D
  {
// public methods

  public:

    // constructors

    inline          CVect2D             (void);
    inline          CVect2D             (const FLOAT    fD);                              // x=y=fD
    inline          CVect2D             (const FLOAT    fX, const FLOAT fY);              // x=fX; y=fY
    inline          CVect2D             (const CVect2D& v2V);                             // copy 
    inline          CVect2D             (const CVect3D& v3V,const BYTE bDropAxis);        // cast v3 to v2 (drop an axis)
  #ifdef _DANGEROUS_CAST_
    inline          CVect2D             (const CVect3D& v3V);                             // cast v3 to v2 (perspective)
  #endif

    // assignment operators

    CVect2D&               operator   = (const CVect2D& v2V);                             // assignment
    CVect2D&               operator  += (const CVect2D& v2V);                             // incrementation
    CVect2D&               operator  -= (const CVect2D& v2V);                             // decrementation
    CVect2D&               operator  *= (const FLOAT    fK);                              // multiplication by a constant
    CVect2D&               operator  *= (const CVect2D& v2V);                             // term by term *
    CVect2D&               operator  /= (const FLOAT    fK);                              // division by a constant
    CVect2D&               operator  /= (const CVect2D& v2V);                             // term by term /
    FLOAT&                 operator  [] (const DWORD    dwN);                             // indexing
    const FLOAT            operator  [] (const DWORD    dwN) const;                       // indexing                            

    // friends operators

    friend CVect2D  inline operator   - (const CVect2D& v2V);                             // -V
    friend CVect2D  inline operator   * (const FLOAT    fK,  const CVect2D& v2V);         // k*V 
    friend CVect2D  inline operator   * (const CVect2D& v2V, const FLOAT    fK);          // V*k
    friend CVect2D  inline operator   / (const CVect2D& v2V, const FLOAT    fK);          // V/k
    friend CVect2D  inline operator   - (const CVect2D& v2V1,const CVect2D& v2V2);        // V1-V2
    friend CVect2D  inline operator   + (const CVect2D& v2V1,const CVect2D& v2V2);        // V1+V2
    friend FLOAT    inline operator   * (const CVect2D& v2V1,const CVect2D& v2V2);        // V1*V2 (dot   product)
    friend FLOAT    inline operator   ^ (const CVect2D& v2V1,const CVect2D& v2V2);        // V1^V2 (cross product)
    friend bool     inline operator  == (const CVect2D& v2V1,const CVect2D& v2V2);        // V1==V2
    friend bool     inline operator  != (const CVect2D& v2V1,const CVect2D& v2V2);        // V1!=V2

    friend ostream&        operator  << (ostream& s,      CVect2D& v2V);                  // output
    friend ostream&        operator  << (ostream& s,const CVect2D& v2V);                  // output

    // friends functions

    friend void     inline vSwap        (      CVect2D& v2V1,      CVect2D& v2V2);        // swap
    friend CVect2D  inline v2Mini       (const CVect2D& v2V1,const CVect2D& v2V2);        // mini(V1,V2) (smallest vector)
    friend CVect2D  inline v2Maxi       (const CVect2D& v2V1,const CVect2D& v2V2);        // mini(V1,V2) (longest  vector) 
    friend CVect2D  inline v2Prod       (const CVect2D& v2V1,const CVect2D& v2V2);        // term by term *
    friend CVect2D  inline v2Div        (const CVect2D& v2V1,const CVect2D& v2V2);        // term by term /

    // special functions

    FLOAT  inline   fLength             (void) const;                                     // vector length d
    FLOAT  inline   fLength2            (void) const;                                     // vector squared length d^2
    bool   inline   boNormalize         (const FLOAT fNorm = 1.0f);                       // normalize
    bool   inline   boNormalize1        (void);                                           // normalize to 1.f - dedicated version

    // set

    inline void     vSet                (const FLOAT    fD);
    inline void     vSet                (const FLOAT    fX, const FLOAT fY);
    inline void     vSet                (const CVect2D& v2V);
    inline void     vSet                (const CVect3D& v3V,const BYTE  bDropAxis);

    // friends classes

    friend class    CVect3D;

// protected members

  protected:

    FLOAT           m_fV[2];                                                              // x/y coords
  };

/**********************************************************************************************************************/
/*                                                3D VECTOR                                                           */
/**********************************************************************************************************************/

class     CVect3D
  {
// public methods

  public:

    // constructors

    inline          CVect3D             (void);                                               
    inline          CVect3D             (const FLOAT    fD);                              // x=y=z=fD
    inline          CVect3D             (const FLOAT    fX, const FLOAT fY,const FLOAT fZ);// x=fX; y=fY; z=fZ
    inline          CVect3D             (const CVect3D& v3V);                             // copy 
    inline          CVect3D             (const CVect4D& v4V,const BYTE  bDropAxis);       // cast v4 to v3 (drop an axis)
    inline          CVect3D             (const CVect2D& v2V,const FLOAT fZ);              // cast v2 to v3; z=fZ
  #ifdef _DANGEROUS_CAST_
    inline          CVect3D             (const CVect4D& v4V);                             // cast v4 to v3 (perspective)
    inline          CVect3D             (const CVect2D& v2V);                             // cast v2 to v3; z=1.0
  #endif

    // assignment operators

    CVect3D&               operator   = (const CVect3D& v3V);                             // assignment
    CVect3D&               operator  += (const CVect3D& v3V);                             // incrementation
    CVect3D&               operator  -= (const CVect3D& v3V);                             // decrementation
    CVect3D&               operator  *= (const FLOAT    fK);                              // multiplication by a constant
    CVect3D&               operator  *= (const CVect3D& v3V);                             // term by term *
    CVect3D&               operator  /= (const FLOAT    fK);                              // division by a constant
    CVect3D&               operator  /= (const CVect3D& v3V);                             // term by term /
    FLOAT&                 operator  [] (const DWORD    dwN);                             // indexing
    const FLOAT            operator  [] (const DWORD    dwN) const;                       // indexing

    // friends operators

    friend CVect3D  inline operator   - (const CVect3D& v3V);                             // -V
    friend CVect3D  inline operator   * (const FLOAT    fK,  const CVect3D& v3V);         // k*V 
    friend CVect3D  inline operator   * (const CVect3D& v3V, const FLOAT    fK);          // V*k
    friend CVect3D  inline operator   / (const CVect3D& v3V, const FLOAT    fK);          // V/k
    friend CVect3D  inline operator   - (const CVect3D& v3V1,const CVect3D& v3V2);        // V1-V2
    friend CVect3D  inline operator   + (const CVect3D& v3V1,const CVect3D& v3V2);        // V1+V2
    friend FLOAT    inline operator   * (const CVect3D& v3V1,const CVect3D& v3V2);        // V1*V2 (dot   product)
    friend CVect3D  inline operator   ^ (const CVect3D& v3V1,const CVect3D& v3V2);        // V1^V2 (cross product)
    friend bool     inline operator  == (const CVect3D& v3V1,const CVect3D& v3V2);        // V1==V2
    friend bool     inline operator  != (const CVect3D& v3V1,const CVect3D& v3V2);        // V1!=V2
    friend CVect3D  inline operator   * (const CMat3x3& m3M, const CVect3D& v3V);         // M3*V3
    friend CVect3D  inline operator   * (const CVect3D& v3V, const CMat3x3& m3M);         // V3*M3

    friend CVect4D  inline operator   * (const CMat4x4& m4M, const CVect3D& v3V);         // M4*V3
    friend CVect4D  inline operator   * (const CVect3D& v3V, const CMat4x4& m4M);         // V3*M4
    friend FLOAT    inline operator   * (const CVect4D& v4V, const CVect3D& v3V);         // V4*V3 (dot product)
    friend FLOAT    inline operator   * (const CVect3D& v3V, const CVect4D& v4V);         // V3*V4 (dot product)

    friend ostream&        operator  << (ostream& s,      CVect3D& v3V);                  // output
    friend ostream&        operator  << (ostream& s,const CVect3D& v3V);                  // output

    // friends functions

    friend void     inline vSwap        (      CVect3D& v3V1,      CVect3D& v3V2);        // swap
    friend CVect3D  inline v3Mini       (const CVect3D& v3V1,const CVect3D& v3V2);        // mini(V1,V2) (smallest vector) 
    friend CVect3D  inline v3Maxi       (const CVect3D& v3V1,const CVect3D& v3V2);        // mini(V1,V2) (longest  vector) 
    friend CVect3D  inline v3Prod       (const CVect3D& v3V1,const CVect3D& v3V2);        // term by term *
    friend CVect3D  inline v3Div        (const CVect3D& v3V1,const CVect3D& v3V2);        // term by term /
    friend CVect3D  inline v3ModifyAxis (      CVect3D& v3V, const SBYTE bX,const SBYTE bY, const SBYTE bZ);       
  
    // special functions

    FLOAT  inline   fLength             (void) const;                                     // vector length d
    FLOAT  inline   fLength2            (void) const;                                     // vector squared length d^2
    bool   inline   boNormalize         (const FLOAT fNorm = 1.0f);                       // normalize
    bool   inline   boNormalize1        (void);                                           // normalize to 1.f - dedicated version

    // set

    inline void     vSet                (const FLOAT    fD);
    inline void     vSet                (const FLOAT    fX, const FLOAT fY,const FLOAT fZ);
    inline void     vSet                (const CVect3D& v3V);
    inline void     vSet                (const CVect4D& v4V,const BYTE  bDropAxis);
    inline void     vSet                (const CVect2D& v2V,const FLOAT fZ);

    // friends classes

    friend class    CVect2D;
    friend class    CVect4D;
    friend class    CMat3x3;
    friend class    CMat4x4;

    friend CMat3x3  inline operator  *  (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1*M2 accesses fV

// protected members

  protected:

    FLOAT           m_fV[3];                                                              // x/y/z coords
  };

/**********************************************************************************************************************/
/*                                                4D VECTOR                                                           */
/**********************************************************************************************************************/

class     CVect4D
  {
// public methods

  public:

    // constructors

    inline          CVect4D             (void);
    inline          CVect4D             (const FLOAT    fD);                              // x=y=z=w=fD
    inline          CVect4D             (const FLOAT    fX, const FLOAT fY,const FLOAT fZ,const FLOAT fW);        
                                                                                          // x=fX; y=fY; z=fZ; w=fW
    inline          CVect4D             (const CVect4D& v4V);                             // copy 
    inline          CVect4D             (const CVect3D& v3V,const FLOAT fW);              // cast v3 to v4; w=fW
  #ifdef _DANGEROUS_CAST_
    inline          CVect4D             (const CVect3D& v3V);                             // cast v3 to v4; w=1.0
  #endif

    // assignment operators

    CVect4D&               operator   = (const CVect4D& v4V);                             // assignment
    CVect4D&               operator  += (const CVect4D& v4V);                             // incrementation
    CVect4D&               operator  -= (const CVect4D& v4V);                             // decrementation
    CVect4D&               operator  *= (const FLOAT    fK);                              // multiplication by a constant
    CVect4D&               operator  *= (const CVect4D& v4V);                             // term by term *
    CVect4D&               operator  /= (const FLOAT    fK);                              // division by a constant
    CVect4D&               operator  /= (const CVect4D& v4V);                             // term by term /
    FLOAT&                 operator  [] (const DWORD    dwN);                             // indexing
    const FLOAT            operator  [] (const DWORD    dwN) const;                       // indexing

    // friends operators

    friend CVect4D  inline operator   - (const CVect4D& v4V);                             // -V
    friend CVect4D  inline operator   * (const FLOAT    fK,  const CVect4D& v4V);         // k*V 
    friend CVect4D  inline operator   * (const CVect4D& v4V, const FLOAT    fK);          // V*k
    friend CVect4D  inline operator   / (const CVect4D& v4V, const FLOAT    fK);          // V/k
    friend CVect4D  inline operator   - (const CVect4D& v4V1,const CVect4D& v4V2);        // V1-V2
    friend CVect4D  inline operator   + (const CVect4D& v4V1,const CVect4D& v4V2);        // V1+V2
    friend FLOAT    inline operator   * (const CVect4D& v4V1,const CVect4D& v4V2);        // V1*V2 (dot product)
    friend bool     inline operator  == (const CVect4D& v4V1,const CVect4D& v4V2);        // V1==V2
    friend bool     inline operator  != (const CVect4D& v4V1,const CVect4D& v4V2);        // V1!=V2
    friend CVect4D  inline operator   * (const CMat4x4& m4M, const CVect4D& v4V);         // M*V
    friend CVect4D  inline operator   * (const CVect4D& v4V, const CMat4x4& m4M);         // V*M

    friend ostream&        operator  << (ostream& s,      CVect4D& v4V);                  // output
    friend ostream&        operator  << (ostream& s,const CVect4D& v4V);                  // output

    // friends functions

    friend void     inline vSwap        (      CVect4D& v4V1,      CVect4D& v4V2);        // swap
    friend CVect4D  inline v4Mini       (const CVect4D& v4V1,const CVect4D& v4V2);        // mini(V1,V2) (smallest vector) 
    friend CVect4D  inline v4Maxi       (const CVect4D& v4V1,const CVect4D& v4V2);        // mini(V1,V2) (longest  vector) 
    friend CVect4D  inline v4Prod       (const CVect4D& v4V1,const CVect4D& v4V2);        // term by term *
    friend CVect4D  inline v4Div        (const CVect4D& v4V1,const CVect4D& v4V2);        // term by term /

    // special functions

    FLOAT  inline   fLength             (void) const;                                     // vector length d
    FLOAT  inline   fLength2            (void) const;                                     // vector squared length d^2
    bool   inline   boNormalize         (const FLOAT fNorm = 1.0f);                       // normalize
    bool   inline   boNormalize1        (void);                                           // normalize to 1.f - dedicated version

    // set

    inline void     vSet                (const FLOAT    fD);
    inline void     vSet                (const FLOAT    fX, const FLOAT fY,const FLOAT fZ,const FLOAT fW);
    inline void     vSet                (const CVect4D& v4V); 
    inline void     vSet                (const CVect3D& v3V,const FLOAT fW);

    // friends classes

    friend class    CVect3D;
    friend class    CMat4x4;

    friend CMat4x4  inline operator  *  (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1*M2 accesses fV

    friend FLOAT    inline operator  *  (const CVect4D& v4V, const CVect3D& v3V);         // V4*V3 accesses v3V
    friend FLOAT    inline operator  *  (const CVect3D& v3V, const CVect4D& v4V);         // V3*V4 accesses v3V
    friend CVect4D  inline operator  *  (const CMat4x4& m4M, const CVect3D& v3V);         // M4*V3 accesses fV
    friend CVect4D  inline operator  *  (const CVect3D& v3V, const CMat4x4& m4M);         // V3*M4 accesses fV

// protected members

  protected:

    FLOAT           m_fV[4];                                                              // x/y/z/w coords
  };

/**********************************************************************************************************************/
/*                                                3*3 MATRIX                                                          */
/**********************************************************************************************************************/

class     CMat3x3
  {
// public methods

  public:

    // constructors

    inline          CMat3x3             (void);
    inline          CMat3x3             (const FLOAT    fD);                              // all values=fD
    inline          CMat3x3             (const CVect3D& v3V0,const CVect3D& v3V1,const CVect3D& v3V2);        
                                                                                          // 3 lines of the matrix
    inline          CMat3x3             (const CMat3x3& m3M);                             // copy 
  #ifdef _DANGEROUS_CAST_
    inline          CMat3x3             (const CMat4x4& m4M);                             // cast m4 to m3 (rotation part)
  #endif

    // assignment operators

    CMat3x3&               operator   = (const CMat3x3& m3M);                             // assignment
    CMat3x3&               operator  += (const CMat3x3& m3M);                             // incrementation
    CMat3x3&               operator  -= (const CMat3x3& m3M);                             // decrementation
    CMat3x3&               operator  *= (const FLOAT    fK);                              // multiplication by a constant
    CMat3x3&               operator  *= (const CMat3x3& m3M);                             // multiplication
    CMat3x3&               operator  /= (const FLOAT    fK);                              // division by a constant
    CMat3x3&               operator  /= (const CMat3x3& m3M);                             // multiplication by inverse
    CVect3D&               operator  [] (const DWORD    dwN);                             // indexing line
    const CVect3D          operator  [] (const DWORD    dwN) const;                       // indexing line

    // friends operators

    friend CMat3x3  inline operator   - (const CMat3x3& m3M);                             // -M
    friend CMat3x3  inline operator   * (const FLOAT    fK,  const CMat3x3& m3M);         // k*M
    friend CMat3x3  inline operator   * (const CMat3x3& m3M, const FLOAT    fK);          // M*k
    friend CMat3x3  inline operator   / (const CMat3x3& m3M, const FLOAT    fK);          // M/k
    friend CMat3x3  inline operator   - (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1-M2
    friend CMat3x3  inline operator   + (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1+M2
    friend CMat3x3  inline operator   * (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1*M2 
    friend bool     inline operator  == (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1==M2
    friend bool     inline operator  != (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1!=M2

    friend ostream&        operator  << (ostream& s,      CMat3x3& m3M);                  // output
    friend ostream&        operator  << (ostream& s,const CMat3x3& m3M);                  // output

    // friends functions

    friend void     inline vSwap        (      CMat3x3& m3M1,      CMat3x3& m3M2);        // swap
    friend CMat3x3  inline m3Prod       (const CMat3x3& m3M1,const CMat3x3& m3M2);        // term by term *
    friend CMat3x3  inline m3Div        (const CMat3x3& m3M1,const CMat3x3& m3M2);        // term by term /
    friend CMat3x3  inline m3MxtM       (const CMat3x3& m3M1,const CMat3x3& m3M2);        // M1*tM2
    friend CMat3x3  inline m3tMxM       (const CMat3x3& m3M1,const CMat3x3& m3M2);        // tM1*M2

    // special functions

    CMat3x3         inline m3Transpose  (void) const;                                     // transpose
    CMat3x3                m3Inverse    (void) const;                                     // inverse
    void            inline vIdentity3D  (void);                                           // 3D identity

    // friends classes

    friend class    CVect3D;
    friend class    CMat4x4;

    friend CVect3D  inline operator  *  (const CMat3x3& m3M,const CVect3D& v3V);          // M*V accesses v3V
    friend CVect3D  inline operator  *  (const CVect3D& v3V,const CMat3x3& m3M);          // V*M accesses v3V

// protected members

  protected:

    CVect3D         m_v3V[3];                                                             // 3 x/y/z lines
  };

/**********************************************************************************************************************/
/*                                                4*4 MATRIX                                                          */
/**********************************************************************************************************************/

class     CMat4x4
  {
// public methods

  public:

    // constructors

    inline          CMat4x4             (void);
    inline          CMat4x4             (const FLOAT    fD);                              // all values=fD
    inline          CMat4x4             (const CVect4D& v4V0,const CVect4D& v4V1,const CVect4D& v4V2,const CVect4D& v4V3);        
                                                                                          // 4 lines of the matrix
    inline          CMat4x4             (const CMat4x4& m4M);                             // copy 
  #ifdef _DANGEROUS_CAST_
    inline          CMat4x4             (const CMat3x3& m3M);                             // cast m3 to m4
  #endif

    // assignment operators

    CMat4x4&               operator   = (const CMat4x4& m4M);                             // assignment
    CMat4x4&               operator  += (const CMat4x4& m4M);                             // incrementation
    CMat4x4&               operator  -= (const CMat4x4& m4M);                             // decrementation
    CMat4x4&               operator  *= (const FLOAT    fK);                              // multiplication by a constant
    CMat4x4&               operator  *= (const CMat4x4& m4M);                             // multiplication
    CMat4x4&               operator  /= (const FLOAT    fK);                              // division by a constant
    CMat4x4&               operator  /= (const CMat4x4& m4M);                             // multiplication by inverse
    CVect4D&               operator  [] (const DWORD    dwN);                             // indexing line
    const CVect4D          operator  [] (const DWORD    dwN) const;                       // indexing

    // friends operators

    friend CMat4x4  inline operator   - (const CMat4x4& m4M);                             // -M
    friend CMat4x4  inline operator   * (const FLOAT    fK,  const CMat4x4& m4M);         // k*M
    friend CMat4x4  inline operator   * (const CMat4x4& m4M, const FLOAT    fK);          // M*k
    friend CMat4x4  inline operator   / (const CMat4x4& m4M, const FLOAT    fK);          // M/k
    friend CMat4x4  inline operator   - (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1-M2
    friend CMat4x4  inline operator   + (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1+M2
    friend CMat4x4  inline operator   * (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1*M2 
    friend bool     inline operator  == (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1==M2
    friend bool     inline operator  != (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1!=M2

    friend ostream&        operator  << (ostream& s,      CMat4x4& m4M);                  // output
    friend ostream&        operator  << (ostream& s,const CMat4x4& m4M);                  // output

    // friends functions

    friend void     inline vSwap        (      CMat4x4& m4M1,      CMat4x4& m4M2);        // swap
    friend CMat4x4  inline m4Prod       (const CMat4x4& m4M1,const CMat4x4& m4M2);        // term by term *
    friend CMat4x4  inline m4Div        (const CMat4x4& m4M1,const CMat4x4& m4M2);        // term by term /
    friend CMat4x4  inline m4MxtM       (const CMat4x4& m4M1,const CMat4x4& m4M2);        // M1*tM2
    friend CMat4x4  inline m4tMxM       (const CMat4x4& m4M1,const CMat4x4& m4M2);        // tM1*M2

    // special functions

    CMat4x4         inline m4Transpose  (void) const;                                     // transpose
    CMat4x4                m4Inverse    (void) const;                                     // inverse
    void            inline vIdentity4D  (void);                                           // 4D identity

    // friends classes

    friend class    CVect4D;
    friend class    CMat3x3;

    friend CVect4D  inline operator  *  (const CMat4x4& m4M,const CVect4D& v4V);          // M*V accesses v4V
    friend CVect4D  inline operator  *  (const CVect4D& v4V,const CMat4x4& m4M);          // V*M accesses v4V
    friend CVect4D  inline operator  *  (const CMat4x4& m4M,const CVect3D& v3V);          // M*V accesses v3V
    friend CVect4D  inline operator  *  (const CVect3D& v3V,const CMat4x4& m4M);          // V*M accesses v3V

    FLOAT* lpf() { return (FLOAT*)&m_v4V[0]; }

// protected members

  protected:

    CVect4D         m_v4V[4];                                                             // 4 x/y/z/w lines
  };

/************************************************ FUNCTIONS PROTOS ****************************************************/

// transformations

inline    CMat3x3             m3Identity3D                  (void);                       // 3D identity
inline    CMat4x4             m4Identity4D                  (void);                       // 4D identity

CMat3x3   m3Translation2D     (const CVect2D& v2V);                                       // 2D translation (3*3 matrix)
CMat4x4   m4Translation3D     (const CVect3D& v3V);                                       // 3D translation (4*4 matrix)
CMat3x3   m3Rotation2D        (const CVect2D& v2Center,const FLOAT fAngleRad);            // 2D rotation around a point (3*3 matrix)
CMat4x4   m4Rotation3D        (const CVect3D& v3Axis,  const FLOAT fAngleRad);            // 3D rotation around an axis (4*4 matrix)
CMat3x3   m3Scale2D           (const CVect2D& v2Scale);                                   // 2D scale (3*3 matrix)
CMat4x4   m4Scale3D           (const CVect3D& v3Scale);                                   // 3D scale (4*4 matrix)
CMat4x4   m4Perspective3D     (const FLOAT    fD);                                        // 3D perspective (4*4 matrix)

// calculations

void      vMultHomoMatrix     (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // homogeneous multiplication
void      vTransposeHomoMatrix(CMat4x4* lpm4Mt,const CMat4x4& m4M);                       // transpose a 4*4 transformation
void      vRotate3DVector4x4  (CVect3D* lpv3Vp,const CMat4x4& m4M, const CVect3D& v3V);   // rotate 3D vector with 4*4 matrix
void      vCalcRotMatrix      (CMat4x4* lpm4M, const CVect3D& v3Angles);                  // 4x4 rotation matrix from Bryant angles
void      vCalcRotMatrix      (CMat3x3* lpm3M, const CVect3D& v3Angles);                  // 3x3 rotation matrix from Bryant angles

// 2D

void      vAdd                (CVect2D* lpv2V, const CVect2D& v2V1);                      // V+=V1
void      vSub                (CVect2D* lpv2V, const CVect2D& v2V1);                      // V-=V1
void      vMul                (CVect2D* lpv2V, const FLOAT    fK);                        // V*=k
void      vMul                (CVect2D* lpv2V, const CVect2D& v2V1);                      // V*=V1 (term by term)
void      vDiv                (CVect2D* lpv2V, const FLOAT    fK);                        // V/=k
void      vDiv                (CVect2D* lpv2V, const CVect2D& v2V1);                      // V/=V1 (term by term)

void      vNeg                (CVect2D* lpv2V, const CVect2D& v2V1);                      // V=-V1
void      vMul                (CVect2D* lpv2V, const CVect2D& v2V1,const FLOAT    fK);    // V=V1*k
void      vDiv                (CVect2D* lpv2V, const CVect2D& v2V1,const FLOAT    fK);    // V=V1/k
void      vSub                (CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2);  // V=V1-V2
void      vAdd                (CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2);  // V=V1+V2
FLOAT     fDotProd            (                const CVect2D& v2V1,const CVect2D& v2V2);  // f=V1*V2
FLOAT     fCrossProd          (                const CVect2D& v2V1,const CVect2D& v2V2);  // f=V1^V2

void      vMul                (CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2);  // V=V1*V2 (term by term)
void      vDiv                (CVect2D* lpv2V, const CVect2D& v2V1,const CVect2D& v2V2);  // V=V1/V2 (term by term)

// 3D

void      vAdd                (CVect3D* lpv3V, const CVect3D& v3V1);                      // V+=V1
void      vSub                (CVect3D* lpv3V, const CVect3D& v3V1);                      // V-=V1
void      vMul                (CVect3D* lpv3V, const FLOAT    fK);                        // V*=k
void      vMul                (CVect3D* lpv3V, const CVect3D& v3V1);                      // V*=V1 (term by term)
void      vDiv                (CVect3D* lpv3V, const FLOAT    fK);                        // V/=k
void      vDiv                (CVect3D* lpv3V, const CVect3D& v3V1);                      // V/=V1 (term by term)

void      vNeg                (CVect3D* lpv3V, const CVect3D& v3V1);                      // V=-V1
void      vMul                (CVect3D* lpv3V, const CVect3D& v3V1,const FLOAT    fK);    // V=V1*k
void      vDiv                (CVect3D* lpv3V, const CVect3D& v3V1,const FLOAT    fK);    // V=V1/k
void      vSub                (CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2);  // V=V1-V2
void      vAdd                (CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2);  // V=V1+V2
FLOAT     fDotProd            (                const CVect3D& v3V1,const CVect3D& v3V2);  // f=V1*V2
FLOAT     fDotProd            (                const CVect4D& v4V1,const CVect3D& v3V2);  // f=V1*V2
void      vCrossProd          (CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2);  // V=V1^V2

void      vMul                (CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2);  // V=V1*V2 (term by term)
void      vDiv                (CVect3D* lpv3V, const CVect3D& v3V1,const CVect3D& v3V2);  // V=V1/V2 (term by term)
void      vMul                (CVect3D* lpv3V, const CMat3x3& m3M ,const CVect3D& v3V1);  // V=M*V1
void      vMul                (CVect3D* lpv3V, const CVect3D& v3V1,const CMat3x3& m3M);   // V=V1*M
void      vMul                (CVect4D* lpv4V, const CMat4x4& m4M ,const CVect3D& v3V1);  // V=M*V1
void      vMul                (CVect4D* lpv4V, const CVect3D& v3V1,const CMat4x4& m4M);   // V=V1*M

// 4D

void      vAdd                (CVect4D* lpv4V, const CVect4D& v4V1);                      // V+=V1
void      vSub                (CVect4D* lpv4V, const CVect4D& v4V1);                      // V-=V1
void      vMul                (CVect4D* lpv4V, const FLOAT    fK);                        // V*=k
void      vMul                (CVect4D* lpv4V, const CVect4D& v4V1);                      // V*=V1 (term by term)
void      vDiv                (CVect4D* lpv4V, const FLOAT    fK);                        // V/=k
void      vDiv                (CVect4D* lpv4V, const CVect4D& v4V1);                      // V/=V1 (term by term)

void      vNeg                (CVect4D* lpv4V, const CVect4D& v4V1);                      // V=-V1
void      vMul                (CVect4D* lpv4V, const CVect4D& v4V1,const FLOAT    fK);    // V=V1*k
void      vDiv                (CVect4D* lpv4V, const CVect4D& v4V1,const FLOAT    fK);    // V=V1/k
void      vSub                (CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2);  // V=V1-V2
void      vAdd                (CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2);  // V=V1+V2
FLOAT     fDotProd            (                const CVect4D& v4V1,const CVect4D& v4V2);  // f=V1*V2

void      vMul                (CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2);  // V=V1*V2 (term by term)
void      vDiv                (CVect4D* lpv4V, const CVect4D& v4V1,const CVect4D& v4V2);  // V=V1/V2 (term by term)
void      vMul                (CVect4D* lpv4V, const CMat4x4& m4M ,const CVect4D& v4V1);  // V=M*V1
void      vMul                (CVect4D* lpv4V, const CVect4D& v4V1,const CMat4x4& m4M);   // V=V1*M

// 3*3

void      vAdd                (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M+=M1
void      vSub                (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M-=M1
void      vMul                (CMat3x3* lpm3M, const FLOAT    fK);                        // M*=k
void      vMul                (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M*=M1
void      vDiv                (CMat3x3* lpm3M, const FLOAT    fK);                        // M/=k
void      vDiv                (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M/=M1 (* inverse)

void      vNeg                (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M=-M1
void      vMul                (CMat3x3* lpm3M, const CMat3x3& m3M1,const FLOAT    fK);    // M=M1*k
void      vDiv                (CMat3x3* lpm3M, const CMat3x3& m3M1,const FLOAT    fK);    // M=M1/k
void      vSub                (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1-M2
void      vAdd                (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1+M2
void      vMul                (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1*M2

void      vTranspose          (CMat3x3* lpm3M, const CMat3x3& m3M1);                      // M=tM1
void      vMul                (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1*M2 (term by term)
void      vDiv                (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1/M2 (term by term)
void      vMulMxtM            (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=M1*tM2
void      vMultMxM            (CMat3x3* lpm3M, const CMat3x3& m3M1,const CMat3x3& m3M2);  // M=tM1*M2

// 4*4

void      vAdd                (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M+=M1
void      vSub                (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M-=M1
void      vMul                (CMat4x4* lpm4M, const FLOAT    fK);                        // M*=k
void      vMul                (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M*=M1
void      vDiv                (CMat4x4* lpm4M, const FLOAT    fK);                        // M/=k
void      vDiv                (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M/=M1 (* inverse)

void      vNeg                (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M=-M1
void      vMul                (CMat4x4* lpm4M, const CMat4x4& m4M1,const FLOAT    fK);    // M=M1*k
void      vDiv                (CMat4x4* lpm4M, const CMat4x4& m4M1,const FLOAT    fK);    // M=M1/k
void      vSub                (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1-M2
void      vAdd                (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1+M2
void      vMul                (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1*M2

void      vTranspose          (CMat4x4* lpm4M, const CMat4x4& m4M1);                      // M=tM1
void      vMul                (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1*M2 (term by term)
void      vDiv                (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1/M2 (term by term)
void      vMulMxtM            (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=M1*tM2
void      vMultMxM            (CMat4x4* lpm4M, const CMat4x4& m4M1,const CMat4x4& m4M2);  // M=tM1*M2

/************************************************ INLINE **************************************************************/

//#include  "CVectMat.inl"

/**********************************************************************************************************************/

#endif
